/*  $Id: CLex.cc,v 1.26 1998/07/05 20:32:54 shaggy Exp $
 *      C/C++ Lexical Analysis
 *      
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

  
//  $Log: CLex.cc,v $
//  Revision 1.26  1998/07/05 20:32:54  shaggy
//  tons of bugfixes, enhancements and parser improvements
//
//  Revision 1.25  1998/07/04 20:38:56  shaggy
//  lots of bug fixes; fn, aggr, etc improvements, new/delete
//
//  Revision 1.24  1998/06/30 18:19:00  shaggy
//  more fixes for member pointers
//
//  Revision 1.23  1998/06/29 01:36:51  shaggy
//  fixed inline attr, output of pure virtuals, etc...
//
//  Revision 1.22  1998/06/05 03:34:17  shaggy
//  parser fixes
//
//  Revision 1.21  1998/06/04 13:40:36  shaggy
//  more fixes
//
//  Revision 1.20  1998/06/03 13:39:11  shaggy
//  function re-parsing almost works (for within classes)
//
//  Revision 1.19  1998/05/25 23:28:42  shaggy
//  major rework of func/param info
//
//  Revision 1.18  1998/05/17 23:49:20  shaggy
//  can handle GNU-style __attribute__ clause (so we can parse the
//  standard headers)
//
//  Revision 1.17  1998/05/17 23:14:55  shaggy
//  expressions, expression lists, etc
//
//  Revision 1.16  1998/05/16 23:30:38  shaggy
//  added CTOR_ID and some rules to help fix ctor ambiguity problem
//  output turned off for #include'ed files
//  better error checking, and ability to list locations
//  of symbol declaration and definition
//
//  Revision 1.15  1998/05/15 00:17:24  shaggy
//  ctor initializers work, and enums are fixed so declarations
//  in an enum can be immediately re-used
//
//  Revision 1.14  1998/05/10 18:53:40  shaggy
//  just about ready to start fixing scoping
//
//  Revision 1.13  1998/05/10 00:03:30  shaggy
//  string concatenation works, and comments are being
//  added to get ready to fix scoping
//
//  Revision 1.12  1998/05/07 14:46:37  shaggy
//  const/volatile partially working
//
//  Revision 1.11  1998/04/26 14:08:56  shaggy
//  cleanup
//
//  Revision 1.10  1998/04/15 13:39:43  shaggy
//  added types to constants
//
//  Revision 1.9  1998/04/13 03:15:25  shaggy
//  various updates
//
//  Revision 1.8  1998/04/12 03:19:59  shaggy
//  improvments for function output
//
//  Revision 1.7  1998/03/23 14:17:01  shaggy
//  lots more changes
//
//  Revision 1.6  1998/02/08 19:07:05  shaggy
//  about to start splitting up everything
//
//  Revision 1.5  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.4  1998/01/11 05:06:33  shaggy
//  about to switch to standard ref/ptr semantics
//
//
//  Revision 1.3  1996/10/19  21:48:45  shaggy
//  removed action tokens
//
//  Revision 1.2  1996/10/19  16:43:11  shaggy
//  about to move around token declarations
//
//  Revision 1.1  1996/03/04  21:26:43  shaggy
//  Initial revision
//


#define IMPLEMENTATION_CLex
#include <CLex>


CLex::CLex(char *s, size_t ssize)
{
	str = s; str_size = ssize;
	buf = s; buf_size = ssize;
	name = s; name_size = ssize;
	state = S_NEXT;
	_kw = _cpp = _eol = _ws = _indent = false;
	_cxx = true;
	_first = true;
	_include_level = 0;
}


// XXX/OPT - to speed keyword matching, create a huge group of
//	 nested switch statements that go through the identifier to
//	 try to match it to a known keyword
//	 Or make another class for high-speed searching that
//	 groups everything according to each successive letter

// XXX - should join together all 'id' chars into one token
//	 first, then see if it's a number, id, keyword,
//	 or invalid

// XXX - create separate 'L' state that checks for ' or "
//	 and sets _long flag for the string


// Notes:
//
// Yes, this is a decidedly ugly and difficult-to-update
// way of doing this, but since this is specifically designed
// for doing lexical analysis of C/C++ programs, and 
// since the lexical specification of those languages doesn't
// change too often, it was written this way to be as fast
// as possible (and not dependent upon any extra tools,
// such as 'lex') and to be callable in a way that didn't
// require the lexer to allocate memory, or have knowledge
// of how to obtain its data
//
// This is also a completely re-entrant way of doing this,
// since multiple CLex objects can be created
//
// This makes it really easy to use CLex to recursively
// parse when necessary, and to have complete control
// over the lexer
//
// Note that most compilers will convert the huge 'switch'
// statements into tables with only a couple direct compares
// needed, so it won't be an inefficient list of individual compares,
// and will instead be table-driven, just like the output of 'lex'
// except that the code will be called directly instead of being
// first looked up and then used to select the correct code
// Also, the order of the entries doesn't usually matter, as
// most compilers will reorder the entries, and will ignore
// small gaps, generating separate tables only when the size of
// the gap is larger that the separate comparison code 
// Perhaps some compilers might even allow optimizing
// the switch statement into one table for speed, despite
// the holes.


// XXX - need support for floats
//	 .9  3.  3.4  
//	 12E3  1.2E4
//	 12E+3  1.2E-4
//	 4.52f  4.34l

// XXX - Long string literals
//	L"blah" L's'


CLex::Token
CLex::next_token()
{
// Macros for processing tokens and switching states:
//	RC: Return/Continue: return token
//	RBS: Return/Break/State: return token, eat char, state = s
//	RCS: Return/Continue/State: return token, state = s
//	RBNEXT: Return/Break/NEXT: return token, eat char, state = S_NEXT
//	RCNEXT: Return/Cont/NEXT: return token, state = S_NEXT
//	BS: Break/State: eat char, state = s
//	OBS: Old/Skip/State: old_state = state, eat char, state = s
//	CS: Continue/State: state = s
//	HANDLE_WS: Eat/Store whitespace, returning it if necessary
#define RC(t) { return (token = t); }
#define RBS(t,s) { old_state = s; state = S_SKIP; return (token = t); }
#define RCS(t,s) { state = s; return (token = t); }
#define RBNEXT(t) { state = S_SKIP_NEXT; return (token = t); }
#define RCNEXT(t) { state = S_NEXT; return (token = t); }
#define BS(s) { state = s; break; }
#define OBS(s) { old_state = state; state = s; break; }
#define CS(s) { state = s; continue; }
#define OCS(s) { old_state = state; state = s; continue; }
#define HANDLE_WS { if(is_whitespace()) { strp = str; OCS(S_SPACE) }}

bool was_first;

    while(!is_empty()) {
	switch(state) {
	case S_NEXT:
		HANDLE_WS
		was_first = _first;
		_first = false;
		switch(get_char()) {
		case 'A': case 'B': case 'C': case 'D':
		case 'E': case 'F': case 'G': case 'H':
		case 'I': case 'J': case 'K': case 'L':
		case 'M': case 'N': case 'O': case 'P':
		case 'Q': case 'R': case 'S': case 'T':
		case 'U': case 'V': case 'W': case 'X':
		case 'Y': case 'Z': 
		case 'a': case 'b': case 'c': case 'd':
		case 'e': case 'f': case 'g': case 'h':
		case 'i': case 'j': case 'k': case 'l':
		case 'm': case 'n': case 'o': case 'p':
		case 'q': case 'r': case 's': case 't':
		case 'u': case 'v': case 'w': case 'x':
		case 'y': case 'z': 
		case '_':
			strp = str;
			*strp++ = get_char();
			BS(S_ID)
		case '0':
			num = 0;
			_signed = true; _long = 0;
			BS(S_NUM_START)
		case '1': case '2': case '3':
		case '4': case '5': case '6': case '7':
		case '8': case '9':
			num = get_digit();
			_signed = true; _long = 0;	
			BS(S_NUM)
		case '(': RBNEXT(LPAREN)
		case ')': RBNEXT(RPAREN)
		case '[': RBNEXT(LBRACK)
		case ']': RBNEXT(RBRACK)
		case '{': RBNEXT(LBRACE)
		case '}': RBNEXT(RBRACE)
		case ',': RBNEXT(COMMA)
		case '?': RBNEXT(QMARK)
		case ';': RBNEXT(EOS)
		case '~': RBNEXT(OP_NOT)
		case '@': RBNEXT(AT)
		case '\\': RBNEXT(BSLASH)
		case '\'':
			_long = 0; strp = str; BS(S_SQUOTE)
		case '"':
			_long = 0; strp = str; BS(S_DQUOTE)
		case '`':
			_long = 0; strp = str; BS(S_BQUOTE)
		case '.': BS(S_DOT)
		case ':': BS(S_COLON)
		case '!': BS(S_OP_LNOT)
		case '+': BS(S_OP_ADD)
		case '-': BS(S_OP_SUB)
		case '*': BS(S_OP_MUL)
		case '/': BS(S_OP_DIV)
		case '^': BS(S_OP_XOR)
		case '&': BS(S_OP_AND)
		case '|': BS(S_OP_OR)
		case '<': BS(S_OP_LSS)
		case '>': BS(S_OP_GTR)
		case '=': BS(S_OP_ASN)
		case '%': 
			if(sol()) { _qwrap = false; RBS(QLINE, S_LINE) }
			else BS(S_OP_MOD)
		case '#':
			if(was_first) {
				_in_cpp = _cpp_first = true;
				if(_cpp) RBS(CPP, S_CPP)
				else BS(S_NOCPP)
			} else RBNEXT(POUND)
		default:
			strp = str;
			*strp++ = get_char();
			*strp = 0;
			size = strp - str;
			RBNEXT(INVALID)
		}
		break;
	case S_SKIP:
		BS(old_state)
	case S_SKIP_NEXT:
		BS(S_NEXT)
	case S_SPACE:
		while(is_whitespace()) {
			if(eol()) break;
			if(_indent || _ws)
				*strp++ = get_char();
			if(!next_char()) RC(INCOMPLETE)
		}
		if(eol()) {
			if(_in_cpp || _eol) {
				_in_cpp = false;
				old_state = S_NEXT;
				RCS(EOL, S_SPACE_EOL)
			} else CS(S_SPACE_EOL)
		}
		if(strp > str && ((_first && _indent) || _ws)) {
			*strp = 0; size = strp - str;
			RCS(SPACE, old_state)
		} else CS(old_state)
	case S_SPACE_EOL:
		_first = true; if(_indent || _ws) strp = str;
		BS(S_SPACE)
	case S_EOL:
		_first = true; strp = str;
		if((_in_cpp && _cpp) || _eol) {
			_in_cpp = false; RBNEXT(EOL) 
		} else BS(S_NEXT)
	case S_NUM:
		if(is_num()) num = num*10 + get_digit();
		else if(match('.')) { fnum = num; num = 1; BS(S_FLOAT) }
		else { size = 1; CS(S_NUM_END) }
		break;
	case S_BIN:
		if(is_bin()) num = num*2 + get_digit();
		else { size = 1; CS(S_NUM_END) }
		break;
	case S_OCT:
		if(is_oct()) num = num*8 + get_digit();
		else { size = 1; CS(S_NUM_END) }
		break;
	case S_HEX:
		if(is_hex()) num = num*16 + get_hex();
		else { size = 1; CS(S_NUM_END) }
		break;
	case S_NUM_START:
		if(match('x')) BS(S_HEX)
		else if(match('b')) BS(S_BIN)
		else CS(S_OCT)
	case S_NUM_END:
		if(match('u') || match('U')) { _signed = false; break; }
		else if(match('l') || match('L')) { ++_long; break; }
		else RCNEXT(NUM)
	case S_FLOAT:
		if(is_num()) {
			num /= 10;
			fnum += get_digit()*num;
		} else RCNEXT(FLOAT)
	case S_ID:
		if(is_id()) {
			*strp++ = get_char();
			size++;
		} else {
			*strp = 0; size = strp-str;
			if(_kw) {
			    if(_in_cpp) {
				if(!strcmp(str,
				  token_table[CPP_DEFINED-ID]))
					RCNEXT(CPP_DEFINED)
				else
					RCNEXT(ID)
			    }
			    int stop = (_dcxx ? END_DCXX : _cxx ? END_CXX : END_C);
			    for(int t = START_C; t<=stop; t++)
				if(!strcmp(token_table[t-ID], str)) 
					RCNEXT(Token(t))
			}
			RCNEXT(ID)
		}
		break;
	case S_SQUOTE:
		if(match('\\')) OBS(S_QUOTE_ESC)
		else if(match('\'')) {
			*strp = 0; size = strp-str; RBNEXT(CHAR)
		} else { *strp++ = get_char(); break; }
	case S_DQUOTE:
		if(match('\\')) OBS(S_QUOTE_ESC)
		else if(match('"')) BS(S_DQUOTE2)
		else { *strp++ = get_char(); break; }
	case S_DQUOTE2:
		if(is_whitespace()) OCS(S_SPACE);
		if(match('"')) BS(S_DQUOTE)
		else { *strp = 0; size = strp-str; RCNEXT(STR) }
	case S_BQUOTE:
		if(match('\\')) OBS(S_QUOTE_ESC)
		else if(match('`')) {
			*strp = 0; size = strp-str; RBNEXT(BACKSTR)
		} else { *strp++ = get_char(); break; }
	case S_AQUOTE:
		if(match('\\')) OBS(S_QUOTE_ESC)
		else if(match('>')) {
			*strp = 0; size = strp-str; RBNEXT(ANGSTR)
		} else { *strp++ = get_char(); break; }
	case S_QUOTE_ESC:
		state = old_state;
		if(_in_cpp) {
			if(!eol()) {
				*strp++ = '\\';
				*strp++ = get_char();
			}
		} else switch(get_char()) {
		case 'a': *strp++ = '\a'; break;
		case 'b': *strp++ = '\b'; break;
		case 'd': *strp++ = '\177'; break;
		case 'e': *strp++ = '\033'; break;
		case 'f': *strp++ = '\f'; break;
		case 'n': *strp++ = '\n'; break;
		case 'r': *strp++ = '\r'; break;
		case 't': *strp++ = '\t'; break;
		case 'v': *strp++ = '\v'; break;
		case '\n': break;
		case 'x': num = 0; BS(S_QUOTE_HEX)
		case '0' ... '9':
			num = get_digit(); BS(S_QUOTE_OCT)
		default:
			*strp++ = get_char();
		}
		break;
	case S_QUOTE_OCT:
		if(is_oct()) { num = num*8 + get_digit(); break; }
		else { *strp++ = num; CS(old_state) }
	case S_QUOTE_HEX:
		if(is_hex()) { num = num*16 + get_hex(); break; }
		else { *strp++ = num; CS(old_state) }
	case S_OP_ADD:
		if(match('+')) RBNEXT(OP_INC)
		else if(match('=')) RBNEXT(OP_ASNADD)
		else RCNEXT(OP_ADD)
	case S_OP_SUB:
		if(match('-')) RBNEXT(OP_DEC)
		else if(match('=')) RBNEXT(OP_ASNSUB)
		else if(match('>')) BS(S_PTRTO)
		else RCNEXT(OP_SUB)
	case S_OP_MUL:
		if(match('=')) RBNEXT(OP_ASNMUL)
		else RCNEXT(OP_MUL)
	case S_OP_DIV:
		if(match('*')) BS(S_COMMENT)
		else if(match('/')) { _qwrap = false; BS(S_EAT) }
		else if(match('=')) RBNEXT(OP_ASNDIV)
		else RCNEXT(OP_DIV)
	case S_OP_MOD:
		if(match('=')) RBNEXT(OP_ASNMOD)
		else RCNEXT(OP_MOD)
	case S_OP_AND:
		if(match('&')) RBNEXT(OP_LAND)
		else if(match('=')) RBNEXT(OP_ASNAND)
		else RCNEXT(OP_AND)
	case S_OP_OR:
		if(match('|')) RBNEXT(OP_LOR)
		else if(match('=')) RBNEXT(OP_ASNOR)
		else RCNEXT(OP_OR)
	case S_OP_XOR:
		if(match('^')) RBNEXT(OP_LXOR)
		else if(match('=')) RBNEXT(OP_ASNXOR)
		else RCNEXT(OP_XOR)
	case S_OP_LNOT:
		if(match('=')) RBNEXT(OP_NEQ)
		else RCNEXT(OP_LNOT)
	case S_OP_ASN:
		if(match('=')) RBNEXT(OP_EQU)
		else RCNEXT(OP_ASN)
	case S_OP_LSS:
		if(match('<')) BS(S_OP_LSHIFT)
		else if(match('=')) RBNEXT(OP_LEQ)
		else RCNEXT(OP_LSS)
	case S_OP_GTR:
		if(match('>')) BS(S_OP_RSHIFT)
		else if(match('=')) RBNEXT(OP_GEQ)
		else RCNEXT(OP_GTR)
	case S_OP_LSHIFT:
		if(match('=')) RBNEXT(OP_ASNLSHIFT)
		else RCNEXT(OP_LSHIFT)
	case S_OP_RSHIFT:
		if(match('=')) RBNEXT(OP_ASNRSHIFT)
		else RCNEXT(OP_RSHIFT)
	case S_COLON:
		if(match(':')) RBNEXT(OP_SCOPE)
		else RCNEXT(COLON)
	case S_DOT:
		if(match('.')) BS(S_DOT2)
		else if(match('*')) RBNEXT(OP_MDOT)
		else RCNEXT(OP_DOT)
	case S_DOT2:
		if(match('.')) RBNEXT(ELIPSIS)
		else RCNEXT(DOTDOT)
	case S_PTRTO:
		if(match('*')) RBNEXT(OP_MPTRTO)
		else RCNEXT(OP_PTRTO)
	case S_CPP:
		HANDLE_WS
		if(is_id_start()) {
			strp = str;
			*strp++ = get_char();
			BS(S_CPP_CMD)
		} else if(is_num()) {
			num = get_digit();
			BS(S_CPP_NUM)
		} else if(match('"')) {
			namep = name; strp = str;
			BS(S_CPP_NAME)
		} else
			CS(S_NEXT)
	case S_NOCPP:
		HANDLE_WS
		if(is_num()) {
			num = get_digit();
			BS(S_CPP_NONUM)
		} else if(match('"')) {
			namep = name; 
			BS(S_CPP_NONAME)
		} else { _qwrap = true; BS(S_EAT) }
	case S_CPP_NUM:
	case S_CPP_NONUM:
		if(is_num()) { num = num*10 + get_digit(); break; }
		else {
			if(_cpp_first) set_line_num(num-1);
			_cpp_first = false;
			if(state == S_CPP_NUM) RCS(NUM, S_CPP)
			else CS(S_NOCPP)
		}
	case S_CPP_NAME:
		if(match('"')) {
			*namep = *strp = 0; size = strp - str;
			num = 0;
			RBS(STR, S_CPP_OPTNUM) 
		} else { *namep++ = *strp++ = get_char(); break; }
	case S_CPP_NONAME:
		if(match('"')) {
			*namep = 0;
			num = 0;
			BS(S_CPP_NOOPTNUM)
		} else { *namep++ = get_char(); break; }
	case S_CPP_OPTNUM:
	case S_CPP_NOOPTNUM:
		if(num == 0) HANDLE_WS
		if(is_num()) { num = num*10 + get_digit(); break; }
		else {	/* XXX - will never get here if num==0 */
			/* note: relying on a GNU-cpp feature here... */
			if(num == 1)
				++_include_level;
			else if(num == 2)
				--_include_level;
			HANDLE_WS
			BS(S_NEXT);
		}
	case S_CPP_CMD:
		if(is_id()) {
			*strp++ = get_char();
			++size;
		} else {
			*strp = 0; size = strp-str;
			if(_kw) {
			    for(int t = START_CPP; t<=END_CPP; t++)
				if(!strcmp(token_table[t-ID], str)) 
					RCS(Token(t), S_CPP_ARG1)
			}
			RCS(ID, S_CPP_ARG1)
		}
		break;
	case S_CPP_ARG1:
		HANDLE_WS
		if(match('<')) { strp = str; BS(S_AQUOTE) }
		else CS(S_NEXT)
	case S_COMMENT:
		if(match('*')) BS(S_COMMENT2)
		break;
	case S_COMMENT2:
		if(match('/')) BS(S_NEXT)
		else BS(S_COMMENT)
	case S_EAT:
		if(match('\\')) OBS(S_SKIP)
		else if(match('"') && _qwrap) BS(S_EAT_STR)
		else if(eol()) CS(S_EOL)
		else break;
	case S_EAT_STR:
		if(match('\\')) OBS(S_SKIP)
		else if(match('"')) BS(S_EAT)
		else break;
	case S_LINE:
		if(match('\\')) OBS(S_QUOTE_ESC)
		else if(match('"') && _qwrap) BS(S_LINE_STR)
		else if(eol()) {
			_in_cpp = true;
			*strp = 0; size = strp-str;
			RCS(LINE, S_EOL);
		} else { *strp++ = get_char(); break; }
	case S_LINE_STR:
		if(match('\\')) OBS(S_QUOTE_ESC)
		else if(match('"')) BS(S_LINE)
		else { *strp++ = get_char(); break; }
// #define TRY_KW(k,n) { token = k; matchp = &token[k].str[n]; BS(S_KW) }
//	case NEXT:
//		if(match('v')) BS(S_V)
//	case S_V: if(match('o')) BS(S_VO) else BS(S_ID)
//	case S_VO: if(match('i')) TRY_KW(KW_VOID,3) 
//		else if(match('l')) TRY_KW(KW_VOLATILE,3)
//	case S_KW:
//		if(!*matchp) { if(!is_id()) RCNEXT(token) }
//		else if(match(*matchp++)) break;
//		BS(S_ID)
//	CASE_KW(VO,'i',KW_VOID) OR_KW('i',KW_VOLATILE) KW_BREAK_KW ??
	}
	next_char();
    }
    if(state == S_NEXT || state == S_SPACE || state == S_SPACE_EOL)
	if(eof()) RC(EOF_COMPLETE)
	else RC(NONE)
    else
	if(eof()) RC(EOF_INCOMPLETE)
	else RC(INCOMPLETE)

#undef RC
#undef RBS
#undef RCS
#undef RBNEXT
#undef RCNEXT
#undef BS
#undef OBS
#undef CS
#undef HANDLE_WS
}


uint
CLex::last_char() const
{
uint num = 0;
csz *cp = str;
	if(*cp) do {
		num += *cp;
	} while(*++cp && ((num <<= 8), true)); 
	return num;
}


csz*
CLex::token_string()
{
	switch(token) {
	case NONE: case INCOMPLETE:
		return "";
	case NUM:
		return token_string(token, last_num(), last_signed(),
					last_long());
	case BOOL:
		return token_string(token, last_bool());
	case INVALID:
	case CHAR: case STR: case CSTR: case BACKSTR: case ANGSTR:
	case ID: case LINE: case SPACE:
		return token_string(token, last_str(), last_size());
	case EOL: return "\n"; 
	case EOS: return ";";
	case QLINE: return "%";
	case CPP: return "#"; 
	default:
		return token_string(token);
	}
}


csz*
CLex::token_string(Token t) const
{
	if(t >= ID && t < NUM_TOKENS+23)	/* XXX - +23 = hack for CAction */
		return token_table[t-ID];
	else if(t == NONE)
		return "<NONE>";
	else
		return "<??>";
}


csz*
CLex::token_string(Token t, bool b) const
{
	if(t != BOOL)
		return "<??>";
	if(b)
		return "true";
	else
		return "false";
}


csz*
CLex::token_string(Token t, uint64 val,
		bool is_signed, int is_long)
{
	if(t != NUM)
		return "<??>";
	bufp = buf;
	if(val < 10) 
		*bufp++ = '0'+val;
	else {
		*bufp++ = '0';
		*bufp++ = 'x';
		register char *endp = bufp;
		for(uint64 n = val;
		    (*endp = "0123456789abcdef"[n&0xf]), n&~0x0f; n>>=4)
			endp++;
		register char *bp = endp++;
		while(bufp < bp) {
			char t = *bufp; *bufp++ = *bp; *bp-- = t;
		}
		bufp = endp;
	}
	if(!is_signed) *bufp++ = 'U';
	while(is_long--) *bufp++ = 'L';
	*bufp = 0;
	return buf;
}


csz*
CLex::token_string(Token t, register csz *s, size_t len)
{
	char quote_start, quote_end;

	switch(t) {
	case INVALID:
	case CHAR: quote_start = quote_end = '\''; break;
	case CSTR:
	case STR: quote_start = quote_end = '"'; break;
	case BACKSTR: quote_start = '`'; quote_end = '`'; break;
	case ANGSTR: quote_start = '<'; quote_end = '>'; break;
	case ID: quote_start = quote_end = 0; break;
	case LINE: case SPACE: return s;
	default:
		return "<??>";
	}
	if(len == NPOS) {
		len = 0;
		for(register csz *sp = s; *sp; sp++)
			len++;
	}
	bufp = buf;
	if(quote_start)
		*bufp++ = quote_start;
	while(len--) {
		char ch = *s++;
		if(ch < 32 || ch > 126) {
			*bufp++ = '\\';
			switch(ch) {
			case '\a': *bufp++ = 'a'; break;
			case '\b': *bufp++ = 'b'; break;
			case '\f': *bufp++ = 'f'; break;
			case '\n': *bufp++ = 'n'; break;
			case '\r': *bufp++ = 'r'; break;
			case '\t': *bufp++ = 't'; break;
			case '\v': *bufp++ = 'v'; break;
			case '\0': *bufp++ = '0'; break;
			default:
				uint n = (uchar)ch;
				*bufp++ = '0'+(n>>6);
				*bufp++ = '0'+((n>>3)&7);
				*bufp++ = '0'+(n&7);
			}
		} else if(ch == quote_end || ch == '\\') {
			*bufp++ = '\\';
			*bufp++ = ch;
		} else {
			*bufp++ = ch;
		}
	}
	if(quote_end)
		*bufp++ = quote_end;
	*bufp = 0;
	return buf;
}


csz*
CLex::token_spacing(Token prev, Token next) const
{
	switch(prev) {
	case NONE: case INCOMPLETE: case INVALID:
	case STR: case CSTR: case ANGSTR: case BACKSTR: case CHAR:
	case LINE: case QLINE:
	case CPP: case EOL:
	case OP_DOT: case OP_SCOPE: case ELIPSIS:
	case KW_FOR: case KW_IF: case KW_WHILE:
	case KW_SWITCH:
	case OP_SIZEOF: case OP_TYPEOF:
	case OP_NOT: case OP_LNOT:
	case OP_INC: case OP_DEC:
	case LPAREN: case LBRACK:
		if(next == OP_ASN)
			return " ";
	case SPACE: 
		return "";
	case NUM: case ID:
		switch(next) {
		case OP_ADD: case OP_SUB: case OP_MUL: case OP_DIV:
		case OP_AND: case OP_OR: case OP_XOR:
			return "";
		default:
			goto def;
		}
	case OP_ADD: case OP_SUB: case OP_MUL: case OP_DIV:
	case OP_AND: case OP_OR: case OP_XOR:
		switch(next) {
		case NUM: case ID:
			return "";
		default:
			goto def;
		}
	default:
	def:
		switch(next) {
		case EOL: case EOS:
		case OP_DOT: case COMMA: case COLON: case OP_SCOPE:
		case LBRACK: case RBRACK:
		case LPAREN: case RPAREN:
		case OP_INC: case OP_DEC:
			if(prev == OP_ASN)
				return " ";
		case SPACE:
			return "";
		default:
			return " ";
		}
	}
}


// Note that all of the tokens in this table
// must be in the exact same order as their
// corresponding token names in CParse.y

csz *CLex::token_table[] = {
	"<ID>", "<NUM>", "<CHAR>", "<FLOAT>", "<STR>", "<CSTR>", "<BOOL>",
	"<BACKSTR>", "<ANGSTR>", "<LINE>", "<SPACE>", "<EOL>",

	"<CPP>", "##", "defined", 

	"define", "undef", "if", "ifdef", "ifndef", "endif", "elif",
	"include", "pragma", "assert", "unassert",
	"line", "error", "warning", "ident",
	"__FILE__", "__LINE__", "__DATE__", "__TIME__",
	"__STDC__", "__STDC_VERSION__",

	"<EOS>", ":", ",", "?", "#",
	"{", "}", "[", "]", "(", ")",

	"<QLINE>", "..", "@", "\\",

	"+", "-", "*", "/", "%",
	"~", "&", "|", "^", "!", "&&", "||", "^^",
	"<", ">", "<=", ">=", "==", "!=",
	"++", "--", "<<", ">>",
	"=", "+=", "-=", "*=", "/=", "%=", "&=", "|=", "^=", "<<=", ">>=",
	".", "->",

	"::", "...", ".*", "->*",

	"?=",

	"__attribute__",
	"const", "volatile", "auto", "static", "extern", "register",
	"long", "short", "signed", "unsigned",
	"int", "char", "float", "double", "void",
	"if", "else", "do", "while", "for", "switch",
	"case", "default", "break", "continue", "goto", "return",
	"enum", "union", "struct", "typedef",
	"sizeof",
	"__FUNCTION__", "__PRETTY_FUNCTION__",

	"operator", 
	"class", "this",
	"typeid", "typename", "namespace", "using",
	"template", "public", "private", "protected", "friend",
	"virtual", "mutable", "explicit",
	"bool", "true", "false", "wchar_t", "throw", "catch", "try",
	"static_cast", "dynamic_cast", "const_cast", "reinterpret_cast",
	"inline", "asm",
	"typeof", "new", "delete", "new[]", "delete[]",

	"dynamic", 

// for debugging only: (not generated by lexer)
	"<VAL_ID>", "<AGGR_ID>", "<ENUM_ID>", "<TYPE_ID>",
	"<CTOR_ID>", "<DTOR_ID>", "<OPER_ID>", "<CONV_ID>",
	"<NSPACE_ID>", "<ASPACE_ID>", "<TPLATE_ID>",
	"<SCOPE>", "<TOKENLIST>", 

	"<OP_REF>", "<OP_PTR>", "<OP_MPTR>", "<OP_UMINUS>", "<OP_UPLUS>", 
	"<OP_INCP>", "<OP_DECP>", "<OP_TSIZEOF>",
	"<OP_COMMA>", "<OP_FUNC>", "<OP_INDEX>", "<OP_CAST>",

	"dclass", "dtype",

	"<NEXT_INSN>", "<BACKPATCH>", "<ACTION>", "<TBC>",
	"<LBRACE_SAVE>", "<COLON_SAVE>", "<PARSE_COMPOUND>", "<DEFINE>",

// from CAction:
	"<TEMP>", "<LITERAL>", "<RESULT>", "<INSN>",
	"<LIST>", "<TREE>", "<ARRAY>", "<OBJ>",
	"<TYPE>", "<TYPELIST>", "<PARM>", "<PARMLIST>",
	"<EXPR>", "<EXPRLIST>", "<STMT>", "<STMTLIST>",
	"<ERR>", "<ERRLIST>",
	"<DECL>",
	"<STRUCT>", "<CLASS>", "<VAR>", "<FUNC>",
	"<CTORINIT>", 
};

