/*  $Id: CParse.y,v 1.90 1998/07/11 23:06:46 shaggy Exp $
 *      C/C++ Parser (shacc grammar)
 *      
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  
 *  $Log: CParse.y,v $
 *  Revision 1.90  1998/07/11 23:06:46  shaggy
 *  fixed labels and goto
 *
 *  Revision 1.89  1998/07/11 19:02:58  shaggy
 *  declarations in for() work now (even multiple decls)
 *
 *  Revision 1.88  1998/07/10 13:45:07  shaggy
 *  working on decls in conditions
 *
 *  Revision 1.87  1998/07/09 14:34:38  shaggy
 *  first phase of fixes for decl's in if(), while(), etc
 *
 *  Revision 1.86  1998/07/07 13:41:40  shaggy
 *  enum cleanup, and can define types inside functions now
 *
 *  Revision 1.85  1998/07/06 13:39:03  shaggy
 *  many more changes/fixes for first pre-release
 *
 *  Revision 1.84  1998/07/05 20:32:55  shaggy
 *  tons of bugfixes, enhancements and parser improvements
 *
 *  Revision 1.83  1998/07/04 20:38:58  shaggy
 *  lots of bug fixes; fn, aggr, etc improvements, new/delete
 *
 *  Revision 1.82  1998/07/04 01:23:29  shaggy
 *  anonymous unions work now
 *
 *  Revision 1.81  1998/06/30 18:19:01  shaggy
 *  more fixes for member pointers
 *
 *  Revision 1.80  1998/06/30 12:22:31  shaggy
 *  many fixes for "this"
 *
 *  Revision 1.79  1998/06/29 14:14:16  shaggy
 *  more DECO conversions, support for 'this', etc
 *
 *  Revision 1.78  1998/06/29 01:36:52  shaggy
 *  fixed inline attr, output of pure virtuals, etc...
 *
 *  Revision 1.77  1998/06/08 04:07:30  shaggy
 *  more fixes
 *
 *  Revision 1.76  1998/06/07 15:45:10  shaggy
 *  realtype() works, and . and -> now work
 *
 *  Revision 1.75  1998/06/05 03:34:18  shaggy
 *  parser fixes
 *
 *  Revision 1.74  1998/06/04 13:40:38  shaggy
 *  more fixes
 *
 *  Revision 1.73  1998/06/04 12:52:21  shaggy
 *  reparsing seems to work now
 *
 *  Revision 1.72  1998/06/03 13:39:13  shaggy
 *  function re-parsing almost works (for within classes)
 *
 *  Revision 1.71  1998/06/02 14:50:40  shaggy
 *  in the midst of more full parsing updates
 *
 *  Revision 1.70  1998/06/01 13:28:59  shaggy
 *  switchable code save/parse
 *
 *  Revision 1.69  1998/06/01 12:28:58  shaggy
 *  Fixed dangling-else problem (for IF/IF/ELSE)
 *
 *  Revision 1.68  1998/06/01 02:10:34  shaggy
 *  more code parsing working
 *
 *  Revision 1.67  1998/05/31 20:58:01  shaggy
 *  got reduction precedence working
 *
 *  Revision 1.66  1998/05/26 05:02:53  shaggy
 *  parameter scoping works correctly
 *  most basic C/C++ statements now work
 *
 *  Revision 1.65  1998/05/26 01:39:46  shaggy
 *  params almost working right
 *
 *  Revision 1.64  1998/05/25 23:28:43  shaggy
 *  major rework of func/param info
 *
 *  Revision 1.63  1998/05/24 18:41:59  shaggy
 *  improved regression testing
 *  proper scoping for initializers
 *  decls available for their initializer
 *
 *  Revision 1.62  1998/05/17 23:49:21  shaggy
 *  can handle GNU-style __attribute__ clause (so we can parse the
 *  standard headers)
 *
 *  Revision 1.61  1998/05/17 23:14:55  shaggy
 *  expressions, expression lists, etc
 *
 *  Revision 1.60  1998/05/16 23:30:41  shaggy
 *  added CTOR_ID and some rules to help fix ctor ambiguity problem
 *  output turned off for #include'ed files
 *  better error checking, and ability to list locations
 *  of symbol declaration and definition
 *
 *  Revision 1.59  1998/05/15 07:03:36  shaggy
 *  dependency ordering fixed; constructor initializers; lots of bug fixes, etc
 *
 *  Revision 1.58  1998/05/15 00:17:25  shaggy
 *  ctor initializers work, and enums are fixed so declarations
 *  in an enum can be immediately re-used
 *
 *  Revision 1.57  1998/05/14 18:02:03  shaggy
 *  fixed constant handling, enums, function cv tags, etc
 *
 *  Revision 1.56  1998/05/14 05:08:40  shaggy
 *  inheritence works and lots of bugs fixed
 *
 *  Revision 1.55  1998/05/13 00:21:15  shaggy
 *  fixed enums (mostly)
 *
 *  Revision 1.54  1998/05/12 05:00:12  shaggy
 *  more scope work
 *
 *  Revision 1.53  1998/05/10 18:53:41  shaggy
 *  just about ready to start fixing scoping
 *
 *  Revision 1.52  1998/05/10 00:03:31  shaggy
 *  string concatenation works, and comments are being
 *  added to get ready to fix scoping
 *
 *  Revision 1.51  1998/05/07 14:46:38  shaggy
 *  const/volatile partially working
 *
 *  Revision 1.50  1998/05/05 13:00:41  shaggy
 *  better function parameter matching
 *
 *  Revision 1.49  1998/05/04 03:09:18  shaggy
 *  support for ctors, dtors, conversion operators, operator overloading, etc
 *
 *  Revision 1.48  1998/05/03 02:47:07  shaggy
 *  added inheritence to parser
 *
 *  Revision 1.47  1998/04/27 04:36:47  shaggy
 *  started auto-generation of code
 *
 *  Revision 1.46  1998/04/26 15:21:23  shaggy
 *  better implicit int handling
 *
 *  Revision 1.45  1998/04/26 14:08:57  shaggy
 *  cleanup
 *
 *  Revision 1.44  1998/04/23 13:47:50  shaggy
 *  changes to make enums and constants work better
 *
 *  Revision 1.42  1998/04/22 13:06:38  shaggy
 *  better enum handling
 *
 *  Revision 1.41  1998/04/16 13:35:44  shaggy
 *  access control
 *
 *  Revision 1.40  1998/04/15 14:48:11  shaggy
 *  array parsing works
 *
 *  Revision 1.39  1998/04/15 13:39:44  shaggy
 *  added types to constants
 *
 *  Revision 1.38  1998/04/14 13:19:16  shaggy
 *  improved declaration parsing
 *
 *  Revision 1.37  1998/04/14 03:57:19  shaggy
 *  forward reference fix
 *
 *  Revision 1.36  1998/04/13 13:16:47  shaggy
 *  added %uniq and reorged tasks
 *
 *  Revision 1.35  1998/04/13 03:15:26  shaggy
 *  various updates
 *
 *  Revision 1.34  1998/04/12 03:20:00  shaggy
 *  improvments for function output
 *
 *  Revision 1.33  1998/04/11 21:49:27  shaggy
 *  added token saving (for regurgitation)
 *
 *  Revision 1.32  1998/04/11 01:47:40  shaggy
 *  deco uses cpp
 *  tests fixed for dc++
 *  dc++ grammar fixed
 *
 *  Revision 1.31  1998/04/06 14:28:07  shaggy
 *  got enums almost working
 *
 *  Revision 1.29  1998/03/23 14:17:02  shaggy
 *  lots more changes
 *
 *  Revision 1.28  1998/03/22 03:14:56  shaggy
 *  almost there...
 *
 *  Revision 1.27  1998/03/21 00:23:49  shaggy
 *  All kinds of new changes...
 *  and it works!!!
 *
 *  Revision 1.26  1998/03/20 21:52:21  shaggy
 *  most skeleton code now added (except templates, exceptions)
 *
 *  Revision 1.25  1998/03/20 17:36:26  shaggy
 *  all notes from Appendix A are in
 *
 *  Revision 1.24  1998/03/19 14:00:23  shaggy
 *  switching over to new parser
 *
 *  Revision 1.1  1998/02/23 14:33:52  shaggy
 *  working on "shacc"
 *
 *
 */


%{//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#pragma Make Exclusive

#include "ActionStack.h"
#include <TokenList>

#define IMPLEMENTATION_CParse
#include "CParse.hh"


#define YYSTYPE CParse::Item


#define yyassert(c)	{ if(!(c)) INTERNAL_ERROR; }
#define yyerror 	_caction->error
#define yywarning 	_caction->warning

#define YYCLASS		CParse
#define YYPARSE		_do_parse
#define YYLEX		_peekahead_lex
#define YYDEBUGFLAG	_yydebug
#define YYEPRINTF	_caction->eprintf
#define YYPARSE_VARS	\
	ActionStack actionstack(_caction);

//#define YYPDEBUG


// XXX - eventually, all of the constant strings (for errors/warnings)
// should be removed and replaced with error ID's
// that can be passed to CAction where they can have defaults
// and overrides.  This will also aid internationalization.
// The position specifier needs to be added to printf for this
// to entirely work correctly.



#define swap_items(a,b)	{ Item tmp = a; a = b; b = tmp; }


#define eval_decl(r, a, t) \
	{ actionstack.pop_lifo(&(r), (a), decl_actions, (t)); }
#define eval_pdecl(r, a, t) \
	{ actionstack.pop_lifo(&(r), (a), pdecl_actions, (t)); }

#define declSym(r, s, d) \
	{ actionstack.push(&(r), none, ACTION_Decl, T(TBC), (d), (s)); }
#define defineSym(r, s, v) \
	{ actionstack.push(&(r), none, ACTION_Define, (s), (v), \
						none); }
#define declPtr(r, op, d) \
	{ actionstack.push(&(r), (d), ACTION_Ptr, (op), T(TBC)); }
#define declArray(r, d, i) \
	{ actionstack.push(&(r), (d), ACTION_Array, T(TBC), i, 0); }
#define declFunc(r, d, p, c) \
	{ actionstack.push(&(r), (d), ACTION_Func, T(TBC), (p), (c)); }
#define declInit(r, d, v) \
	{ actionstack.push(&(r), (d), ACTION_Init, T(TBC), (p)); }
#define buildParamList(r, l, i) \
	{ actionstack.push(&(r), none, ACTION_BuildList, (l), (i)); }
#define buildDeclList(r, l, i) \
	{ actionstack.push(&(r), none, ACTION_None, (l), (i)); }

#define declInherit(r, t, b) \
	{ actionstack.push(&(r), none, ACTION_Inherit, T(TBC), (t), (b)); }


#define UNSUPPORTED yyerror("unsupported syntax (%d)", __LINE__);
#define NOSCOPE(s) if(s.scope) yyerror("scope not allowed");



enum ActionIndex {
	ACTION_None = 0,
	ACTION_Decl,
	ACTION_Define,
	ACTION_Ptr,
	ACTION_Array,
	ACTION_Func,
	ACTION_BuildList,
	ACTION_Inherit,
};


static const CAction::Action decl_actions[] = {
	0,		/* none */
	&A(DeclSym),	/* declarator */
	&A(Define),	/* define initial value */
	&A(DeclPtr),	/* pointer (*) or reference (&) declaration */
	&A(DeclArray),	/* array declaration ([n]) */
	&A(DeclFunc),	/* function declaration */
	&A(BuildParamList), /* parameter list (,) */
	&A(Inherit)	/* inherit from a base class */
};

static const CAction::Action pdecl_actions[] = {
	0,		/* none */
	&A(DeclParam),	/* declarator */
	&A(DefParam),	/* default value */
	&A(DeclPtr),	/* pointer (*) or reference (&) declaration */
	&A(DeclArray),	/* array declaration ([n]) */
	&A(DeclFunc),	/* function declaration */
	&A(BuildParamList), /* parameter list (,) */
	0,		/* none */
};


/* Note: when defining parameter lists, the current scope,
	which is used for each parameter, can either be some arbitrary
	undefined scope, for arbitrary declarations,
	or the scope of the function, for the header on the definition */

/* Need a 'kill_scope' to get rid of unneeded scopes later, and
   reference counts should probably be used on all items so
   that those items referenced in a scope which are still needed 
   will be available */


#define curscope	_current_scope
#define savescope_new(s,f,t,i)	{ s = _current_scope; \
		do3(_current_scope, f, t, i, s); }
#define savescope_newlocal(s)	{ s = _current_scope; \
		do3(_current_scope, NewLocal, none, none, s); }
#define savescope_newfunc(s,d)	{ s = _current_scope; \
		do2(_current_scope, FuncScope, d, s); }
#define savescope_set(s, n)	{ s = _current_scope; _current_scope = n; }
#define savescope(s)		{ s = _current_scope; }
#define getscope_restore(g, r)	savescope_set(g, r)
#define restorescope(s) { _current_scope = s; }

%}//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


// The exact sequence of the following tokens must
// exactly match the sequence in token_table[] in CLex.C 
// Any additional tokens should be added above
// this line, or added appropriately to CLex.C

// Generic tokens:

%token _ID		// identifier
%token _NUM		// integer-literal
%token _CHAR		// character-literal
%token _FLOAT		// floating-literal
%token _STR		// string-literal
%token _CSTR		// constant string (for internal use)
%token _BOOL		// boolean-literal
%token _BACKSTR		// `string`
%token _ANGSTR		// <string>
%token _LINE		// % line
%token _SPACE		// whitespace
%token _EOL		// end of line

// CPP tokens:

%token _CPP		// #
%token _CPP_JOIN	// ##
%token _CPP_DEFINED 	// defined

// CPP keywords:

%token _CPP_DEFINE	// #define
%token _CPP_UNDEF	// #undef
%token _CPP_IF		// #if
%token _CPP_IFDEF	// #ifdef
%token _CPP_IFNDEF	// #ifndef
%token _CPP_ENDIF	// #endif
%token _CPP_ELIF	// #elif
%token _CPP_INCLUDE	// #include
%token _CPP_PRAGMA	// #pragma
%token _CPP_ASSERT	// #assert
%token _CPP_UNASSERT	// #unassert
%token _CPP_LINE	// #line
%token _CPP_ERROR	// #error
%token _CPP_WARNING	// #warning
%token _CPP_IDENT	// #ident

// CPP special substitution tokens
%token _CPP__FILE	// __FILE__
%token _CPP__LINE	// __LINE__
%token _CPP__DATE	// __DATE__
%token _CPP__TIME	// __TIME__
%token _CPP__STDC	// __STDC__
%token _CPP__STDCVER	// __STDC_VERSION__

// standard C punctuation:

%token _EOS _COLON _COMMA		// ; : ,
%token _QMARK _POUND _LBRACE _RBRACE	// ? # { }
%token _LBRACK _RBRACK _LPAREN _RPAREN	// [ ] ( )

// extended punctuation:

%token _QLINE _DOTDOT _AT _BSLASH	// % .. @ \ 

// standard C operators:

%token _OP_ADD _OP_SUB			// + -
%token _OP_MUL _OP_DIV _OP_MOD		// * / %
%token _OP_NOT _OP_AND _OP_OR _OP_XOR	// ~ & | ^
%token _OP_LNOT _OP_LAND _OP_LOR _OP_LXOR // ! && || ^^
%token _OP_LSS _OP_GTR _OP_LEQ _OP_GEQ	// < > <= >=
%token _OP_EQU _OP_NEQ _OP_INC _OP_DEC	// == != ++ --
%token _OP_LSHIFT _OP_RSHIFT		// << >>
%token _OP_ASN				// =
%token _OP_ASNADD _OP_ASNSUB		// += -=
%token _OP_ASNMUL _OP_ASNDIV _OP_ASNMOD	// *= /= %=
%token _OP_ASNAND _OP_ASNOR _OP_ASNXOR	// &= |= ^=
%token _OP_ASNLSHIFT _OP_ASNRSHIFT	// <<= >>=
%token _OP_DOT _OP_PTRTO		// . ->

// C++ operators:

%token _OP_SCOPE _ELIPSIS _OP_MDOT _OP_MPTRTO	// :: ... .* ->*

// DC++ operators:

%token _OP_CONDASN 			// ?=

// standard C keywords:

%token _KW_ATTRIBUTE			// GNU-ish extension
%token _KW_CONST _KW_VOLATILE
%token _KW_AUTO _KW_STATIC _KW_EXTERN _KW_REGISTER
%token _KW_LONG _KW_SHORT _KW_SIGNED _KW_UNSIGNED
%token _KW_INT _KW_CHAR _KW_FLOAT _KW_DOUBLE _KW_VOID
%token _KW_IF _KW_ELSE _KW_DO _KW_WHILE _KW_FOR
%token _KW_SWITCH _KW_CASE _KW_DEFAULT
%token _KW_BREAK _KW_CONTINUE _KW_GOTO
%token _KW_RETURN
%token _KW_ENUM _KW_UNION _KW_STRUCT _KW_TYPEDEF
%token _OP_SIZEOF

// special C substitution strings:

%token _KW__FUNC	// __FUNCTION__
%token _KW__PFUNC	// __PRETTY_FUNCTION__

// C++ keywords:

%token _KW_OPERATOR 
%token _KW_CLASS _KW_THIS
%token _KW_TYPEID _KW_TYPENAME
%token _KW_NAMESPACE _KW_USING
%token _KW_TEMPLATE
%token _KW_PUBLIC _KW_PRIVATE _KW_PROTECTED _KW_FRIEND
%token _KW_VIRTUAL _KW_MUTABLE _KW_EXPLICIT
%token _KW_BOOL _KW_TRUE _KW_FALSE
%token _KW_WCHAR
%token _KW_THROW _KW_CATCH _KW_TRY
%token _KW_SCAST _KW_DCAST _KW_CCAST _KW_RCAST 
%token _KW_INLINE _KW_ASM
%token _OP_TYPEOF
%token _OP_NEW _OP_DELETE
%token _OP_NEWA	_OP_DELETEA	// new[] delete[]

// extended (DC++) keywords:

%token _KW_DYNAMIC

/////////////////////////////////////////////////////////

// Additional tokens not generated by original lexer,
// but substituted in, and used for semantic parsing:

%token _VAL_ID		// value (not type) identifier
%token _AGGR_ID		// aggregate identifier (class,struct,union)
%token _ENUM_ID		// enumerator identifier
%token _TYPE_ID		// typedef'ed identifier
%token _CTOR_ID		// constructor ID (matches enclosing scope)
%token _DTOR_ID		// destructor ID ( _OP_NOT _AGGR_ID )
%token _OPER_ID		// operator ID	(operator*  , etc...)
%token _CONV_ID		// conversion ID  (operator int  , etc...)
%token _NSPACE_ID	// namespace identifier
%token _ASPACE_ID	// alias namespace identifier
%token _TPLATE_ID	// template identifier
%token _SCOPE		// scope specification
%token _TOKENLIST	// a list of tokens saved for parsing later

/////////////////////////////////////////////////////////

// Additional tokens not generated by lexer, and not used for parsing:


// unary C/C++ operators

%token _OP_REF				// &
%token _OP_PTR				// *
%token _OP_MPTR				// ::*
%token _OP_UMINUS			// -
%token _OP_UPLUS			// +
%token _OP_INCP				// ++
%token _OP_DECP				// --
%token _OP_TSIZEOF			// sizeof(type)

// other C/C++ operators

%token _OP_COMMA	// ,
%token _OP_FUNC		// ()
%token _OP_INDEX	// []
%token _OP_CAST		// (old_style_cast)()

// various compound keywords

%token _KW_DCLASS	// dynamic class (dclass)
%token _KW_DTYPE	// dynamic virtual class (dtype)

// special tokens

%token _NEXT_INSN	// use next instruction in sequence
%token _BACKPATCH	// this will get backpatched later
%token _ACTION		// used to maintain internal action lists
%token _TBC		// to be computed (later)

// special tokens for parse control
%token _LBRACE_SAVE
%token _COLON_SAVE
%token _PARSE_COMPOUND	// to parse a compound statement (function body)
%token _DEFINE

%%//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


// For the syntax descriptions,
// the section numbers refer to "The C++ Programming Language" Third Edition
// by Bjarne Stroustrup. (1997)

// unquoted names represent other syntactical parts
// (which are italicized in the document), and quoted names
// represent literal text (keywords)
// any item followed by `(opt)' is optional.

// suffixes:
//	_opt		optional
//	_nodecl 	cannot look like a declarator
//	_nosdec		cannot look like a simple_declaration
//	_notype		cannot look like a type
//	_noptype	cannot have a type in parenthesis

//////////////////////////////////////////////////////////////////////////////
// PROGRAMS (A.4)
//
// translation-unit:
//	declaration-seq(opt)

translation_unit:
	  declaration_seq_opt
	| _PARSE_COMPOUND compound_statement
		{ /* not used yet */ }
	;

//////////////////////////////////////////////////////////////////////////////
// DECLARATIONS (A.7)
//
// declaration-seq:
//	declaration
//	declaration-seq declaration

declaration_seq:
	  declaration_seq_opt { _caction->complete(false); } declaration
		{ _caction->complete(true);
		  if(_deftok && !_savecode) {
			inject(_deftok);
			_deftok = 0;
		  } }
	;

declaration_seq_opt: { $$ = none; } | declaration_seq ;

// declaration:
//	block-declaration
//	function-definition
//	template-declaration
//	explicit-instantiation
//	explicit-specialization
//	linkage-specification
//	namespace-definition

none: { $$ = none; };

declaration:
	  block_declaration 	
	| none init_declarator_list _EOS
	| function_definition 
//x	| template_declaration 
//x	| explicit_instantiation 
//x	| explicit_specialization 
	| linkage_specification 
//x	| namespace_definition 	
	| _DEFINE save_scope id_or_ctor 
		{ if($3.scope) {
			/* XXX - hack */
			curscope.token = T(SCOPE);
			curscope.tuple.right = $3.scope;
		  }
		  $$ = $3; }
	    ctor_initializer_opt function_body
		{ if($5.token != T(NONE))
			do2($6, CtorInit, $5, $6);
		  do2($$, Define, $3, $6);
		  restorescope($2); }
	;

id_or_ctor: identifier | _CTOR_ID;

// block-declaration:
//	simple-declaration
//	asm-definition
//	namespace-alias-definition
//	using-declaration
//	using-directive

block_declaration:
	  simple_declaration
	| asm_directive
//x	| namespace_alias_definition
//x	| using_declaration
//x	| using_directive
	;

// simple-declaration:
//	decl-specifier-seq(opt) init-declarator-list(opt) ;

simple_declaration:
	  decl_specifier_seq init_declarator_list _EOS %rprec 1 %uniq
	| decl_specifier_seq _EOS %rprec 2 %uniq
		{ do3($$, DeclSym, $1, none, curscope); }
	;

// decl-specifier:
//	storage-class-specifier
//	type-specifier
//	function-specifier
//	"friend"
//	"typedef"

decl_specifier_notypename:
	  storage_class_specifier
	| type_specifier_notypename
	| function_specifier
	| _KW_FRIEND
	| _KW_TYPEDEF
	;

// decl-specifier-seq:
//	decl-specifier-seq(opt) decl-specifier

decl_specifier_notypename_seq:
	  decl_specifier_notypename
		{ do2($$, BuildTypeList, none, $1); }
	| decl_specifier_notypename_seq decl_specifier_notypename
		{ do2($$, BuildTypeList, $1, $2); }
	;

decl_specifier_seq:
	  decl_specifier_notypename_seq
	| type_name
		{ do2($$, BuildTypeList, none, $1); }
	| decl_specifier_notypename_seq type_name
		{ do2($$, BuildTypeList, $1, $2); }
	| type_name decl_specifier_notypename_seq
		{ do2($$, BuildTypeList, $2, $1); }
	| decl_specifier_notypename_seq type_name decl_specifier_notypename_seq
		{ Item i;
	          do2(i, BuildTypeList, $1, $2); 
		  do2($$, BuildTypeList, i, $3); }
	;

decl_specifier_seq_opt: { $$ = none; } | decl_specifier_seq ;

// storage-class-specifier:
//	"auto"
//	"register"
//	"static"
//	"extern"
//	"mutable"

storage_class_specifier:
	  _KW_AUTO
	| _KW_REGISTER
	| _KW_STATIC
	| _KW_EXTERN
	| _KW_MUTABLE
	;

// function-specifier:
//	"inline"
//	"virtual"
//	"explicit"

function_specifier:
	  _KW_INLINE
	| _KW_VIRTUAL
	| _KW_EXPLICIT
	;

// type-specifier:
//	simple-type-specifier
//	class-specifier
//	enum-specifier
//	elaborated-type-specifier
//	cv-qualifier

type_specifier_notypename:
	  simple_type_specifier_notypename
	| class_specifier
	| enum_specifier
	| elaborated_type_specifier
	| cv_qualifier
	;

// simple-type-specifier:
//	::(opt) nested-name-specifier(opt) type-name
//	"char"
//	"wchar_t"
//	"bool"
//	"short"
//	"int"
//	"long"
//	"signed"
//	"unsigned"
//	"float"
//	"double"
//	"void"

simple_type_specifier_notypename:
	  _KW_CHAR
	| _KW_WCHAR
	| _KW_BOOL
	| _KW_SHORT
	| _KW_INT
	| _KW_LONG
	| _KW_SIGNED
	| _KW_UNSIGNED
	| _KW_FLOAT
	| _KW_DOUBLE
	| _KW_VOID
	;

// type-name:
//	class-name
//	enum-name
//	typedef-name

type_name:
	  class_name
	| enum_name
	| typedef_name
	;

// elaborated-type-specifier:
//	class-key ::(opt) nested-name-specifier(opt) identifier
//	"enum" ::(opt) nested-name-specifier(opt) identifier
//	"typename" ::(opt) nested-name-specifier identifier
//	"typename" ::(opt) nested-name-specifier identifier 
//				< template-argument-list >

elaborated_type_specifier:
	  class_key identifier
		{ do3($$, AggrRef, $1, $2, curscope);
		  newdecl(&_stmtlist, $$); }
	| _KW_ENUM identifier
		{ do3($$, EnumRef, $1, $2, curscope);
		  newdecl(&_stmtlist, $$); }
//x	| _KW_TYPENAME identifier
//x	| _KW_TYPENAME identifier _OP_LSS template_argument_list _OP_GTR
	;

// enum-specifier:
//	"enum" identifier(opt) { enumerator-list(opt) }

enum_specifier:
	  _KW_ENUM identifier_opt lbrace %uniq
		{ do3(_enum_type, NewEnum, $1, $2, curscope);
		  _enum_val = none;	
		  newdecl(&_stmtlist, _enum_type); }
	    enumerator_list_opt comma_opt _RBRACE
		{ $$ = _enum_type; }
	;

// enumerator-list:
//	enumerator-definition
//	enumerator-list , enumerator-definition

enumerator_list:
	  enumerator_definition
	| enumerator_list _COMMA enumerator_definition
	;

enumerator_list_opt: { $$ = none; } | enumerator_list ;

// enumerator-definition:
//	enumerator
//	enumerator = constant-expression

// enumerator:
//	identifier

enumerator_definition:
	  identifier
		{ Item s, v;
		  do3(s, DeclSym, _enum_type, $1, curscope);
		  do1(v, EnumNext, _enum_val);
		  _enum_val = v;
		  do2($$, Define, s, v); }
	| identifier _OP_ASN constant_expression
		{ do3($2, DeclSym, _enum_type, $1, curscope);
		  _enum_val = $3;
		  do2($$, Define, $2, $3); }
	;

// namespace-definition:
//	named-namespace-definition
//	unnamed-namespace-definition

namespace_definition:
	  named_namespace_definition
	| unnamed_namespace_definition
	;

// named-namespace-definition:
//	original-namespace-definition
//	extension-namespace-definition

named_namespace_definition:
	  original_namespace_definition
	| extension_namespace_definition
	;

// original-namespace-definition:
//	"namespace" identifier { namespace-body }

original_namespace_definition:
	  _KW_NAMESPACE identifier lbrace %uniq
		{ savescope_new($$, NewNspace, $1, $2); } // $3
	      namespace_body _RBRACE %uniq
		{ getscope_restore($$, $3); }
	;

// extension-namespace-definition:
//	"namespace" original-namespace-name { namespace-body }

extension_namespace_definition:
	  _KW_NAMESPACE original_namespace_name lbrace %uniq
		{ savescope_set($$, $2); } // $$ = $4 
	      namespace_body _RBRACE %uniq
		{ getscope_restore($$, $4); }
	;

// unnamed-namespace-definition:
//	"namespace" { namespace-body }

unnamed_namespace_definition:
	  _KW_NAMESPACE lbrace %uniq
		{ savescope_new($$, NewNspace, $1, none); } // $3
	      namespace_body _RBRACE %uniq
		{ getscope_restore($$, $3); }
	;

// namespace-body:
//	declaration-seq(opt)

namespace_body:
	  declaration_seq_opt
	;

// namespace-alias-definition:
//	"namespace" identifier = qualifier-namespace-specifier ;

namespace_alias_definition:
	  _KW_NAMESPACE identifier _OP_ASN qualifier_namespace_specifier _EOS %uniq
		{ UNSUPPORTED /* 1 */ }
	;

// qualified-namespace-specifier:
//	::(opt) nested-name-specifier (opt) namespace-name


// using-declaration:
//	"using" "typename"(opt) ::(opt) nested-name-specifier unqualified-id ;
//	"using" :: unqualified-id ;

using_declaration:
	  _KW_USING _KW_TYPENAME unqualified_id _EOS %uniq
		{ UNSUPPORTED /* 2 */ }
	| _KW_USING unqualified_id _EOS %uniq
		{ UNSUPPORTED /* 3 */ }
	| _KW_USING _SCOPE unqualified_id _EOS %uniq
		{ UNSUPPORTED /* 4 */ }
	;

// using-directive:
//	"using" "namespace" ::(opt) nested-name-specifier(opt) namespace-name ;

using_directive:
	  _KW_USING _KW_NAMESPACE namespace_name _EOS %uniq
		{ UNSUPPORTED /* 5 */ }
	;

// asm-directive:
//	"asm" ( string-literal ) ;

asm_directive:
	  _KW_ASM _LPAREN _STR _RPAREN _EOS %uniq
		{ UNSUPPORTED /* 6 */ }
	;

// linkage-specification:
//	"extern" string-literal { declaration-seq(opt) }
//	"extern" string-literal declaration

linkage_specification:
	  _KW_EXTERN _STR lbrace declaration_seq_opt _RBRACE
	| _KW_EXTERN _STR declaration
	;

//////////////////////////////////////////////////////////////////////////////
// Declarators (A.7.1)

// init-declarator-list:
//	init-declarator
//	init-declarator-list , init-declarator

// init-declarator:
//	declarator initializer(opt)

save_scope:
	  { savescope($$); }
	;

init_declarator_list1:
	  save_scope init_declarator _COMMA save_scope
		{ $$ = $4; }
	| init_declarator_list1 init_declarator _COMMA save_scope
		{ $$ = $4; }
	;

init_declarator_list:
	  save_scope init_declarator
		{ $$ = $1; }
	| init_declarator_list1 init_declarator
		{ $$ = $1; }
	;

init_declarator_list_opt: { $$ = none; } | init_declarator_list;


init_declarator:
	  declarator attribute_opt 
		{ eval_decl($$, $1, $-1);
		  newdecl(&_stmtlist, $$);
		  restorescope($0); }
	| declarator attribute_opt 
		{ eval_decl($$, $1, $-1); 
		  newdecl(&_stmtlist, $$); }
	     initializer 
		{ do2($$, Define, $3, $4); 
		  restorescope($0); 
		  $$ = $3; }
	;
 
// declarator:
//	direct-declarator
//	ptr-operator declarator

declarator:
	  direct_declarator
	| ptr_operator declarator
		{ declPtr($$, $1, $2); }
	;

// direct-declarator:
//	declarator-id
//	direct-declarator ( parameter-declaration-clause ) 
//			cv-qualifier-seq(opt) exception-specification(opt)
//	direct-declarator [ constant-expression(opt) ]
//	( declarator )

direct_declarator:
	  declarator_id
	| direct_declarator _LPAREN parameter_declaration_clause _RPAREN
			cv_qualifier_seq_opt //x exception_specification_opt
		{ declFunc($$, $1, $3, $5); }
	| direct_declarator _LBRACK constant_expression _RBRACK
		{ declArray($$, $1, $3); }
	| direct_declarator _LBRACK _RBRACK
		{ declArray($$, $1, none); }
	| _LPAREN declarator _RPAREN
		{ $$ = $2; }
	;

ctor_declarator_id:
	  _CTOR_ID
		{ do3($1, SpecialId, T(CTOR_ID), $1, curscope);
		  declSym($$, curscope, $1); 
		  if($1.scope) {
			if(curscope.token == T(NONE)) {
				/* XXX - hack */
				curscope.token = T(SCOPE);
				curscope.tuple.right = $1.scope;
			} else 
				yyerror("scope cannot be specified for"
					" non-global constructor");
		  }
		}
	;

ctor_declarator:
	  ctor_declarator_id _LPAREN parameter_declaration_clause _RPAREN
			cv_qualifier_seq_opt //x exception_specification_opt
		{ declFunc($$, $1, $3, $5); }
	;

declarator_or_ctor:
	  ctor_declarator	%rprec 1
	| declarator		%rprec 2
	;

// ptr-operator:
//	* cv-qualifier-seq(opt)
//	&
//	::(opt) nested-name-specifier * cv-qualifier-seq(opt)

ptr_operator:
	  _OP_MUL cv_qualifier_seq_opt
		{ $$ = T(OP_PTR); $$.val = $2.val; }
	| _OP_AND
		{ $$ = T(OP_REF); }
	| _SCOPE _OP_MUL cv_qualifier_seq_opt
		{ $$ = T(OP_MPTR); $$.scope = $1.scope; $$.val = $3.val; }
	;

// cv-qualifier-seq:
//	cv-qualifier cv-qualifier-seq(opt)

cv_qualifier_seq:
	  cv_qualifier cv_qualifier_seq_opt
		{ do2($$, BuildTypeList, $2, $1); }
	;

cv_qualifier_seq_opt: { $$ = none; } | cv_qualifier_seq ;

// cv-qualifier:
//	"const"
//	"volatile"

cv_qualifier:
	  _KW_CONST
	| _KW_VOLATILE
	;

// declarator-id:
//	::(opt) id-expression
//	::(opt) nested-name-specifier(opt) type-name

declarator_id:
	  id_expression
		{ 
		  declSym($$, curscope, $1); 
		  if($1.scope) {
			if(curscope.token == T(NONE)) {
				/* XXX - hack */
				curscope.token = T(SCOPE);
				curscope.tuple.right = $1.scope;
			} else 
				yyerror("scope cannot be specified for"
					" non-global declarator");
		  }
		}
	;

// type-id:
//	type-specifier-seq abstract-declarator(opt)

type_id:
	  type_specifier_seq abstract_declarator_opt
		{ eval_decl($$, $2, $1); }
	;

// type-specifier-seq:
//	type-specifier type-specifier-seq(opt)

type_specifier_notypename_seq:
	  type_specifier_notypename
		{ do2($$, BuildTypeList, none, $1); }
	| type_specifier_notypename_seq type_specifier_notypename
		{ do2($$, BuildTypeList, $1, $2); }
	;

type_specifier_seq:
	  cv_qualifier_seq_opt type_name
		{ do2($$, BuildTypeList, $1, $2); }
	| type_specifier_notypename_seq
	;

// abstract-declarator:
//	ptr-operator abstract-declarator(opt)
//	direct-abstract-declarator

abstract_declarator:
	  ptr_operator abstract_declarator_opt
		{ declPtr($$, $1, $2); }
	| direct_abstract_declarator
	;

abstract_declarator_opt: { $$ = none; } | abstract_declarator ;

// direct-abstract-declarator:
//	direct-abstract-declarator(opt) ( parameter-declaration-clause )
//			cv-qualifier-seq(opt) exception-specification(opt)
//	( abstract-declarator )

direct_abstract_declarator:
	  _LPAREN parameter_declaration_clause
		_RPAREN cv_qualifier_seq_opt //x exception_specification_opt
		{ declFunc($$, none, $2, $4); }
	| direct_abstract_declarator _LPAREN parameter_declaration_clause
		_RPAREN cv_qualifier_seq_opt //x exception_specification_opt
		{ declFunc($$, $1, $3, $5); }
	| _LPAREN abstract_declarator _RPAREN
		{ $$ = $2; }
	;

// parameter-declaration-clause:
//	parameter-declaration-list(opt) ...(opt)
//	parameter-declaration-list , ...

parameter_declaration_clause:	// optional
		{ do3($$, BuildParamList, none, none, curscope); }
	| parameter_declaration_list
	| parameter_declaration_list comma_opt _ELIPSIS
		{ do3($$, BuildParamList, $1, T(ELIPSIS), curscope); }
	| _ELIPSIS
		{ do3($$, BuildParamList, none, T(ELIPSIS), curscope); }
	;

// parameter-declaration-list:
//	parameter-declaration
//	parameter-declaration-list , parameter-declaration

parameter_declaration_list:
	  parameter_declaration
		{ do3($$, BuildParamList, none, $1, curscope); }
	| parameter_declaration_list _COMMA parameter_declaration
		{ do3($$, BuildParamList, $1, $3, curscope); }
	;

// parameter-declaration:
//	decl-specifier-seq declarator
//	decl-specifier-seq declarator = assignment-expression
//	decl-specifier-seq abstract-declarator(opt)
//	decl-specifier-seq abstract-declarator(opt) = assignment-expression

parameter_declaration:
	  decl_specifier_seq declarator _OP_ASN assignment_expression %rprec 1
		{ eval_pdecl($3, $2, $1);
		  do2($$, DefParam, $3, $4); }
	| decl_specifier_seq declarator %rprec 1
		{ eval_pdecl($$, $2, $1); }
	| decl_specifier_seq abstract_declarator_opt
				 _OP_ASN assignment_expression %rprec 2
		{ eval_pdecl($3, $2, $1);
		  do2($1, DeclParam, $3, none);
		  do2($$, DefParam, $1, $4); }
	| decl_specifier_seq abstract_declarator_opt %rprec 2
		{ eval_pdecl($$, $2, $1); $2 = $$;
		  do2($$, DeclParam, $2, none); }
	; 

// function-definition:
//	decl-specifier-seq(opt) declarator ctor-initializor(opt) function-body
//	decl-specifier-seq(opt) declarator function-try-block

eval_declarator_or_ctor:
	  declarator_or_ctor
		{ eval_decl($$, $1, $-1); 
			yywarning("%p,%p,%p\n", &$-1, &$0, &$1); }
	;

function_definition:
	  decl_specifier_seq_opt save_scope declarator_or_ctor 
		{ eval_decl($$, $3, $1); }
	    ctor_initializer_opt function_body
		{ if($5.token == T(TOKENLIST)) {
		  	((TokenList*)$5.ptr)->append_delete((TokenList*)$6.ptr);
			$6 = $5;
		  } else if($5.token != T(NONE))
			do2($6, CtorInit, $5, $6);
		  if(_savecode && !_gsavecode) {
			if(!_deftok)
				_deftok = new TokenList;
			_deftok->append(T(DEFINE));
			$4.scope = curscope.scope;
			_deftok->append($4);
			_deftok->append_delete((TokenList*)$6.ptr);
			$$ = $4;
		  } else 
		  	do2($$, Define, $4, $6);
		  restorescope($2); }
//x	| decl_specifier_seq_opt declarator function_try_block
//x		{ UNSUPPORTED /* 7 */ }
	;


// function-body:
//	compound-statement

function_body:
		{ savescope_newfunc($$, $-1); }	// $1
	    compound_statement
		{ restorescope($1); $$ = $2; }
	| save_compound_statement
	;

// initializer:
//	= initializer-clause
//	( expression-list )

initializer:
	  _OP_ASN initializer_clause
		{ $$ = $2; }
	| _LPAREN expression_list _RPAREN
		{ $$ = $2; }
	;

// initializer-clause:
//	assignment-expression
//	{ initializer-list ,(opt) }
//	{ }

initializer_clause:
	  assignment_expression
	| lbrace initializer_list comma_opt _RBRACE %uniq
		{ $$ = $2; }
	| lbrace _RBRACE %uniq
		{ do2($$, BuildArray, none, none); }
	;

comma_opt: | _COMMA ;

// initializer-list:
//	initializer-clause
//	initializer-list , initializer-clause

initializer_list:
	  initializer_clause
		{ do2($$, BuildArray, none, $1); }
	| initializer_list _COMMA initializer_clause
		{ do2($$, BuildArray, $1, $3); }
	;

//////////////////////////////////////////////////////////////////////////////
// CLASSES (A.8)


// class-specifier:
//	class-head { member-specification(opt) }

// class-head:
//	class-key identifier(opt) base-clause(opt)
//	class-key nested-name-specifier identifier base-clause(opt)

class_specifier:
	  class_key identifier_opt base_clause_opt lbrace %uniq
		{ Item s; do3(s, NewAggr, $1, $2, curscope); 
		  newdecl(&_stmtlist, s);
		  if($3.token != T(NONE)) { Item t;
		  	eval_decl(t, $3, s); s = t; }
		  savescope_set($$, s);
		  $4.bval = _savecode; _savecode = true; } // $5
	     member_specification_opt _RBRACE %uniq
		{ do2($$, EndAggr, curscope, $5);
		  restorescope($5);
		  _savecode = $4.bval; }
	;


// class-key:
//	"class"
//	"struct"
//	"union"
//    DC++ extension:
//	"dynamic" "class"
//	"dynamic" "virtual" "class"

class_key:
	  _KW_CLASS
	| _KW_STRUCT
	| _KW_UNION

	// DC++ extension:
	| _KW_DYNAMIC _KW_CLASS
		{ $$ = T(KW_DCLASS); }
	| _KW_DYNAMIC _KW_VIRTUAL _KW_CLASS
		{ $$ = T(KW_DTYPE); }
	// for re-parsing:
	| _KW_DCLASS
	| _KW_DTYPE
	;

// member-specification:
//	member-declaration member-specification(opt)
//	access-specifier : member-specification(opt)
//    DC++ extension:
//	class-name : member-specification(opt)

member_specification:
	  member_declaration member_specification_opt
	| access_specifier colon
		{ do2($$, SetAccess, curscope, $1); }
		member_specification_opt
	// DC++ extension:
	| class_name colon
		{ do3($1, FindSym, T(AGGR_ID), $1, curscope); 
		  if($1.token != T(NONE)) do2($$, SetAccess, curscope, $1); }
		member_specification_opt
	| _KW_DYNAMIC class_name colon
		{ do3($2, FindSym, T(AGGR_ID), $2, curscope);
		  if($1.token != T(NONE)) do2($$, SetAccess, curscope, $2); }
		member_specification_opt	/* XXX - need to tag with DYNAMIC */
	;

member_specification_opt: { $$ = none; } | member_specification ;

// member-declaration:
//	decl-specifier-seq(opt) member-declarator-list(opt) ;
//	function-definition ;(opt)
//	qualified-id ;
//	using-declaration
//	template-declaration

member_declaration:
	  _EOS %uniq
	| decl_specifier_seq _EOS %uniq
		{ do3($$, DeclSym, $1, none, curscope); }
	| decl_specifier_seq member_declarator_list _EOS %uniq
		{ eval_decl($$, $2, $1); }
	| member_declarator_list _EOS %uniq
		{ eval_decl($$, $1, none); }
	| function_definition
		{ do3($$, SetAttr, $1, T(KW_INLINE), none); }
//x	| qualified_id _EOS %uniq
//x	| using_declaration
//x	| template_declaration
	;
 
// member-declarator-list:
//	member-declarator
//	member-declarator-list , member-declarator

member_declarator_list:
	  member_declarator
	| member_declarator_list _COMMA member_declarator
		{ buildDeclList($$, $1, $3); }
	;

// member-declarator:
//	declarator pure-specifier(opt)
//	declarator constant-initializer(opt)
//	identifier(opt) : constant-expression

// pure-specifier:
//	= 0

// constant-initializer:
//	= constant-expression

member_declarator:
	  declarator_or_ctor
	| declarator _OP_ASN constant_expression
		{ defineSym($$, $1, $3); }
	| colon constant_expression
		{ $$ = none; }
	| identifier colon constant_expression
		{ declSym($$, curscope, $1); }
	;


//////////////////////////////////////////////////////////////////////////////
// Derived Classes (A.8.1)

// base-clause:
//	: base-specifier-list

base_clause_opt:  // optional
		{ $$ = none; }
	| colon base_specifier_list
		{ $$ = $2; }
	;

// base-specifier-list:
//	base-specifier
//	base-specifier-list , base-specifier

base_specifier_list:
	  base_specifier
	| base_specifier_list _COMMA base_specifier
		{ buildDeclList($$, $1, $3); }
	;

// base-specifier:
//	::(opt) nested-name-specifier(opt) class-name
//	"virtual" access-specifier(opt) ::(opt) nested-name-specifier(opt) 
//						class-name
//	access-specifier "virtual"(opt) ::(opt) nested-name-specifier(opt)
//						class-name

base_specifier:
	  class_name
		{ declInherit($$, none, $1); }
	| _KW_VIRTUAL class_name
		{ declInherit($$, $1, $2); }
	| _KW_VIRTUAL access_specifier class_name
		{ $2.tval = T(KW_VIRTUAL); declInherit($$, $2, $3); }
	| access_specifier class_name
		{ Item n = none; n.tval = T(KW_VIRTUAL);
		  declInherit($$, $1, $2); }
	| access_specifier _KW_VIRTUAL class_name
		{ $1.tval = T(KW_VIRTUAL); declInherit($$, $1, $3); }
	;

// access-specifier:
//	"private"
//	"protected"
//	"public"

access_specifier:
	  _KW_PRIVATE
	| _KW_PROTECTED
	| _KW_PUBLIC
	;

//////////////////////////////////////////////////////////////////////////////
// Special Member Functions (A.8.2)

// conversion-function-id:
//	"operator" conversion-type-id

conversion_function_id:
	  _KW_OPERATOR conversion_type_id
		{ do3($$, SpecialId, T(CONV_ID), $2, curscope); }
	;

// conversion-type-id:
//	type-specifier-seq conversion-declarator(opt)

conversion_type_id:
	  type_specifier_seq conversion_declarator_opt
		{ eval_decl($$, $2, $1); }
	;

// conversion-declarator:
//	ptr-operator conversion-declarator(opt)

conversion_declarator:
	  ptr_operator conversion_declarator_opt
		{ declPtr($$, $1, $2); }
	;

conversion_declarator_opt: { $$ = none; } | conversion_declarator ;

// ctor-initializer:
//	: mem-initializer-list

ctor_initializer:
	  _COLON_SAVE 
		{ $$.bval = _rawid; _rawid = true; }
	     save_mem_initializer_list
		{ _rawid = $2.bval;
		  $1.token = T(COLON);
		  ((TokenList*)$3.ptr)->prepend($1); 
		  $$ = $3; }
	| _COLON 
		{ savescope_newfunc($$, $0); } // $2
	     mem_initializer_list
		{ $$ = $3; restorescope($2); }
	;

ctor_initializer_opt: { $$ = none; } | ctor_initializer;

// mem-initializer-list:
//	mem-initializer
//	mem-initializer , mem-initializer-list

save_mem_initializer_list:
	  save_mem_initializer
	| save_mem_initializer_list _COMMA save_mem_initializer
		{ ((TokenList*)$1.ptr)->append($2); 
		  ((TokenList*)$1.ptr)->append_delete((TokenList*)$3.ptr); 
		  $$ = $1; }
	;

mem_initializer_list:
	  mem_initializer
		{ do3($$, BuildStmtList, none, $1, curscope); }
	| mem_initializer_list _COMMA mem_initializer
		{ do3($$, BuildStmtList, $1, $3, curscope); }
	;

// mem-initializer:
//	mem-initializer-id ( expression-list(opt) )

// mem-initializer-id:
//	::(opt) nested-name-specifier(opt) class-name
//	identifier

save_mem_initializer:
	  identifier _LPAREN save_statement_seq_opt _RPAREN
		{ $1.token = T(ID); 
		  ((TokenList*)$3.ptr)->prepend($2);
		  ((TokenList*)$3.ptr)->prepend($1);
		  ((TokenList*)$3.ptr)->append($4);
		  $$ = $3; }
	;

mem_initializer:
	  identifier _LPAREN expression_list_opt _RPAREN
		{ do2($$, CtorCall, $1, $3); }
	;


//////////////////////////////////////////////////////////////////////////////
// Overloading (A.8.3)

// operator-function-id:
//	"operator" operator

operator_function_id:
	  _KW_OPERATOR operator
		{ do3($$, SpecialId, T(OPER_ID), $2, curscope); }
	;

// operator: one of
//   "new" 	"delete" 	"new[]" 	"delete[]"
//   +     -     *     /     %     ^     &     |     ~     !     =     <     >
//   +=    -=    *=    /=    %=    ^=    &=    |=    <<    >>    >>=   <<=   ==
//   !=    <=    >=    &&    ||    ++    --    ,     ->*   ->    ()    []

simple_operator:
	  _OP_NEW | _OP_DELETE | _OP_NEWA | _OP_DELETEA
	| _OP_ADD | _OP_SUB | _OP_MUL | _OP_DIV | _OP_MOD | _OP_XOR | _OP_AND
	| _OP_OR | _OP_NOT | _OP_LNOT | _OP_ASN | _OP_LSS | _OP_GTR
	| _OP_ASNADD | _OP_ASNSUB | _OP_ASNMUL | _OP_ASNDIV | _OP_ASNMOD
	| _OP_ASNXOR | _OP_ASNAND | _OP_ASNOR | _OP_LSHIFT | _OP_RSHIFT
	| _OP_ASNRSHIFT | _OP_ASNLSHIFT | _OP_EQU
	| _OP_NEQ | _OP_LEQ | _OP_GEQ | _OP_LAND | _OP_LOR | _OP_INC
	| _OP_DEC | _COMMA | _OP_MPTRTO | _OP_PTRTO 
	;

operator:
	  simple_operator
	| _LPAREN _RPAREN
		{ $$ = T(OP_FUNC); }
	| _LBRACK _RBRACK
		{ $$ = T(OP_INDEX); }
	;


//////////////////////////////////////////////////////////////////////////////
// TEMPLATES (A.9)

// template-declaration:
//	"export"(opt) "template" < template-parameter-list > declaration

// template-parameter-list:
//	template-parameter
//	template-parameter-list , template-parameter

// template-parameter:
//	type-parameter
//	parameter-declaration

// type-parameter:
//	"class" identifier(opt)
//	"class" identifier(opt) = type-id
//	"typename" identifier(opt)
//	"typename" identifier(opt) = type-id
//	"template" < template-parameter-list > "class" identifier(opt)
//	"template" < template-parameter-list > "class" identifier(opt)
//					 = template-name

// template-id:
//	template-name < template-argument-list >

// template-argument-list:
//	template-argument
//	template-argument-list , template-argument

// template-argument:
//	assignment-expression
//	type-id
//	template-name

// explicit-instantiation:
//	template declaration

// explicit-specialization:
//	"template" < > declaration


//////////////////////////////////////////////////////////////////////////////
// EXCEPTION HANDLING (A.10)

// try-block:
//	"try" compound-statement handler-seq

// function-try-block:
//	"try" ctor-initializer(opt) function-body handler-seq

// handler-seq:
//	handler handler-seq(opt)

// handler:
//	"catch" ( exception-declaration ) compound-statement

// exception-declaration:
//	type-specifier-seq declarator
//	type-specifier-seq abstract-declarator
//	type-specifier-seq
//	...

// throw-expression:
//	"throw" assignment-expression(opt)

// exception-specification:
//	"throw" ( type-id-list(opt) )

// type-id-list:
//	type-id
//	type-id-list , type-id



//////////////////////////////////////////////////////////////////////////////
// EXPRESSIONS (A.5)

// primary-expression:
//	literal
//	"this"
//	:: identifier
//	:: operator-function-id
//	:: qualified-id
//	( expression )
//	id-expression

literal:
	  _NUM
	| _CHAR
	| _STR
	| _FLOAT
	| _BOOL
	;

primary_expression:
	  literal
		{ do1($$, Literal, $1); }
	| _KW_THIS
		{ do3($$, SpecialId, $1, $1, curscope); }
	| _LPAREN expression _RPAREN
		{ $$ = $2; }
	| id_expression_val
	;

// id-expression:
//	unqualified-id
//	qualified-id

id_expression:
	  unqualified_id
//x	| qualified_id
	;

id_expression_val:
	  unqualified_id_val
	;

// unqualified-id:
//	identifier
//	operator-function-id
//	conversion-function-id
//	~ class-name
//	template-id

unqualified_id:
	  identifier
	| operator_function_id
	| conversion_function_id
	| _DTOR_ID
		{ do3($$, SpecialId, T(DTOR_ID), $1, curscope); }
//x	| template_id
	;

unqualified_id_val:
	  _VAL_ID
	| _ID
		{ do2($$, Undeclared, $1, curscope); }
	| operator_function_id
	| conversion_function_id
	| _DTOR_ID
		{ do3($$, SpecialId, T(DTOR_ID), $1, curscope); }
	;

// qualified-id:
//	nested-name-specifier "template"(opt) unqualified-id

// nested-name-specifier:
//	class-or-namespace-name :: nested-name-specifier(opt)

// class-or-namespace-name:
//	class-name
//	namespace-name

// NOTE: this probably won't be necessary since CParse.cc handles scoping



// postfix-expression:
//	primary-expression
//	postfix-expression [ expression ]
//	postfix-expression ( expression-list(opt) )
//	simple-type-specifier ( expression-list(opt) )
//	postfix-expression . "template"(opt) ::(opt) id-expression
//	postfix-expression -> "template"(opt) ::(opt) id-expression
//	postfix-expression . pseudo-destructor-name
//	postfix-expression -> pseudo-destructor-name
//	postfix-expression ++
//	postfix-expression --
//	"dynamic_cast" < type-id > ( expression )
//	"static_cast" < type-id > ( expression )
//	"reinterpret_cast" < type-id > ( expression )
//	"const_cast" < type-d > ( expression )
//	"typeid" ( expression )
//	"typeid" ( type-id )

postfix_expression:
	  primary_expression
	| postfix_expression _LBRACK expression _RBRACK
		{ do2($$, Index, $1, $3); }
	| postfix_expression _LPAREN expression_list_opt _RPAREN
		{ do2($$, Call, $1, $3); }
	| simple_type_specifier_notypename _LPAREN expression_list_opt _RPAREN
		{ do2($$, CtorCall, $1, $3); }
	| type_name _LPAREN expression_list_opt _RPAREN
		{ do2($$, CtorCall, $1, $3); }
	| postfix_expression _OP_DOT id_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| postfix_expression _OP_PTRTO id_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| postfix_expression _OP_DOT pseudo_destructor_name
		{ do3($$, BinaryOp, $2, $1, $3); }
	| postfix_expression _OP_PTRTO pseudo_destructor_name
		{ do3($$, BinaryOp, $2, $1, $3); }
	| postfix_expression _OP_INC
		{ do2($$, UnaryOp, T(OP_INCP), $1); }
	| postfix_expression _OP_DEC
		{ do2($$, UnaryOp, T(OP_DECP), $1); }
	| newcast _OP_LSS type_id _OP_GTR _LPAREN expression _RPAREN
		{ do3($$, Cast, $6, $3, $1); }
	| _KW_TYPEID _LPAREN expression _RPAREN
		{ UNSUPPORTED /* 8 */ }
	| _KW_TYPEID _LPAREN type_id _RPAREN
		{ UNSUPPORTED /* 9 */ }
	;

newcast:
	  _KW_DCAST
	| _KW_SCAST
	| _KW_RCAST
	| _KW_CCAST
	;

// expression-list:
//	assignment-expression
//	expression-list , assignment-expression

expression_list:
	  assignment_expression
		{ do2($$, BuildExprList, none, $1); }
	| expression_list _COMMA assignment_expression
		{ do2($$, BuildExprList, $1, $3); }
	;

expression_list_opt: { do2($$, BuildExprList, none, none); } | expression_list;

// pseudo-destructor-name:
//	::(opt) nested-name-specifier(opt) type-name :: ~ type-name
//	::(opt) nested-name-specifier(opt) ~ type-name

pseudo_destructor_name:
	  _DTOR_ID
		{ do3($$, SpecialId, T(DTOR_ID), $1, curscope); }
/*
	  type_name _SCOPE _OP_NOT type_name
	| _OP_NOT type_name
*/
	;

// unary-expression:
//	postfix-expression
//	++ cast-expression
//	-- cast-expression
//	unary-operator cast-expression
//	"sizeof" unary-expression
//	"sizeof" ( type-id )
//	new-expression
//	delete-expression

unary_expression:
	  postfix_expression
	| _OP_INC cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	| _OP_DEC cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	| unary_operator cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	| _OP_SIZEOF unary_expression
		{ do2($$, UnaryOp, $1, $2); }
	| _OP_SIZEOF _LPAREN type_id _RPAREN
		{ do2($$, UnaryOp, T(OP_TSIZEOF), $3); }
	| new_expression
	| delete_expression
	;

// unary-operator: one of
//	* & + - ! ~

unary_operator:
	  _OP_MUL	{ $$ = T(OP_PTR); }
	| _OP_AND	{ $$ = T(OP_REF); }
	| _OP_ADD	{ $$ = T(OP_UPLUS); }
	| _OP_SUB	{ $$ = T(OP_UMINUS); }
	| _OP_LNOT
	| _OP_NOT
	;

// new-expression:
//	::(opt) "new" new-placement(opt) new-type-id new-initializer(opt)
//	::(opt) "new" new-placement(opt) ( type-id ) new-initializer(opt)

new_expression:
	  _OP_NEW new_placement_opt new_type_id new_initializer_opt
		{ do3($$, OpNew, $3, $4, $2); }
	| _OP_NEW new_placement_opt _LPAREN type_id _RPAREN new_initializer_opt
		{ do3($$, OpNew, $4, $6, $2); }
	;

// new-placement:
//	( expression-list )

new_placement:
	  _LPAREN expression_list _RPAREN
		{ $$ = $2; }
	;

new_placement_opt: { $$ = none; } | new_placement;

// new-type-id:
//	type-specifier-seq new-declarator(opt)

new_type_id:
	  type_specifier_seq new_declarator_opt
		{ eval_decl($$, $2, $1); }
	;

// new-declarator:
//	ptr-operator new-declarator(opt)
//	direct-new-declarator

new_declarator:
	  ptr_operator new_declarator_opt
		{ declPtr($$, $1, $2); }
	| direct_new_declarator
	;

new_declarator_opt: { $$ = none; } | new_declarator;

// direct-new-declarator:
//	[ expression ]
//	direct-new-declarator [ constant-expression ]

direct_new_declarator:
	  _LBRACK expression _RBRACK
		{ declArray($$, none, $2); }
	| direct_new_declarator _LBRACK constant_expression _RBRACK
		{ declArray($$, $1, $3); }
	;

// new-initializer:
//	( expression-list(opt) )

new_initializer:
	  _LPAREN expression_list_opt _RPAREN
		{ $$ = $2; }
	;

new_initializer_opt: { $$ = none; } | new_initializer;


// delete-expression:
//	::(opt) "delete" cast-expression
//	::(opt) "delete" [ ] cast-expression

delete_expression:
	  _OP_DELETE cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	| _OP_DELETEA cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	;

// cast-expression:
//	unary-expression
//	( type-id ) cast-expression

cast_expression:
	  unary_expression
	| _LPAREN type_id _RPAREN cast_expression
		{ do3($$, Cast, $4, $2, none); }
	;

// pm-expression:
//	cast-expression
//	pm-expression .* cast-expression
//	pm-expression ->* cast-expression

pm_expression:
	  cast_expression
	| pm_expression _OP_MDOT cast_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| pm_expression _OP_MPTRTO cast_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// multiplicative-expression:
//	pm-expression
//	multiplicative-expression * pm-expression
//	multiplicative-expression / pm-expression
//	multiplicative-expression % pm-expression

multiplicative_expression:
	  pm_expression
	| multiplicative_expression _OP_MUL pm_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| multiplicative_expression _OP_DIV pm_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| multiplicative_expression _OP_MOD pm_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// additive-expression:
//	multiplicative-expression
//	additive-expression + multiplicative-expression
//	additive-expression - multiplicative-expression

additive_expression:
	  multiplicative_expression
	| additive_expression _OP_ADD multiplicative_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| additive_expression _OP_SUB multiplicative_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;
 
// shift-expression:
//	additive-expression
//	shift-expression << additive-expression
//	shift-expression >> additive-expression

shift_expression:
	  additive_expression
	| shift_expression _OP_LSHIFT additive_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| shift_expression _OP_RSHIFT additive_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;
	
// relational-expression:
//	shift-expression
//	relational-expression < shift-expression
//	relational-expression > shift-expression
//	relational-expression <= shift-expression
//	relational-expression >= shift-expression

relational_expression:
	  shift_expression
	| relational_expression _OP_LSS shift_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| relational_expression _OP_GTR shift_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| relational_expression _OP_LEQ shift_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| relational_expression _OP_GEQ shift_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// equality-expression:
//	relational-expression
//	equality-expression == relational-expression
//	equality-expression != relational-expression

equality_expression:
	  relational_expression
	| equality_expression _OP_EQU relational_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| equality_expression _OP_NEQ relational_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// and-expression:
//	equality-expression
//	and-expression & equality-expression

and_expression:
	  equality_expression
	| and_expression _OP_AND equality_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// exclusive-or-expression:
//	and-expression
//	exclusive-or-expression ^ and-expression


exclusive_or_expression:
	  and_expression
	| exclusive_or_expression _OP_XOR and_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// inclusive-or-expression:
//	exclusive-or-expression
//	inclusive-or-expression | exclusive-or-expression

inclusive_or_expression:
	  exclusive_or_expression
	| inclusive_or_expression _OP_OR exclusive_or_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// logical-and-expression:
//	inclusive-or-expression
//	logical-and-expression && inclusive-or-expression

logical_and_expression:
	  inclusive_or_expression
	| logical_and_expression _OP_LAND inclusive_or_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// logical-or-expression:
//	logical-and-expression
//	logical-or-expression || logical-and-expression

logical_or_expression:
	  logical_and_expression
	| logical_or_expression _OP_LOR logical_and_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// conditional-expression:
//	logical-or-expression
//	logical-or-expression ? expression : assignment-expression

conditional_expression:
	  logical_or_expression
	| logical_or_expression _QMARK expression colon assignment_expression
		{ do3($$, CondExpr, $1, $3, $5); }
	;

// assignment-expression:
//	conditional-expression
//	logical-or-expression assignment-operator assignment-expression
//	throw-expression

assignment_expression:
	  conditional_expression
	| logical_or_expression assignment_operator assignment_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
//x	| throw_expression
	;

// assignment-operator: one of
//	= *= /= %= += -= >>= <<= &= ^= |=

assignment_operator:
	  _OP_ASN
	| _OP_ASNMUL
	| _OP_ASNDIV
	| _OP_ASNMOD
	| _OP_ASNADD
	| _OP_ASNSUB
	| _OP_ASNRSHIFT
	| _OP_ASNLSHIFT
	| _OP_ASNAND
	| _OP_ASNXOR
	| _OP_ASNOR
	;

// expression:
//	assignment-expression
//	expression , assignment-expression

expression:
	  assignment_expression
	| expression _COMMA assignment_expression
		{ do3($$, BinaryOp, T(OP_COMMA), $1, $3); }
	;

expression_opt: { $$ = none; } | expression;

// constant-expression:
//	conditional-expression

constant_expression:
	  conditional_expression
		{ do1($$, Literal, $1); }
	;


//////////////////////////////////////////////////////////////////////////////
// STATEMENTS (A.6)

// statement:
//	labeled-statement
//	expression-statement
//	compound-statement
//	selection-statement
//	iteration-statement
//	jump-statement
//	declaration-statement
//	try-block

// note: this got a little messy to avoid the dangling-else problem
// once shift/reduce precedence is added to shacc, this can be cleaned up

statement:
	  matched_statement
	| unmatched_statement
	;

other_statement:
	  labeled_statement
	| expression_statement
	| compound_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	| declaration_statement
//x	| try_block
	;

// selection-statement:
//	"if" ( condition ) statement
//	"if" ( condition ) statement "else" statement
//	"switch" ( condition ) statement

matched_statement:
	  other_statement
	| _KW_IF _LPAREN condition _RPAREN matched_statement _KW_ELSE matched_statement
		{ do3($$, IfElse, $3, $5, $7); }
	;

unmatched_statement:
	  _KW_IF _LPAREN condition _RPAREN statement
		{ do3($$, IfElse, $3, $5, none); }
	| _KW_IF _LPAREN condition _RPAREN matched_statement _KW_ELSE unmatched_statement
		{ do3($$, IfElse, $3, $5, $7); }
	;

selection_statement:
	  _KW_SWITCH _LPAREN condition _RPAREN statement
		{ do2($$, Switch, $3, $5); }
	;

// labeled-statement:
//	identifier : statement
//	"case" constant-expression : statement
//	"default" : statement

labeled_statement:
	  identifier colon statement
		{ NOSCOPE($1); do3($$, Label, none, $1, $3); }
	| _KW_CASE constant_expression colon statement
		{ do3($$, Label, $1, $2, $4); }
	| _KW_DEFAULT colon statement
		{ do3($$, Label, $1, none, $3); }
	;

// expression-statement:
//	expression(opt) ;

expression_statement:
	  expression_opt _EOS %uniq
	;

// compound-statement:
//	{ statement-seq(opt) }

compound_statement:
	  _LBRACE %uniq
		{ savescope_newlocal($$);
		  $1 = _stmtlist;
		  do3(_stmtlist, BuildStmtList, none, none, curscope);
		  _stmtlist.scope = curscope.scope; }
	    statement_seq_opt _RBRACE %uniq
		{ $$ = _stmtlist; _stmtlist = $1; 
		  restorescope($2); }
	;

// statement-seq:
//	statement
//	statement-seq statement

statement_seq_opt:	// optional
	| statement_seq_opt statement
		{ do3(_stmtlist, BuildStmtList, _stmtlist, $2, curscope); }
	;

// condition:
//	expression
//	type-specifier-seq declarator = assignment-expression

condition:
	  expression
	| type_specifier_seq declarator _OP_ASN assignment_expression
		{ 
		  defineSym($3, $2, $4); 
		  eval_decl($2, $3, $1);
		  do3($$, BuildStmtList, none, none, curscope);
		  $$.scope = curscope.scope; 
		  newdecl(&$$, $2);
		}
	;

condition_opt: { $$ = none; } | condition;

// iteration-statement:
//	"while" ( condition ) statement
//	"do" statement "while" ( expression ) ;
//	"for" ( for-init-statement condition(opt) ; expression(opt) ) statement

iteration_statement:
	  _KW_WHILE _LPAREN 
		{ savescope_newlocal($2); }
	     condition _RPAREN statement
		{ do3($$, While, $4, $6, none);
		  restorescope($2); }
	| _KW_DO statement _KW_WHILE _LPAREN expression _RPAREN _EOS %uniq
		{ do3($$, DoWhile, none, $2, $5); }
	| _KW_FOR _LPAREN
		{ savescope_newlocal($2); }
	     for_init_statement condition_opt _EOS expression_opt _RPAREN statement 
		{ do3($1, ForExpr, $4, $5, $7);
		  do2($$, For, $1, $9);
		  restorescope($2); }
	;

// for-init-statement:
//	expression-statement
//	simple-declaration

for_init_statement:
	  expression_statement
	| 
		{ $$ = _stmtlist;
	    	  do3(_stmtlist, BuildStmtList, none, none, curscope);
		  _stmtlist.scope = curscope.scope; } // $1
	   simple_declaration
		{ $$ = _stmtlist; _stmtlist = $1; }
	;

// jump-statement:
//	"break" ;
//	"continue" ;
//	"return" expression(opt) ;
//	"goto" identifier ;

jump_statement:
	  _KW_BREAK _EOS %uniq
		{ do1($$, Break, $1); }
	| _KW_CONTINUE _EOS %uniq
		{ do1($$, Continue, $1); }
	| _KW_RETURN expression_opt _EOS %uniq
		{ do1($$, Return, $2); }
	| _KW_GOTO identifier _EOS %uniq
		{ NOSCOPE($2); do2($$, Goto, none, $2); }
	;

// declaration-statement:
//	block-declaration

declaration_statement:
	  block_declaration
		{ $$ = none; }
	;

//////////////////////////////////////////////////////////////////////////////
// KEYWORDS (A.2)

// typedef-name:
//	identifier

typedef_name:
	  _TYPE_ID
	;

// namespace-name:
//	original-namespace-name
//	namespace-alias

namespace_name:
	  original_namespace_name
	| namespace_alias
	;

// original-namespace-name:
//	identifier

original_namespace_name:
	  _NSPACE_ID
	;

// namespace-alias:
//	identifier

namespace_alias:
	  _ASPACE_ID
	;

// class-name:
//	identifier
//	template-id

class_name:
	  _AGGR_ID
//x	| _CTOR_ID
//x		{ $$.token = T(AGGR_ID); }
//x	| template_id
	;

// enum-name:
//	identifier

enum_name:
	  _ENUM_ID
	;

// template-name:
//	identifier

template_name:
	  _TPLATE_ID
	;


identifier:
	  _ID
	| _VAL_ID
	| _TYPE_ID
	| _ENUM_ID
	| _AGGR_ID
//x	| _CTOR_ID
		{ $$.token = T(AGGR_ID); }
	| _NSPACE_ID
	| _ASPACE_ID
	| _TPLATE_ID
	;

identifier_opt: { $$ = none; } | identifier ;


//////////////////////////////////////////////////////////////////////////////
// save up expressions/compound statements for processing later

lbrace: _LBRACE | _LBRACE_SAVE { $$.token = T(LBRACE); } ;

colon: _COLON | _COLON_SAVE { $$.token = T(COLON); } ;

ignore_savecode: /* optional */ | _TOKENLIST;

save_compound_statement:
		{ $$.bval = _rawid; _rawid = true; }
	  _LBRACE_SAVE save_statement_seq_opt _RBRACE %uniq
		{ _rawid = $1.bval;
		  $2.token = T(LBRACE);
		  ((TokenList*)$3.ptr)->prepend($2); 
		  ((TokenList*)$3.ptr)->append($4);
		  $$ = $3; }
	;

save_statement_seq:
	  save_statement_seq_opt any_statement_token
		{ ((TokenList*)$$.ptr)->append($2); }
	| save_statement_seq_opt save_statement_list
		{ ((TokenList*)$$.ptr)->append_delete((TokenList*)$2.ptr); }
	;

save_statement_seq_opt:
		{ $$.token = T(TOKENLIST); $$.ptr = new TokenList; }
	| save_statement_seq
	;

save_statement_list:
	  lbrace save_statement_seq_opt _RBRACE %uniq
		{ ((TokenList*)$2.ptr)->prepend($1); 
		  ((TokenList*)$2.ptr)->append($3);
		  $$ = $2; }
	| _LPAREN save_statement_seq_opt _RPAREN
		{ ((TokenList*)$2.ptr)->prepend($1); 
		  ((TokenList*)$2.ptr)->append($3);
		  $$ = $2; }
	| _LBRACK save_statement_seq_opt _RBRACK
		{ ((TokenList*)$2.ptr)->prepend($1); 
		  ((TokenList*)$2.ptr)->append($3);
		  $$ = $2; }
	;

any_statement_token:
	  _NUM | _CHAR | _STR | _FLOAT | _BOOL
	| _EOS | colon | _QMARK 
	| simple_operator
	| _OP_DOT | _OP_MDOT
	| _OP_SCOPE | _ELIPSIS
	| cv_qualifier
	| simple_type_specifier_notypename
	| storage_class_specifier
	| _KW_IF | _KW_ELSE | _KW_DO | _KW_WHILE | _KW_FOR
	| _KW_SWITCH | _KW_CASE | _KW_DEFAULT
	| _KW_BREAK | _KW_CONTINUE | _KW_GOTO
	| _KW_RETURN
	| _KW_ENUM | _KW_TYPEDEF
	| class_key
	| function_specifier
	| _OP_SIZEOF | _OP_TYPEOF
	| _KW_ASM
	| _KW_OPERATOR | _KW_THIS
	| _KW_TYPEID | _KW_TYPENAME
	| _KW_NAMESPACE | _KW_USING
	| _KW_TEMPLATE
	| access_specifier
	| _KW_FRIEND
	| _KW_THROW | _KW_CATCH | _KW_TRY
	| newcast
	| identifier
		{ $$.token = T(ID); }
	| _DTOR_ID
	;


//////////////////////////////////////////////////////////////////////////////
// PREPROCESSING DIRECTIVES (A.11) [ the book says A.10.1 but
//	    this seems odd, since it's not a subsection of Exceptions


//////////////////////////////////////////////////////////////////////////////
// Attributes (GNU-ish extension)

attribute:
	  _KW_ATTRIBUTE _LPAREN _LPAREN save_statement_seq _RPAREN _RPAREN
		{ /* just ignore it for now */ }
	;

attribute_opt: { $$ = none; } | attribute ;

%%//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

