/*  $Id: CParse_test.cc,v 1.12 1998/04/22 12:02:04 shaggy Exp $
 *      
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: CParse_test.cc,v $
//  Revision 1.12  1998/04/22 12:02:04  shaggy
//  Changed ID to Id to be consistent with naming conventions
//
//  Revision 1.11  1998/01/11 05:06:33  shaggy
//  about to switch to standard ref/ptr semantics
//
//  Revision 1.10  1997/12/09 16:41:20  shaggy
//  param decls fixed
//
//  Revision 1.9  1997/12/09 15:23:05  shaggy
//  *** empty log message ***
//
//  Revision 1.8  1997/12/08 03:50:56  shaggy
//  *** empty log message ***
//
//  Revision 1.7  1996/11/29  18:51:25  shaggy
//  last checkin before removing unnecessary 'List' class
//
//  Revision 1.6  1996/11/23  17:22:39  shaggy
//  last checkin before switching to virtual functions
//
//  Revision 1.5  1996/10/19  21:47:53  shaggy
//  modified to handle new token location
//
//  Revision 1.4  1996/10/19  16:41:24  shaggy
//  supports abstract parameter declarations
//
//  Revision 1.3  1996/10/18  16:37:19  shaggy
//  basic type declarations (including functions) work
//
//  Revision 1.2  1996/10/18  00:53:16  shaggy
//  added ptrs, arrays, basic types
//
//  Revision 1.1  1996/10/06  20:33:57  shaggy
//  Initial revision
//
//      last modified: Fri May 24 00:51:49 1996
//      creation date: Mon Mar  4 00:35:59 1996


// #pragma Make Config_Needed += local
#pragma Make Includes_Banned += -nostdinc -nostdinc++
#pragma Make Flags_Banned += -static

#define DEBUG

#include <defines>
#include <sys/types.h>
#include <unistd.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <fcntl.h>

#include <Object>
#include <CLex>
#include <CParse>
#include <CAction>
#include <StrTab>
#include <SymTab>

#include <posix/assert.h>


/*
class UniqueObj {
	struct _ObjEntry {
	};
	
public:
	insert(
};
*/

// XXX - ideas:
//	- extend Item structure by making it a class that can
//	  be derived from and have methods added to
//	- extend Token by putting base definition in a file that can
//	  be included and then have more entries added to
//	- make sure specific/generic print routines are available for it 
//	  too

// XXX - at some point, any knowledge specific to certain data
// types should be removed from DECO, since it is supposed to be 
// a generic C++ pre-processor that generates code that can compile
// on any architecture. 


class CTest : public CAction {
private:
	const int _bufsize = 4096;
	char _buf[_bufsize];
	int _temp;
	int _input_fd;
	vStrTab::Id strid_class;
public:
	CTest(CLex &clp, vStrTab &strtab, vSymTab &symtab);
	bool open_input(const char *fname = 0);
	void close_input();
	void get_input(bool split_token);
	vStrTab::Id get_string_id(Token, const char *str);
	void abort_exit();
	void internal_error_exit();
	void veprintf(const char *fmt, va_list args);
	int new_temp() { return ++_temp; }
	int new_temp(Item &r)
		{ r.token = TEMP; return (r.ival = ++_temp); }

	void print_type(const type_t type);
	char* type_string(const type_t type, char *s);

// these are part of CAction, but are here for testing:
	type_t decl_type(const Item &item);

	void print_syms(const vSymTab &symtab, int indent = 0);
	void print_syms()
		{ print_syms(_global_scope); }

	action UnaryOp, BinaryOp;
	action DeclSym, DeclParam;
	action DeclPtr, DeclArray, DeclFunc;
	action Define, DefParam;
	action BuildTypeList, BuildParamList;
	action NewScope;
	action FindSym;

	action BuildExprList, BuildStmtList;
	action Initialize;
	action Cast, Call;
/*
	action Test;
	action DefineClass;
	action DefineStruct, DefineEnum, DefineUnion;
	action ScopeBegin, ScopeEnd, GetScope;
	action GetMember, GetThis, GetDestructor;
	action BuildExprList, BuildStmtList;
	action Label, CaseLabel, DefaultLabel, Switch;
	action Return, Goto, Jump;
	action IfElse, CondExpr;
	action Patch;
*/
	void GetSym(Item &r,
			const Item &type, const Item &sym, vSymTab *symtab);
};


static bool debugaction = false;

#ifdef DEBUG
#define PRINT_ACTION(r,p1,p2,p3) \
	PrintAction _p_a_ (this, __FUNCTION__, r, p1, p2, p3)

class PrintAction {
	CAction *_caction;
	CAction::Item &_rval;
public:
	PrintAction(CAction *caction, const char *name,
		CAction::Item &rval,
		const CAction::Item &p1,
		const CAction::Item &p2,
		const CAction::Item &p3)
			: _caction(caction), _rval(rval)
		{ if(debugaction)
			_caction->mprint_action("", name, p1, p2, p3);
		}
	~PrintAction()
		{ if(debugaction)
			_caction->mprint_result("", _rval);
		}
};

#else
#define PRINT_ACTION(r,p1,p2,p3) 	
#endif


struct Param {
	type_t		type;		// type of parameter
	vStrTab::Id	name;		// formal name of parameter
	CAction::Item	defval;		// default value
	Param		*next;		
};


struct ParamList {
	Param	*head;
	uint8	nparam;
	bool	variable;
};


struct Decl {
	enum StorageClass {
		sc_default = 0,
		sc_auto,
		sc_static,
		sc_extern,
		sc_register,
		sc_typedef
	};
	enum DataType {
		dt_default = 0,
		dt_int = 1,
		dt_char = 2,
		dt_wchar = 3,
		dt_bool = 4,
		dt_void = 5,
		dt_float = 6,
		dt_double = 7,
		dt_enum = 8,
		dt_struct = 9,
		dt_class = 10,
		dt_union = 11,
		dt_namespace = 12
		
	};

	bool		is_const:1;
	bool		is_volatile:1;
	bool		is_mutable:1;
	bool		is_unsigned:1;
	bool		is_signed:1;
	bool		is_short:1;
	uint		reserved1:9;
	bool		error:1;

	uint		longcount:3;
	uint		reserved2:5;

	DataType	type:4;
	StorageClass	storage:3;
	uint		reserved3:1;
	
	type_t		type_ptr;

	void clear() { *(uint64*)this = 0; }
};



#define case_ANY_SCOPE_ID  \
			case AGGR_ID: case ENUM_ID: \
			case NSPACE_ID: case ASPACE_ID: \
			case TPLATE_ID

#define case_ANY_DECL_ID  case TYPE_ID: case_ANY_SCOPE_ID 

#define case_ANY_ID	case ID: case_ANY_DECL_ID



CTest::CTest(CLex &clp, vStrTab &strtab, vSymTab &symtab)
	 : CAction(clp, strtab, symtab)
{
	_temp = 0;
	strid_class = strtab.id("class");
}



void
CTest::print_type(const type_t type)
{
	if(type->is_type(Type_Func)) {
		Func *func = (Func*)type->get_ptr();
		printf("Function(");
		for(uint i=0; i<func->get_nparam(); i++) {
			if(i)
				printf(", ");
			if(func->get_param_name(i))
				printf("%s: ",
				   _strtab.string(func->get_param_name(i)));
			print_type(func->get_param_type(i));
		}
		if(func->is_variable()) {
			if(func->get_nparam())
				printf(", ");
			printf("...");
		}
		printf(") returning ");
		print_type(func->get_rtype());
	} else if(type->is_type(Type_Array)) {
		Array *array = (Array*)type->get_ptr();
		if(0 == array->get_lower_bound()) 
			printf("Array[%d] of ", array->get_size());
		else
			printf("Array[%d..%d] of ", array->get_lower_bound(),
					array->get_upper_bound());
		print_type(array->get_rtype());
	} else if(type->is_type(Type_PtrTo)) {
		PtrTo *ptr = (PtrTo*)type->get_ptr();
		printf("Ptr to ");
		print_type(ptr->get_rtype());
	} else if(type->is_type(Type_RefTo)) {
		RefTo *ref = (RefTo*)type->get_ptr();
		printf("Ref to ");
		print_type(ref->get_rtype());
	} else if(type->is_type(Type_Integral)) {
		Integral *integral = (Integral*)type->get_ptr();
		printf((integral->is_unsigned() ? "uint%d" : "int%d"), 
			integral->get_bitsize());
	} else if(type == &Type_bool)
		printf("bool");
	else if(type == &Type_void)
		printf("void");
	else if(type->is_null()) 
		printf("<null>");
	else 
		printf("<unknown:%p/%#llx>",
			type->get_type(), type->get_data());
}



// XXX - this needs to be done with length-safe dynamically 
// extensible strings... 


char * strpcat(char *str, char *prefix)
{
	size_t n = strlen(prefix);
	memmove(str+n, str, strlen(str)+1);
	memcpy(str, prefix, n);
	return str;
}


char *
CTest::type_string(const type_t type, char *s)
{
	char buf[512];

	if(type->is_type(Type_Func)) {
		Func *func = (Func*)type->get_ptr();
		if(s[0] == '*' || s[0] == '&') {
			strpcat(s, "(");
			strcat(s, ")");
		}
		strcat(s, "(");
		for(uint i=0; i<func->get_nparam(); i++) {
			if(i)
				strcat(s, ", ");
			if(func->get_param_name(i))
				strcpy(buf,
				   _strtab.string(func->get_param_name(i)));
			else
				buf[0] = '\0';
			strcat(s, type_string(func->get_param_type(i), buf));
		}
		if(func->is_variable()) {
			if(func->get_nparam())
				strcat(s, ", ");
			strcat(s, "...");
		}
		strcat(s, ")");
		s = type_string(func->get_rtype(), s);
	} else if(type->is_type(Type_Array)) {
		Array *array = (Array*)type->get_ptr();
		if(s[0] == '*' || s[0] == '&') {
			strpcat(s, "(");
			strcat(s, ")");
		}
		if(0 == array->get_lower_bound()) 
			sprintf(buf, "[%d]", array->get_size());
		else
			sprintf(buf, "[%d..%d]", array->get_lower_bound(),
					array->get_upper_bound());
		strcat(s, buf);
		s = type_string(array->get_rtype(), s);
	} else if(type->is_type(Type_PtrTo)) {
		PtrTo *ptr = (PtrTo*)type->get_ptr();
		strpcat(s, "*");
		s = type_string(ptr->get_rtype(), s);
	} else if(type->is_type(Type_RefTo)) {
		RefTo *ref = (RefTo*)type->get_ptr();
		strpcat(s, "&");
		s = type_string(ref->get_rtype(), s);
	} else if(type->is_type(Type_Integral)) {
		Integral *integral = (Integral*)type->get_ptr();
		sprintf(buf, (integral->is_unsigned() ? "uint%d " : "int%d "), 
			integral->get_bitsize());
		strpcat(s, buf);
	} else if(type == &Type_bool)
		strpcat(s, "bool ");
	else if(type == &Type_void)
		strpcat(s, "void ");
	else if(type->is_null())
		strpcat(s, "<null> "); 
	else {
		sprintf(buf, "<unknown:%p/%#llx>",
			type->get_type(), type->get_data());
		strpcat(s, buf);
	}
	return s;
}



void CTest::print_syms(const vSymTab &symtab, int indent)
{
	vSymTab::Id id;
	char buf[512];

	for(int n=0; (id = symtab.index(n)); n++) {
		Decl *decl = (Decl*)symtab.get(id);
	
#if 0	
		printf("%*s%s: ", indent, "", _strtab.string(id));
		print_type(decl->type_ptr);
		printf("\n");
#endif

		printf("%*s", indent, "");
		switch(decl->storage) {
		case Decl::sc_auto:	printf("auto "); break;
		case Decl::sc_static:	printf("static "); break;
		case Decl::sc_extern:	printf("extern "); break;
		case Decl::sc_register:	printf("register "); break;
		case Decl::sc_typedef:	printf("typedef "); break;
		default: break;
		}
		if(decl->is_const)
			printf("const ");
		if(decl->is_volatile)
			printf("volatile ");
		if(decl->is_mutable)
			printf("mutable ");
		
		strcpy(buf, _strtab.string(id));	
		switch(decl->type) {
		case Decl::dt_class:
			printf("class");
			goto dumpscope;
		case Decl::dt_struct:
			printf("struct");
			goto dumpscope;
		case Decl::dt_union:
			printf("union");
			goto dumpscope;
		case Decl::dt_namespace:
			printf("namespace");
		dumpscope:
			printf(" {\n");
			// printf(" %s;\n", type_string(decl->type_ptr, buf));
			print_syms(*(vSymTab*)decl->type_ptr, indent+4); 
			printf("%*s} %s;\n", indent, "", buf);
			break;
		default:
			printf("%s;\n", type_string(decl->type_ptr, buf));
		}
		if(decl->type == Decl::dt_class
		  || decl->type == Decl::dt_struct 
		  || decl->type == Decl::dt_union
		  || decl->type == Decl::dt_namespace) {
		}
		
	}
}






type_t
CTest::decl_type(const Item &item)
{
	assert(item.token == TYPE_LIST);
	
	Decl &decl = *(Decl*)&item.val;
	if(decl.type_ptr)
		return decl.type_ptr;

	switch(decl.type) {
	case Decl::dt_void:
		if(decl.longcount || decl.is_short
			|| decl.is_signed || decl.is_unsigned)
			error("long/short/signed/unsigned invalid for void");
		return &Type_void;
	case Decl::dt_bool:
		if(decl.longcount || decl.is_short
			|| decl.is_signed || decl.is_unsigned)
			error("long/short/signed/unsigned invalid for bool");
		return &Type_bool;
	case Decl::dt_char:
		if(decl.longcount || decl.is_short)
			error("long/short invalid for char");
		if(decl.is_signed) 
			return &Type_schar;
		else if(decl.is_unsigned)
			return &Type_uchar;
		else
			return &Type_char;
	case Decl::dt_default:
	case Decl::dt_int:
		if(decl.is_short)
			return decl.is_unsigned
				? &Type_ushort : &Type_short;
		else switch(decl.longcount) {
			case 0: return decl.is_unsigned
				? &Type_uint : &Type_int; break;
			case 1:	return decl.is_unsigned
				? &Type_ulong : &Type_long; break;
			case 2:	return decl.is_unsigned
				? &Type_ulong2 : &Type_long2; break;
			default:
				error("long long long not supported");
				return 0;
		}
		break;
	default:
		error("unsupported type");
		return 0;
	}
}



void
CTest::UnaryOp(Item &r, const Item &op, const Item &arg, const Item &)
{
	new_temp(r);
	printf("t%d = %s %lld\n", r.ival,
		_clex.token_string((CLex::Token)op.token), arg.val);
}


void
CTest::BinaryOp(Item &r, const Item &op, const Item &arg1, const Item &arg2)
{
	PRINT_ACTION(r, op, arg1, arg2);
	if((arg1.token == NUM || arg1.token == CHAR)
			&& (arg2.token == NUM || arg2.token == CHAR)) {
		switch(op.token) {
		case OP_ADD:  r.val = arg1.val + arg2.val; break;
		case OP_SUB:  r.val = arg1.val - arg2.val; break;
		case OP_MUL:  r.val = arg1.val * arg2.val; break;
		case OP_DIV:  r.val = arg1.val / arg2.val; break;
		case OP_MOD:  r.val = arg1.val % arg2.val; break;
		case OP_AND:  r.val = arg1.val & arg2.val; break;
		case OP_OR:   r.val = arg1.val | arg2.val; break;
		case OP_XOR:  r.val = arg1.val ^ arg2.val; break;
		case OP_LAND: r.val = arg1.val && arg2.val; break;
		case OP_LOR:  r.val = arg1.val || arg2.val; break;
		case OP_LXOR: r.val=!(!arg1.val && !arg2.val); break;
		case OP_LSS:  r.val = arg1.val < arg2.val; break;
		case OP_GTR:  r.val = arg1.val > arg2.val; break;
		case OP_LEQ:  r.val = arg1.val <= arg2.val; break;
		case OP_GEQ:  r.val = arg1.val >= arg2.val; break;
		case OP_EQU:  r.val = arg1.val == arg2.val; break;
		case OP_NEQ:  r.val = arg1.val != arg2.val; break;
		case OP_LSHIFT:  r.val = arg1.val << arg2.val; break;
		case OP_RSHIFT:  r.val = arg1.val >> arg2.val; break;
		default:
			printf("unknown operator\n");
		}
		r.token = NUM;
	} else {
		new_temp(r);
		printf("t%d = %lld %s %lld\n", r.ival,
			arg1.val, _clex.token_string(
				(CLex::Token)op.token), arg2.val);
	}
}




void
CTest::BuildTypeList(Item &r,
		const Item &declist, const Item &item, const Item &)
{
	PRINT_ACTION(r, declist, item, NONE);
	switch(declist.token) {
	case TYPE_LIST:
		r = declist;
		break;
	case NONE:
		r.token = TYPE_LIST;	
		r.val = 0;
		break;
	default:
		assert(0);
		break;
	}

	Decl::StorageClass storage = Decl::sc_default;
	Decl &decl = (Decl&)r.val;

	// XXX - All of these should check for duplicates
	//	Also, all conflict checking needs to be done here

	switch(item.token) {
	case KW_AUTO:	storage = Decl::sc_auto; break;
	case KW_STATIC:	storage = Decl::sc_static; break;
	case KW_EXTERN:	storage = Decl::sc_extern; break;
	case KW_REGISTER:	storage = Decl::sc_register; break;
	case KW_TYPEDEF:	storage = Decl::sc_typedef; printf("got typedef\n"); break;
	case KW_MUTABLE:	decl.is_mutable = true; break;
	case KW_VOLATILE:	decl.is_volatile = true; break;
	case KW_CONST:	decl.is_const = true; break;
	case KW_LONG:	decl.longcount++; break;
	case KW_SHORT:	decl.is_short = true; break;
	case KW_SIGNED:	decl.is_signed = true; break;
	case KW_UNSIGNED:	decl.is_unsigned = true; break;
	case KW_INT:	decl.type = Decl::dt_int; break;
	case KW_CHAR:	decl.type = Decl::dt_char; break;
	case KW_BOOL:	decl.type = Decl::dt_bool; break;
	case KW_VOID:	decl.type = Decl::dt_void; break;
	case_ANY_DECL_ID:
		decl.type_ptr = ((Decl*)item.tuple.right)->type_ptr;
		break;
	case KW_CLASS:
	case KW_STRUCT:
	case KW_UNION:
	case KW_ENUM:
		decl.type_ptr = item.type; break;
	default:
		error("unsupported declaration");
	}
	if(storage) {
		if(!decl.storage)
			decl.storage = storage;
		else if(decl.storage == storage)
			error("repeated storage class");
		else
			error("multiple storage classes");
	}
}


void
CTest::BuildParamList(Item &r,
		const Item &paramlist, const Item &item, const Item &)
{
	PRINT_ACTION(r, paramlist, item, NONE);
	assert((paramlist.token == PARM_LIST)
		| (paramlist.token == NONE));

	r.val = paramlist.val;
	ParamList &plist = (ParamList&)r.val;
	r.token = PARM_LIST;

	Param *p;
	switch(item.token) {
	case PARM:
		p = (Param*)item.ptr;
		p->next = plist.head;
		plist.head = p;
		plist.nparam++;
		break;
	case ELIPSIS:
		plist.variable = true;
		break;
	case NONE:
		break;
	default:
		assertfail(item.token == PARM_LIST|ELIPSIS|NONE);
	}
}


void
CTest::DeclArray(Item &r, const Item &type, const Item &size,
			const Item &/*lower*/)
{
	PRINT_ACTION(r, type, size, NONE);
	r = type;
	if(size.token == NUM)
		((Decl&)r.val).type_ptr =
			new obj_Array(decl_type(type), size.val);
	else
		((Decl&)r.val).type_ptr = new obj_PtrTo(decl_type(type));
	// XXX - this is probably incorrect...
	// an unknown sized array should not necessarily be a pointer
}

void
CTest::DeclPtr(Item &r, const Item &ptrtype, const Item &type, const Item &)
{
	PRINT_ACTION(r, ptrtype, type, NONE);
	r = type;
	switch(ptrtype.token) {
	case OP_PTR:
		((Decl&)r.val).type_ptr = new obj_PtrTo(decl_type(type));
		break;
	case OP_REF:
		((Decl&)r.val).type_ptr = new obj_RefTo(decl_type(type));
		break;
	default:
		assertfail(ptrtype.token == OP_REF|OP_PTR);
	}
}


void
CTest::DeclFunc(Item &r,
		const Item &type, const Item &paramlist, const Item &)
{
	PRINT_ACTION(r, type, paramlist, NONE);
	assert(type.token == TYPE_LIST);
	assert(paramlist.token == PARM_LIST);

	ParamList &plist = (ParamList&)paramlist.val;

	FuncParam *fparam = new FuncParam;
	fparam->nparam = plist.nparam;
	fparam->firstdef = plist.nparam;
	fparam->is_variable = plist.variable;
	fparam->param = new ParamInfo[fparam->nparam];
	fparam->defval = 0;

	Param *param = plist.head;
	for(int i=fparam->nparam-1; i>=0; i--) {
		fparam->param[i].type = param->type;
		fparam->param[i].name = param->name;
		if(param->defval.token != NONE) 
			fparam->firstdef = i;
		param = param->next;
	}

	if((fparam->ndefval = fparam->nparam - fparam->firstdef)) {
		fparam->defval = new obj_t[fparam->ndefval];
		param = plist.head;
		for(int i=fparam->ndefval-1; i>=0; i--) {
			// fparam->defval[i] = param->defval;
			param = param->next;
		}
	}
	
	r = type;
	((Decl&)r.val).type_ptr = new obj_Func(decl_type(type), fparam);
}


void
CTest::DeclSym(Item &r, const Item &def, const Item &sym, const Item &scope)
{
	PRINT_ACTION(r, def, sym, scope);
	Decl *decl = new Decl;
	*decl = (Decl&)def.val;
	decl->type_ptr = decl_type(def);

	switch(sym.token) {
	case_ANY_ID:
		if(!get_symtab(sym, scope).add(sym.id, decl)) {
			error("%s already declared in this scope",
				_strtab.string(sym.id));
			delete decl;
			decl = 0;
		}
		r.token = ID;
		r.tuple.id = sym.id;
		r.tuple.right = decl;
		break;
	default:
		error("unknown assignment type");
		break;
	}
}


void
CTest::DeclParam(Item &r, const Item &type, const Item &name, const Item &)
{
	PRINT_ACTION(r, type, name, NONE);
	Param *p = new Param;
	p->type = decl_type(type);
	p->name = (name.token == NONE) ? 0 : name.id;
	p->defval = NONE;
	p->next = 0;
	r.token = PARM;
	r.ptr = p;
}	

void
CTest::Define(Item &r, const Item &p1, const Item &p2, const Item &p3)
{
	PRINT_ACTION(r, p1, p2, p3);
	r = p1;
}

void
CTest::DefParam(Item &r, const Item &par, const Item &val, const Item &)
{
	PRINT_ACTION(r, par, val, NONE);
	((Param*)par.ptr)->defval = val;
	r = par;
}

void
CTest::NewScope(Item &r,
	const Item &type, const Item &sym, const Item &scope)
{
	PRINT_ACTION(r, type, sym, scope);
	vSymTab &parent = get_symtab(sym, scope);
	SymTab *new_symtab = new SymTab(&parent);
	if(sym.token) {
		Decl *decl = new Decl;
		decl->clear();
		decl->storage = Decl::sc_typedef;
		switch(type.token) {
		case KW_CLASS:	decl->type = Decl::dt_class; break;
		case KW_STRUCT: decl->type = Decl::dt_struct; break;
		case KW_UNION: decl->type = Decl::dt_union; break;
		case KW_NAMESPACE: decl->type = Decl::dt_namespace; break;
		default:
			error("bad type");
		}
		decl->type_ptr = (type_t)new_symtab;
		if(!parent.add(sym.id, decl)) {
			error("%s already declared in this scope",
				_strtab.string(sym.id));
			// delete new_scope;
			// delete decl;
		}
	}
	r.token = type.token;
	r.scope = new_symtab;
}


void
CTest::GetSym(Item &r,
	const Item &type, const Item &sym, vSymTab *symtab)
{
	Decl *decl = (Decl*)symtab->get(sym.id);
	if(decl) {
		r.scope = sym.scope;
		r.tuple.id = sym.id;
		r.tuple.right = decl;
		if(decl->storage == Decl::sc_typedef) {
			switch(decl->type) {
			case Decl::dt_struct: 	
			case Decl::dt_union: 	
			case Decl::dt_class:
				r.token = AGGR_ID; break;
			case Decl::dt_enum:	
				r.token = ENUM_ID; break;
			case Decl::dt_namespace:
				r.token = NSPACE_ID; break;
			default:
				r.token = TYPE_ID;
			}
		}
		if(type.token == SCOPE) {
			switch(r.token) {
			case_ANY_SCOPE_ID:
				r.scope = (vSymTab*)decl->type_ptr;
				r.token = SCOPE;
				break;
			default:
				r.token = NONE;
			}
		}
	} else
		r = NONE;
}


void
CTest::FindSym(Item &r,
	const Item &type, const Item &sym, const Item &scope)
{
	PRINT_ACTION(r, type, sym, scope);
	vSymTab *symtab = &get_symtab(sym, scope);
	do {
		GetSym(r, type, sym, symtab);
		if(r.token != NONE)
			return;
		symtab = symtab->get_parent();
	} while(!sym.scope && symtab);
}


void
CTest::BuildExprList(Item &r, const Item &p1, const Item &p2, const Item &p3)
{
	PRINT_ACTION(r, p1, p2, p3);
}

void
CTest::BuildStmtList(Item &r, const Item &p1, const Item &p2, const Item &p3)
{
	PRINT_ACTION(r, p1, p2, p3);
}

#ifdef ndef
void
CTest::Initialize(Item &r, const Item &p1, const Item &p2, const Item &p3)
{
	PRINT_ACTION(r, p1, p2, p3);
}
#endif

void
CTest::Cast(Item &r, const Item &p1, const Item &p2, const Item &p3)
{
	PRINT_ACTION(r, p1, p2, p3);
}

void
CTest::Call(Item &r, const Item &p1, const Item &p2, const Item &p3)
{
	PRINT_ACTION(r, p1, p2, p3);
}


void
CTest::veprintf(const char *fmt, va_list args)
{
	vfprintf(stderr, fmt, args);
}



void
CTest::abort_exit()
{
	exit(1);
}


void
CTest::internal_error_exit()
{
	mprintf("", "Please send a full bug report to `shaggy@csh.rit.edu'");
	exit(2);
}



vStrTab::Id CTest::get_string_id(Token /*token*/, const char *str)
{
	return _strtab.id(str);
}


bool
CTest::open_input(const char *fname)
{
	if(!fname) 
		_input_fd = 0;
	else 
		_input_fd = open(fname, O_RDONLY);
	return _input_fd >= 0;
}


void
CTest::close_input()
{
	if(_input_fd)
		close(_input_fd);
}


void
CTest::get_input(bool split_token)
{
	if(_input_fd == 0 && !split_token)
		if(complete())
			write(1, "> ", 2);
		else
			write(1, "? ", 2);
	int n = read(_input_fd, _buf, _bufsize);
	if(n <= 0)
		_clex.set_eof();
	else
		_clex.set_src(_buf, n);
}
 

main(int argc, char **argv)
{
char strbuf[4096];
char bufbuf[4096];
char namebuf[1024];
	
	CLex lex(strbuf, sizeof(strbuf));
	lex.set_buf(bufbuf, sizeof(bufbuf));
	lex.set_name(namebuf, sizeof(namebuf));
	lex.parse_keywords().parse_space(false).parse_indent(false);
//	lex.parse_cpp();
//	lex.parse_eol();
//	lex.parse_indent();

	StrTab strtab;
	SymTab symtab;

	CTest test(lex, strtab, symtab);
	CParse cparse(lex, test);
	
	int argnum = 0;
	while(++argnum < argc) {
		if(!strcmp(argv[argnum], "-d")) 
			cparse.set_debug(true);
		else if(!strcmp(argv[argnum], "-da"))
			debugaction = true;
		else if(!strcmp(argv[argnum], "-dy"))
			cparse.set_yydebug(true);
		else
			break;
	}
	do {
		if(!test.open_input((argnum < argc) ? argv[argnum] : 0))
			perror((argnum < argc) ? argv[argnum] : "stdin" );
		if(!cparse.parse()) 
			fprintf(stderr, "cannot recover from previous errors\n");
		test.close_input();
		printf("\nSymbol Table:\n");
		test.print_syms();
	} while(++argnum < argc);
}

