/*  $Id: CAction.hh,v 1.20 1998/05/12 15:53:20 shaggy Exp $
 *      Default C/C++ Parsing Actions
 *      
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996,1997 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: CAction.hh,v $
//  Revision 1.20  1998/05/12 15:53:20  shaggy
//  scoping and class tags almost work
//
//  Revision 1.19  1998/05/12 05:00:12  shaggy
//  more scope work
//
//  Revision 1.18  1998/05/10 18:53:39  shaggy
//  just about ready to start fixing scoping
//
//  Revision 1.17  1998/05/10 00:03:30  shaggy
//  string concatenation works, and comments are being
//  added to get ready to fix scoping
//
//  Revision 1.16  1998/05/04 03:09:18  shaggy
//  support for ctors, dtors, conversion operators, operator overloading, etc
//
//  Revision 1.15  1998/05/03 02:47:07  shaggy
//  added inheritence to parser
//
//  Revision 1.14  1998/04/26 14:08:56  shaggy
//  cleanup
//
//  Revision 1.13  1998/04/22 13:06:38  shaggy
//  better enum handling
//
//  Revision 1.12  1998/04/22 12:02:03  shaggy
//  Changed ID to Id to be consistent with naming conventions
//
//  Revision 1.11  1998/04/16 13:35:43  shaggy
//  access control
//
//  Revision 1.10  1998/04/15 13:39:43  shaggy
//  added types to constants
//
//  Revision 1.9  1998/04/12 03:19:59  shaggy
//  improvments for function output
//
//  Revision 1.8  1998/04/06 14:28:06  shaggy
//  got enums almost working
//
//  Revision 1.7  1998/02/08 19:07:05  shaggy
//  about to start splitting up everything
//
//  Revision 1.6  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.5  1998/01/11 05:08:24  shaggy
//   about to switch to standard ref/ptr semantics
//
//  Revision 1.4  1997/12/09 16:43:19  shaggy
//  param parsing fixed
//
//  Revision 1.3  1996/10/19  21:46:46  shaggy
//  moved non-lex/parse tokens here
//
//  Revision 1.2  1996/10/19  16:42:38  shaggy
//  about to move around token declarations
//
//  Revision 1.1  1996/04/10  12:27:55  shaggy
//  Initial revision
//


#include <defines>
#include <types.h>
#include <inttypes.h>
#include <stdarg.h>
#include <CLex>
#include <vStrTab>
#include <vSymTab>




#ifdef NDEBUG
#	define assertfail(c) internal_error(__FILE__, __LINE__)
#else
#	define assertfail(c) _assertfail(#c, __FILE__, __LINE__)
#endif


class ErrorReport {
public:
	virtual void error(csz *fmt, ...) = 0;
	virtual void warning(csz *fmt, ...) = 0;
	virtual void abort(csz *fmt, ...) = 0;
	virtual void internal_error(csz *file, int line,
				csz *module = 0) = 0;
};


#define CACTION0(fn)	virtual void fn(Item *r, Item_, Item_, Item_);
#define CACTION1(fn,p1)	virtual void fn(Item *r, Item_ p1, Item_, Item_);
#define CACTION2(fn,p1,p2)	\
	virtual void fn(Item *r, Item_ p1, Item_ p2, Item_);
#define CACTION3(fn,p1,p2,p3)	\
	virtual void fn(Item *r, Item_ p1, Item_ p2, Item_ p3);


class CAction : public ErrorReport {
private:
	bool _complete;
protected:
	CLex &_clex;
	vStrTab &_strtab;
	vSymTab &_global_scope;
	int _num_errors;
	int _num_warnings;
public:
	enum Token {
		EOF_COMPLETE = 0,
		EOF_INCOMPLETE,
		NONE, INCOMPLETE, INVALID,
	
#include "CParse.enum.h"

		TEMP,		// used to designate temporaries
		RESULT,		// result of specified code
		INSN,		// pointer to another instruction
		LIST,		// linked list
		TREE,		// binary (or more) tree
		ARRAY,		// linear array
		OBJ,		// generic object
		TYPE,		// generic type
		PARM,		// parameter definition
		PARM_LIST,	// parameter list
		TYPE_LIST,	// declaration list
		EXPR_LIST,	// expression list
		STMT_LIST,	// statement list
		ERROR,		// error information
		ERROR_LIST,	// list of error information
		STRUCT,		// structure information
		CLASS,		// class information
		VAR,		// variable
		FUNC,		// function

		NUM_TOKENS
	};

	struct Item {
		Token token;
		union {
			vSymTab *scope;
			struct {
				bool is_signed;
				char is_long;
			} c;	// for constants/literals only
			void *cptr;	// for constants/literals only
		};
		union {
			uint64 val;
			int64 sval;
			double fval;
			int ival;
			uint uival;
			char cval;
			bool bval;
			Token tval;
			csz *str;
			void *ptr;
			vStrTab::Id id;
			class Object *objptr;	// obj_t*
			class Object *type;	// type_t
			struct {
				union {
					void *prev;
					void *left;
					void *ptr;
					vStrTab::Id id;
					int32 ival;
					uint32 uival;
					int32 lval;
					uint32 ulval;
					Token tval;
				};
				union {
					void *next;
					void *right;
					int32 rval;
					uint32 urval;
				};
			} tuple;
		};

		Item() {}
		Item(Token t) { token = t; val = 0; }
		Item(CLex::Token t) { token = (Token)t; val = 0; }
		Item(Token t, uint64 v) { token = t; val = v; }
		Item(CLex::Token t, uint64 v) { token = (Token)t; val = v; }
		Item(Token t, double f) { token = t; fval = f; }
		Item(CLex::Token t, double f) { token = (Token)t; fval = f; }
		Item(Token t, csz *s) { token = t; str = s; }
		Item(CLex::Token t, csz *s) { token = (Token)t; str = s; }
		Item(Token t, void *p) { token = t; ptr = p; }
		Item(CLex::Token t, void *p) { token = (Token)t; ptr = p; }
		Item(csz *s) { token = CSTR; str = s; }
		Item(int n) { token = NUM; val = n; }
		Item& operator=(Token t) { token = t; val = 0; return *this; }
		Item& operator=(CLex::Token t)
			{ token = (Token)t; val = 0; return *this; }
	};
	typedef const Item &Item_;

	static const Item null_item;

	
	bool complete() const { return _complete; }
	void complete(bool tf) { _complete = tf; }

	int errors() const { return _num_errors; }
	int warnings() const { return _num_warnings; }

	virtual void accept();
	virtual void get_input(bool split_token) = 0;
	virtual vStrTab::Id string_id(Token, csz *str) = 0;
	CAction(CLex &clp, vStrTab &strtab, vSymTab &globlsym);
	virtual void eol();
	virtual csz* string(Token token) const;
	virtual csz* string(vStrTab::Id id) const;
	virtual sz * string(Item_ item, sz *buf, size_t buflen,
				bool show_type = false) const;

	virtual void eprintf(csz *fmt, ...);
	virtual void veprintf(csz *fmt, va_list args) = 0;
	virtual void mprintf(csz *prefix, csz *fmt, ...);
	virtual void vmprintf(csz *prefix,
				csz *fmt, va_list args);
	virtual void _assertfail(csz *cond, csz *filename,
				int line);

	virtual void error(csz *fmt, ...);
	virtual void warning(csz *fmt, ...);
	virtual void abort(csz *fmt, ...);
	virtual void internal_error(csz *file, int line,
				csz *module = 0);

	virtual void abort_exit() = 0;
	virtual void internal_error_exit() = 0;
	
	virtual void mprint_action(csz *prefix, csz *name,
			Item_ p1, Item_ p2, Item_ p3);
	virtual void mprint_result(csz *prefix, Item_ r);
	virtual void unsupported(csz *name, Item_ p1, Item_ p2, Item_ p3);

	virtual vSymTab* get_symtab(Item_ id, Item_ scope);
	
	typedef void action(Item *r, Item_ p1, Item_ p2, Item_ p3);
	typedef action CAction::* Action;

	CACTION3(Test, p1, p2, p3);		/* r = p1 */
	// used for testing parts of the parser during development
	
	CACTION2(Undeclared, sym, scope);	/* r = p1 */
	//	sym: NewId
	//	scope: AnyScope
	// report an error about an undeclared symbol

	CACTION1(Literal, item);		/* r = literal(item) */
	//	item: RawLit|Lit|DeclId|Expr
	//	r: Lit
	// create a descriptor for a literal
	// should return NONE if no valid literal can be generated

	CACTION2(UnaryOp, op, arg); 		/* r = op(arg) */
	//	op: OP_INC|OP_DEC|OP_INCP|OP_DECP|OP_SIZEOF
	//		|OP_MUL|OP_AND|OP_ADD|OP_SUB|OP_LNOT|OP_NOT
	//	arg: Lit|Id|Expr
	//	r: Lit|Expr
	// perform a unary operation
	// should handle NONE without generating additional errors
	
	CACTION3(BinaryOp, op, arg1, arg2);	/* r = op(arg1, arg2) */
	//	op: OP_INDEX|OP_DOT|OP_PTRTO|OP_MDOT|OP_MPTRTO
	//		|OP_MUL|OP_DIV|OP_MOD|OP_ADD|OP_SUB
	//		|OP_LSHIFT|OP_RSHIFT
	//		|OP_LSS|OP_GTR|OP_LEQ|OP_GEQ|OP_EQU|OP_NEQ
	//		|OP_AND|OP_XOR|OP_OR
	//		|OP_LAND|OP_LOR
	//		|OP_COMMA
	//		|OP_ASN|OP_ASNMUL|OP_ASNDIV|OP_ASNMOD
	//		|OP_ASNADD|OP_ASNSUB
	//		|OP_ASNRSHIFT|OP_ASNLSHIFT
	//		|OP_ASNAND|OP_ASNXOR|OP_ASNOR
	//	arg1: Lit|Id|Expr
	//	arg2: Lit|Id|Expr
	// 	r: Lit|Expr
	// perform a binary operation
	// should handle NONE without generating additional errors

	CACTION2(Index, array, index);		/* r = array[index] */
	// 	array: Lit|Id|Expr
	// 	index: Lit|Id|Expr
	// note that if index is a pointer, r = index[array] instead,
	// for backwards compatibility with C
	// should handle NONE without generating additional errors

	CACTION2(DeclParam, type, sym);
	//	type:
	//	sym:
	//	r:

	CACTION3(DeclSym, type, sym, scope);
	//	type: TypeList | EnumType | NONE
	//	sym: Id
	//	scope: AnyScope
	//	r: DeclId
	// declare a symbol in the given scope to be of type `type'

	CACTION2(DeclPtr, ptrtype, type);
	//	ptrtype: OP_PTR | OP_REF 
	//	type: Type
	//	r: Type
	// return a type that is a pointer or reference to type `type'

	CACTION3(DeclArray, type, size, lower);
	//	type: Type
	//	size: Lit|Id|Expr|NONE
	//	lower: Lit|Id|Expr|NONE
	//	r: Type
	// return a type that is an array from [lower..lower+size-1] of `type'
	// a lower bound of NONE implies 0, while a size of NONE
	// implies an array with an unknown/variable size

	CACTION3(DeclFunc, type, plist, scope);
	//	type: Type
	//	plist: ParamList
	//	scope: AnyScope
	//	r: Type
	// return a function type for the specified param list and return value

	CACTION2(Define, var, val);
	//	var: Id
	//	val: Lit|Id|Expr
	//	r: Id
	// define the default value for a variable, or the
	// code body for a function
	// should handle NONE without generating additional errors

	CACTION2(DefParam, par, val);
	//	par: 
	//	val:
	//	r:
	
	CACTION3(AggrRef, type, sym, scope);
	//	type:
	//	sym:
	//	scope:
	//	r:
		
	CACTION3(NewAggr, type, sym, scope);
	//	type: KW_STRUCT | KW_CLASS | KW_UNION | KW_DCLASS | KW_DTYPE
	//	sym: Id | NONE
	//	scope: AnyScope (enclosing scope)
	//	r: AggrId
	// note: if undefined, defaults to calling NewScope,
	// and NewScope is expected to be able to handle NewAggr as well

	CACTION3(NewNspace, type, sym, scope);
	//	type: KW_NAMESPACE
	//	sym: Id | NONE
	//	scope: AnyScope (enclosing scope)
	//	r: NspaceId
	// note: if undefined, defaults to calling NewScope
	// and NewScope is expected to be able to handle NewNspace as well

	CACTION3(NewScope, type, sym, scope);
	//	type: KW_STRUCT | KW_CLASS | KW_UNION | KW_NAMESPACE
	//	sym: Id | NONE
	//	scope: AnyScope (enclosing scope)
	//	r: Scope
	// create a new scope, and enter a handle to it in
	// the enclosing scope, with the specified name, or anonymous

	CACTION1(FreeScope, scope);
	//	scope: Scope
	//	r: Scope (parent of freed scope)
	// free the specified scope to reclaim its memory
	// all symbols created within it can be freed as well

	CACTION3(EnumRef, type, sym, scope);
	//	type:
	//	sym:
	//	scope:
	//	r:
	
	CACTION3(NewEnum, type, sym, scope);
	//	type: KW_ENUM
	//	sym: Id
	//	scope: Scope
	//	r: EnumType

	CACTION2(SetAccess, scope, acc);
	//	scope:
	//	acc:
	//	r:

	CACTION3(Inherit, scope, itype, base);
	//	scope:
	//	itype:
	//	base:
	//	r:

	CACTION2(BuildTypeList, typelist, item);
	//	typelist: TypeList|NONE
	//	item: RawTypeKw|TypeId|TypeList
	//	r: TypeList
	// builds a compound type list from simple type keywords
	// and predefined types
	// note that joining two typelists must be supported

	CACTION2(BuildParamList, plist, item);
	//	plist: ParamList|NONE
	//	item: Id|ELIPSIS|NONE
	//	r: ParamList
	// builds a list of parameters used in formal function definitions
	// if item==ELIPSIS the paramlist should be marked as having
	// a variable number of parameters

	CACTION2(BuildExprList, exprlist, item);
	//	exprlist: ExprList | NONE
	//	item: Lit|Id|Expr|NONE
	//	r: ExprList
	// builds a list of expressions to be used for a function call
	// should handle NONE without generating additional errors

	CACTION3(FindSym, type, sym, scope);
	//	type: NONE|SCOPE|VAL_ID|AGGR_ID
	//		|KW_CLASS|KW_STRUCT|KW_UNION|KW_ENUM
	//	sym: Id|NONE
	//	scope: ANyScope|NONE
	//	r: Id
	// find a symbol, starting in the specified scope
	// if sym.scope is set, only that scope will be used
	// if sym==NONE && type==SCOPE, the global scope should be returned

	CACTION2(Cast, type, val);
	//	type:
	//	val:
	//	r:

	CACTION2(Call, fn, param);
	//	fn:
	//	param:
	//	r:

	CACTION3(SpecialId, type, val, scope);
	//	type:
	//	val:
	//	scope:
	//	r:

	virtual action GetMember, GetThis, GetDestructor;
	virtual action BuildStmtList;
	virtual action Label, CaseLabel, DefaultLabel, Switch;
	virtual action Return, Goto, Jump;
	virtual action IfElse, CondExpr;
	virtual action Patch;
};



//    implementation-defined types:
//
//	(Tokens listed in square brackets can be optionally
//	replaced with any token-- the one listed is just
//	a standard preference, but it doesn't affect the parser---
//	if other tokens are listed as well, then those ones have special
//	meanings, and should not be reused)
//
//	AnyId:		any identifier (might not be declared)
//		= DeclId | NewId
//	AnyDeclId:	any declared indentifier
//		= AnyTypeId | AnyValId | NspaceId | TplateId
//	AnyTypeId:	identifier for any type descriptor
//		= AggrId | EnumId | TypeId
//	AnyScope:	identifier for any scope
//		= AggrId | NspaceId | CodeScope
//	AnyValId:	identifier for any symbol with a value
//		token= ID | DTOR_ID | CONV_ID | OPER_ID
//	AggrId:		identifier for an aggregate (class/struct/union)
//		token= AGGR_ID
//	EnumId:		identifier for an enumeration
//		token= ENUM_ID
//	TypeId:		identifier for a type (not enum/aggr)
//		token= TYPE_ID
//	ValId:		identifier for a symbol with a value
//		token= VAL_ID
//	NewId:		undefined identifier
//		token= ID
//	NspaceId:
//		token= NSPACE_ID | ASPACE_ID
//	CodeScope:
//		token= SCOPE
//	TypeList:	compound list of identifiers to describe a type
//		token= [TYPE_LIST]
//	Lit:		literal (and synthesized literals, which
//				are essentially constants)
//		token= [LITERAL] | NONE
//	Expr:		compound expression
//
//	
//   CParse-defined types:
//
//	RawId:		
//		token= ID
//		id= string id
//
//	RawLit:
//	    integer number:
//		token= NUM 
//		val= 64-bit number
//		c.is_signed= literal has U suffix
//		c.is_long= # of L suffixes
//
//	    character:
//		token= CHAR
//		val= character number
//
//	    string:
//		token= STR
//		id= string id
//
//	    floating-point number:
//		token= FLOAT
//		fval= number (as a double)
//
//	    boolean:
//		token= BOOL
//		val= 0:false, 1:true
//
//	RawTypeKw:
//	    storage class:
//		token= KW_AUTO | KW_REGISTER | KW_STATIC | KW_EXTERN
//			| KW_MUTABLE
//
//	    simple type:
//		token= KW_CHAR | KW_WCHAR | KW_BOOL | KW_INT
//			| KW_FLOAT | KW_DOUBLE | KW_VOID
//
//	    type modifier:
//		token= KW_SHORT | KW_LONG | KW_SIGNED | KW_UNSIGNED
//
//	    function specifier:
//		token= KW_INLINE | KW_VIRTUAL | KW_EXPLICIT
//
//	    special declarations:
//		token= KW_FRIEND | KW_TYPEDEF
//
//	    cv qualifier:
//		token= KW_CONST | KW_VOLATILE	
//
