/*  Lexer.cc                                                    (CRC:65305734)
 *      relative path: ./lib/parse
 *      creation date: Mon Mar  4 00:48:00 1996
 *      last modified: Sat Jun 21 08:24:39 1997
 *      header update: Sat Jun 21 08:24:40 1997
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        Generic parsing functions
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define IMPLEMENTATION_Lexer
#include <Lexer>
#include <new>

Lexer::Lexer()
{
	if(init_ctypes) {
		for(int i=0; i<256; i++) {
			int ct = 0;
			switch(i) {
			case ' ': case '\r': case '\n': case '\t':
			case '\f': case '\v':
				ct |= WHITESPACE; break;
			case 'A' ... 'F': ct |= HEXNUM;
			case 'G' ... 'Z': ct |= UPPER | ID | ID_START; break;
			case 'a' ... 'f': ct |= HEXNUM;
			case 'g' ... 'z': ct |= LOWER | ID | ID_START; break;
			case '0' ... '1': ct |= BINNUM;
			case '2' ... '7': ct |= OCTNUM;
			case '8' ... '9': ct |= NUM | HEXNUM | ID; break;
			case '$': case '_':	ct |= ID | ID_START; break;
			default:
				if(i >= 32 && i <= 126)
					ct |= PUNC;
			}
			ctype[i] = (ctypes)ct;
		}
	}	
	set_empty();
	_sol = true;
	_line_num = 1;
}

Lexer::Lexer(const char *str, size_t len)
{
	new(this)Lexer();
	set_src(str, len);
}

