/*  $Id: atoi.cc,v 1.2 1997/06/21 12:23:56 shaggy Exp $
 *      convert an ascii string to a number
 *      relative path: ./lib/posix
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        Posix-compliant code
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: atoi.cc,v $
//  Revision 1.2  1997/06/21 12:23:56  shaggy
//  header update
//
//  Revision 1.1  1997/01/21 05:04:01  shaggy
//  Initial revision
//
//      creation date: 1992


#include <types.h>
#include <posix/stddef.h>
#include <posix/stdlib.h>

#define DEF_NUMBASE 10
#define digit_value(ch)  (ch<='9'?ch-'0':10+(ch>='A'&&ch<='Z'?ch-'A':ch-'a'))


/* numbers are accepted in the following format:
 *   ddddd   default base
 *   0dddd   octal
 *   0xddd   hex
 *   0tddd   decimal
 *   0nddd   binary
 */

ulong strtoul(const char *str, char **adr, int base)
{
int x;
ulong num = 0;
	if(!base)
	base = DEF_NUMBASE;
	while(*str == ' ' || *str == '\f' || *str == '\n'
			|| *str == '\r' || *str == '\t' || *str == '\v')
		str++;
	if(*str == '0') {
		base = 8;
		str++;
		switch(*str) {
		case 'x':
			base = 16; str++; break;
		case 't':	
			base = 10; str++; break;
		case 'n':
			base = 2; str++; break;
		}
	}
	while((x = digit_value(*str)) >= 0 && (x < base)) {
		num = (num*base)+x;
		str++;
	}
	if(adr)
		*adr = (char*)str;
	return(num);
}


long strtol(const char *str, char **adr, int base)
{
	if(*str == '-')
		return(-(long)strtoul(&str[1], adr, base));
	else
		return((long)strtoul(str, adr, base));
}


int atoi(char *str)
{
	return((long)strtol(str,0,10));
}

long atol(char *str)
{
	return((long)strtol(str,0,10));
}

ulong atoul(char *str)
{
	return((ulong)strtoul(str,0,10));
}


uchar bcd_decode(uchar byte)
{
	return(((byte>>4)&0xf)*10 + (byte&0xf));
}

uchar bcd_encode(uchar byte)
{
	return(((byte/10)<<4) | (byte%10));
}

