/*  $Id: StrTab.cc,v 1.7 1998/04/22 12:02:04 shaggy Exp $
 *      Symbol Table support
 *      
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: StrTab.cc,v $
//  Revision 1.7  1998/04/22 12:02:04  shaggy
//  Changed ID to Id to be consistent with naming conventions
//
//  Revision 1.6  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.5  1998/01/11 05:06:33  shaggy
//  about to switch to standard ref/ptr semantics
//
//  Revision 1.4  1996/11/08  22:23:11  shaggy
//  header update
//
//  Revision 1.3  1996/10/26  19:01:48  shaggy
//  header update
//
//  Revision 1.2  1996/10/26  18:52:19  shaggy
//  header update
//
//  Revision 1.1  1996/10/18  01:41:15  shaggy
//  Initial revision
//
 

#define IMPLEMENTATION_StrTab
#include <StrTab>
#include <posix/string.h>


StrTab::StrTab()
{
	_list = 0;
}


StrTab::~StrTab()
{
	_StrEntry *sp = _list;
	while(sp) {
		_StrEntry *spnext = sp->next;
		delete sp->string;
		delete sp;
		sp = spnext;
	}
}


StrTab::Id StrTab::id(csz *string) 
{
	// Search through list
	_StrEntry *sp = _list;
	while(sp) {
		if(!strcmp(sp->string, string)) 
			return (Id)sp;
		sp = sp->next;
	}

	// If not found, add to head of list
	sp = new _StrEntry;
	sp->string = new char [strlen(string)+1];
	strcpy((sz*)sp->string, string);
	sp->refcount = 1;
	sp->next = _list;
	_list = sp;
	return (Id)sp;
}


uint StrTab::unref(Id id)
{
	_StrEntry *sp = (_StrEntry*)id;
	uint newcount = --(sp->refcount);
	if(!newcount) {
		// Find the pointer to this entry
		_StrEntry **spp = &_list;
		while(*spp != sp)
			spp = &(*spp)->next;

		// And skip over to the next one
		*spp = sp->next;

		// Remove this entry
		delete sp->string;
		delete sp;
	}
	return newcount;
}

