/*  $Id: SymTab.hh,v 1.7 1998/07/06 22:49:42 shaggy Exp $
 *      Symbol Table support
 *      
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: SymTab.hh,v $
//  Revision 1.7  1998/07/06 22:49:42  shaggy
//  reworked symtab, output ordering, etc
//
//  Revision 1.6  1998/07/03 19:06:24  shaggy
//  changed scope semantics; fn name is available in scope now
//
//  Revision 1.5  1998/05/04 03:09:19  shaggy
//  support for ctors, dtors, conversion operators, operator overloading, etc
//
//  Revision 1.4  1998/04/22 12:02:05  shaggy
//  Changed ID to Id to be consistent with naming conventions
//
//  Revision 1.3  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.2  1998/01/11 05:11:03  shaggy
//   about to switch to standard ref/ptr semantics
//
//  Revision 1.1  1996/04/08  00:07:21  shaggy
//  Initial revision
//


#include <defines>
#include <types.h>
#include <vStrTab>
#include <vSymTab>

// just a linked-list for now, although this needs to be changed
// in the future for efficiency

class SymTab : public vSymTab {
private:
	vSymTab *_parent;
	vStrTab *_strtab;
	void *_data;
	struct _SymEntry {
		Id	id;
		Val	val;
		_SymEntry *next;	// next sequential entry
		_SymEntry *cnext;	// next cached entry
	};
	_SymEntry *_first;
	mutable _SymEntry *_cache;
	mutable _SymEntry *_inext;	// next index entry
	mutable uint	   _inum;	// next index number
public:
	SymTab(vStrTab *strtab);
	SymTab(vSymTab *parent);
	SymTab(vSymTab *parent, void *data);
	~SymTab();
	bool get(Id id, Val &val) const;
	bool set(Id id, Val val);
	bool add(Id id, Val val);
	bool del(Id id);
	Id index(uint n) const;
	vSymTab* parent() const;
	void* data() const;
	bool data(void*);
	vStrTab* strtab() const;
};


/*
class obj_SymTab : public Object {
public:
	obj_SymTab() : Object(Type_vSymTab, Alloc)
		{ _ref.ptr = new SymTab(); }
	obj_SymTab(vSymTab *parent) : Object(Type_vSymTab, Alloc)
		{ _ref.ptr = new SymTab(parent); }
	obj_SymTab(SymTab &data) : Object(Type_vSymTab, Auto)
		{ _ref.ptr = &data; }
	obj_SymTab(SymTab *data) : Object(Type_vSymTab, Auto)
		{ _ref.ptr = data; }
};
*/

