/*  $Id: CObj.hh,v 1.8 1998/07/12 19:01:23 shaggy Exp $
 *      
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: CObj.hh,v $
//  Revision 1.8  1998/07/12 19:01:23  shaggy
//  many more cvqual fixes
//
//  Revision 1.7  1998/07/05 02:47:37  shaggy
//  tons of bug fixes
//
//  Revision 1.6  1998/07/03 16:53:51  shaggy
//  new Scope wrapper class, and dcast_XXX functions eliminate lots of casts
//
//  Revision 1.5  1998/06/30 15:38:37  shaggy
//  better support for member fns and member fn ptrs
//
//  Revision 1.4  1998/05/26 21:34:15  shaggy
//  C/C++/DC++ header formats now supported
//  header output includes external requirements
//
//  Revision 1.3  1998/05/26 13:40:13  shaggy
//  many more fixes
//
//  Revision 1.2  1998/05/25 23:28:44  shaggy
//  major rework of func/param info
//
//  Revision 1.1  1998/05/25 20:04:41  shaggy
//  renamed Object.hh to CObj.hh
//  (since it's quite specialized right now)
//
//  Revision 1.14  1998/05/15 00:17:26  shaggy
//  ctor initializers work, and enums are fixed so declarations
//  in an enum can be immediately re-used
//
//  Revision 1.13  1998/05/14 18:02:04  shaggy
//  fixed constant handling, enums, function cv tags, etc
//
//  Revision 1.12  1998/05/14 05:08:41  shaggy
//  inheritence works and lots of bugs fixed
//
//  Revision 1.11  1998/05/08 13:03:23  shaggy
//  fixed const/volatile
//
//  Revision 1.10  1998/05/07 14:46:39  shaggy
//  const/volatile partially working
//
//  Revision 1.9  1998/05/04 04:39:49  shaggy
//  first stage of function overloading
//
//  Revision 1.8  1998/05/04 03:09:20  shaggy
//  support for ctors, dtors, conversion operators, operator overloading, etc
//
//  Revision 1.7  1998/04/22 12:17:50  shaggy
//  changed ID to Id for consistency
//
//  Revision 1.6  1998/04/15 14:47:57  shaggy
//  array output works
//
//  Revision 1.5  1998/04/12 18:38:07  shaggy
//  improved code output
//
//  Revision 1.4  1998/02/11 04:33:44  shaggy
//  added const/volatile to ptr/ref
//
//  Revision 1.3  1998/02/08 19:07:05  shaggy
//  about to start splitting up everything
//
//  Revision 1.2  1998/01/11 05:11:03  shaggy
//   about to switch to standard ref/ptr semantics
//
//  Revision 1.1  1997/12/08 04:16:09  shaggy
//  Initial revision
//


#include <defines>
#include <vStrTab>
#include <vSymTab>


// placement new (should be in new.h)

inline void* operator new(size_t, void *p)  { return p; }

typedef size_t bitsize_t;


typedef class Object obj_t;
typedef class Object *type_t;


#define GEN_DCAST(t) inline t* dcast_##t(Object *o) \
	{ return (o && o->is_type(Type_##t)) ? reinterpret_cast<t*>(o->ptr()) : 0; }
#define GEN_COPYCTOR(o, s) \
	o(const s &data) : Object(s::type(), Inline) { new(&_val)s(data); } 
#define GEN_REFCTOR(o, s) \
	o(s *data) : Object(s::type(), Auto) { _ref.ptr = data; }
#define GEN_COPYREFCTOR(o, s)  GEN_COPYCTOR(o, s) GEN_REFCTOR(o, s)


class Object {
public:
	enum StorageType {
		Inline = 0,	// data is stored inline in object descriptor
		Auto = 1,	// ptr to automatic storage 
		Alloc = 2,	// ptr to allocated storage (must free)
		_Reserved = 3	// reserved for future use
	};
protected:
	// the lower 2 bits of the type are used for the StorageType
	// thus all objects used for types must have 32-bit alignment
	vaddr_t _type;
	union {
		uint64	_val;
		double	_fval;
		struct {
			void *ptr;
			uint32 count;
		} _ref;
	};
	obj_t *_refcnt;	// note: some inline objects are 
			// currently flowing over into refcnt, since no
			// bounds checking is done
	Object(type_t type, StorageType st) 
		{ _type = (vaddr_t)type | st; } 
	Object(obj_t &type, StorageType st) 
		{ _type = (vaddr_t)&type | st; } 

public:
	Object()
		{ _type = 0; _val = 0; }

	// Perhaps all of these functions for operating on the object
	// descriptor (rather than the contained object) should be 
	// prefixed in some way to distinguish them?

	type_t type() const
		{ return (type_t)(_type & ~3); }
	StorageType storagetype() const
		{ return (StorageType)(_type & 3); }
	bool is_null() const
		{ return !type(); }
	bool is_type(obj_t &t) const
		{ return type() == &t; }
	bool is_type(type_t t) const
		{ return type() == t; }
	bool is_of() const;	// isa? is_a?  ISA?
	bool is_inline() const { return storagetype() == Inline; } 
	bool is_auto() const { return storagetype() == Auto; } 
	bool is_alloc() const { return storagetype() == Alloc; } 
	void * ptr() { return storagetype() == Inline
			? (void*)(&_val) : _ref.ptr; } 
	const void * ptr() const { return storagetype() == Inline
			? (const void*)(&_val) : (const void*)_ref.ptr; } 
	uint64 data() const { return storagetype() == Inline
				? _val : *(uint64*)_ref.ptr; } 
	size_t _sizeof() const;
	obj_t operator () (obj_t obj);
	obj_t operator () (obj_t *obj);
	operator bool () { return !is_null(); }
};


#define OBJ_NULL	Object()





// this should be an object describing the type:
extern obj_t Type_Integral;


class Integral {
private:
	uint16	_bits;		// total number of bits
	uint8	_alignment;	// bit alignment, as a power of two
	uint8	_is_signed:1;	// true if signed
	uint8	_is_bigend:1;	// true if big endian
	uint8	_reserved:6;	// reserved
public:
	Integral(int bits, int alignment, bool is_signed, bool is_bigend);
	bool is_signed() const { return _is_signed; }
	bool is_unsigned() const { return !_is_signed; }
	size_t bitsize() const { return _bits; }
	size_t size() const { return ((_bits-1)/8)+1; }
	uint alignment() const { return _alignment; }
	bool is_big_endian() const { return _is_bigend; }
	bool is_little_endian() const { return _is_bigend; }
	static type_t type() { return &Type_Integral; }
};


inline Integral::Integral(int bits, int alignment,
		bool is_signed, bool is_bigend) {
	_bits = bits;
	_alignment = alignment;	
	_is_signed = is_signed;
	_is_bigend = is_bigend;
	_reserved = 0;
}




////////////////////////////////

class obj_Integral : public Object {
public:
	obj_Integral(int bits, int alignment, bool is_signed,
		bool is_bigend) : Object(Integral::type(), Inline)
		{ new(&_val)Integral(bits, alignment, is_signed, is_bigend); }
	operator Integral() { return *(Integral*)ptr(); }
	GEN_COPYREFCTOR(obj_Integral, Integral)
};

GEN_DCAST(Integral)

////////////////////////////////////////////////////////

extern obj_t Type_PtrTo;

class PtrTo {
	const type_t _type;
	bool _is_const:1;
	bool _is_volatile:1;	
public:
	// obj_t* operator*() { return ptr; }
	PtrTo(type_t type, bool c = false, bool v = false) : 
			_type(type), _is_const(c), _is_volatile(v) {}
	type_t rtype() const { return _type; }
	bool is_const() { return _is_const; }
	bool is_volatile() { return _is_volatile; }
	static type_t type() { return &Type_PtrTo; }
};


class obj_PtrTo : public Object {
public:
	obj_PtrTo(type_t type, bool c = false, bool v = false)
			: Object(PtrTo::type(), Inline)
		{ new(&_val)PtrTo(type, c, v); }
	size_t _sizeof() const;
	GEN_COPYREFCTOR(obj_PtrTo, PtrTo)
};


GEN_DCAST(PtrTo)

////////////////////////////////////////////////////////

extern obj_t Type_RefTo;

class RefTo {
	const type_t _type;
	bool _is_const:1;
	bool _is_volatile:1;	
public:
	// obj_t* operator*() { return ptr; }
	RefTo(type_t type, bool c = false, bool v = false) :
			_type(type), _is_const(c), _is_volatile(v) {}
	type_t rtype() const { return _type; }
	bool is_const() { return _is_const; }
	bool is_volatile() { return _is_volatile; }
	static type_t type() { return &Type_RefTo; }
};


class obj_RefTo : public Object {
public:
	obj_RefTo(type_t type, bool c = false, bool v = false)
			: Object(RefTo::type(), Inline)
		{ new(&_val)RefTo(type, c, v); }
	GEN_COPYREFCTOR(obj_RefTo, RefTo)
};

GEN_DCAST(RefTo)

////////////////////////////////////////////////////////

extern obj_t Type_Qualifier;

class Qualifier {
	const type_t _type;
	bool _is_const:1;
	bool _is_volatile:1;	
public:
	Qualifier(type_t type, bool c = false, bool v = false) : 
			_type(type), _is_const(c), _is_volatile(v) {}
	type_t rtype() const { return _type; }
	bool is_const() { return _is_const; }
	bool is_volatile() { return _is_volatile; }
	static type_t type() { return &Type_Qualifier; }
};


class obj_Qualifier : public Object {
public:
	obj_Qualifier(type_t type, bool c = false, bool v = false)
			: Object(Qualifier::type(), Inline)
		{ new(&_val)Qualifier(type, c, v); }
	size_t _sizeof() const;
	GEN_COPYREFCTOR(obj_Qualifier, Qualifier)
};

GEN_DCAST(Qualifier)

////////////////////////////////////////////////////////

extern obj_t Type_Array;


class Array {
	const type_t _type;
	const size_t _size;
public:
	Array(type_t type, size_t size) : _type(type), _size(size) {}
	size_t size() const
		{ return _size; }
	size_t lower_bound() const { return 0; }
	size_t upper_bound() const { return _size-1; }
	type_t rtype() const { return _type; }
	static type_t type() { return &Type_Array; }
};



class obj_Array : public Object {
public:
	obj_Array(type_t type, size_t size) : Object(Array::type(), Inline)
		{ new(&_val)Array(type, size); }
	GEN_COPYREFCTOR(obj_Array, Array)
};

GEN_DCAST(Array)

////////////////////////////////////////////////////////


// Params defines the input signature for a function
// and Func defines the complete I/O signature (Param and return type)

// Note: Limited to 255 specified parameters for any function
// (and if any function needs more than about 8, something is probably 
// terribly wrong with the design and organization of the
// function.  Most should never need more than 2 or 3 max...
// Remember that the human mind usually only keeps 7 plus or minus 2
// items in short term memory, so some sort of chunking should be
// used for to keep track of any more than that.
// Okay, I'll get off my soapbox now...)


extern obj_t Type_Params;

class Params {
private:
	const type_t	*_ptype;
	const uint8	_nparam;
	const bool _is_variable:1;
public:
	Params(int nparam, const type_t *ptype, 
		bool variable = false)
		      : _ptype(ptype), _nparam(nparam),
			_is_variable(variable) {}
	bool is_variable() const { return _is_variable; }
	uint nparam() const { return _nparam; }
	type_t ptype(uint index) const { return _ptype[index]; }
	static type_t type() { return &Type_Params; }
};

class obj_Params : public Object {
public:
	obj_Params(int nparam, const type_t *ptype, 
		bool variable = false)
		: Object(Params::type(), Inline)
		{ new(&_val)Params(nparam, ptype, variable); }
	GEN_COPYREFCTOR(obj_Params, Params)
};

GEN_DCAST(Params)

extern obj_t Type_Func;

class Func {
private:
	const type_t	_rtype;
	const type_t	_psig;
	vSymTab		*_scope;
public:
	Func(type_t rtype, type_t param, vSymTab *scope)
		      : _rtype(rtype), _psig(param), _scope(scope) {}
	bool is_variable() const { return ((Params*)_psig->ptr())->is_variable(); }
	uint nparam() const { return ((Params*)_psig->ptr())->nparam(); }
	type_t ptype(uint n) const { return ((Params*)_psig->ptr())->ptype(n); }
	type_t rtype() const { return _rtype; }
	type_t psig() const { return _psig; }
	vSymTab *scope() const { return _scope; }
	static type_t type() { return &Type_Func; }
	void change_rtype(type_t type) { *(type_t*)&_rtype = type; } /* XXX - HACK */
	void change_scope(vSymTab *scope) { _scope = scope; } /* XXX - HACK */
};


class obj_Func : public Object {
public:
	obj_Func(type_t rtype, type_t param, vSymTab *scope)
		: Object(Func::type(), Inline)
		{ new(&_val)Func(rtype, param, scope); }
	GEN_COPYREFCTOR(obj_Func, Func)
};

GEN_DCAST(Func)

////////////////////////////////////////////////////////

extern obj_t Type_Method;

class Method {
	const type_t _otype;
	const type_t _ftype;
public:
	Method(type_t otype, type_t ftype) :
			_otype(otype), _ftype(ftype) {}
	type_t ftype() const { return _ftype; }
	type_t otype() const { return _otype; }
	static type_t type() { return &Type_Method; }
};


class obj_Method : public Object {
public:
	obj_Method(type_t otype, type_t ftype)
			: Object(Method::type(), Inline)
		{ new(&_val)Method(otype, ftype); }
	size_t _sizeof() const;
	GEN_COPYREFCTOR(obj_Method, Method)
};

GEN_DCAST(Method)

////////////////////////////////////////////////////////

/*
class FuncList {
private:
	struct FuncListNode {
		Func *func;
		FuncListNode *next;
	};
	FuncListNode *list;
public:
};

class obj_FuncList : public Object {
public:
	obj_FuncList() : Object(FuncList::type(), Inline)
		{ new(&_val)FuncList(); }
	GEN_COPYREFCTOR(obj_FuncList, FuncList)
};
*/


/*
func: obj(type_func, (ptr to desc, ptr to code))
or
func: obj(exacttype, ptr to code)          -> object as object instance
				cannot be used as type	 
exacttype: obj(type_func, ptr to desc)	   -> object as type,
				but instance of type_func
				can be used as type



integral: obj(class, ptr to methods)
int16: obj(integral, 16-bits)
enum16: obj(integral, 16-bits... etc)
x: obj(enum16, val)

*/


////////////////////////////////////////////////////////


/*
	for function pointers to work normally,
	a simple attribute list is needed that says the type
	and location of dynamic values in the object,

	and another list is needed for static types
	that are defined for the class

	Methods ARE attributes... just ones whose type 
	has the () operator defined

	need to be careful...
	if func ptrs are publicly available, then anyone can get/set them
	which is probably not the intent.
	
	Perhaps some notion of get/set/execute is needed?

	Or individual access methods MUST be used...

	perhaps a 'const' bit is needed for ObjAttr?
	needs to be passed along with object to any methods

	what about no-read access?

	perhaps all object access should be through functions,
	so get() and set() could theoretically filter undesired access

	should get_xxx() and set_xxx() methods exist,
	or should everything go through get() and set()?

	For full access to 'contained' objects, a means of direct 
	attribute access is needed...
	Or at least a way to get a full pointer/handle to the object

	class xx {
		class foo *f;	// here, getting 'f' is sufficient
	};			// must return Obj(type_f, ptr=f)
				// get_f() can easily do this (not ambiguous)
	vs.

	class xx {
		class foo f;	// here, the address of 'f' is needed
	};			// must return Obj(type_f, ptr=&f)
				// get_f() is ambiguous--- should it return
				// the address of f, or a copy of f?
*/



extern obj_t Type_Aggr;

struct ObjAttr {
	vStrTab::Id	name;		// or 0 for private attributes
	type_t		type;
	bitsize_t	offset;		
	bitsize_t	size;		// XXX - or make special types?
};

struct AggrAttr {
	vStrTab::Id	name;
	obj_t		val;
};


class Aggr {
	vStrTab::Id	_name;
	// bitsize_t	size;
	// uint16		nobj_attr;
	// uint16		nclass_attr;
	// ObjAttr		*obj_attr;
	// AggrAttr	*class_attr;
	vSymTab		*_scope;
public:
	Aggr(vStrTab::Id name, vSymTab *scope) : _name(name), _scope(scope) {}
	vStrTab::Id name() const { return _name; }
	vSymTab *scope() const { return _scope; }
	void set_scope(vSymTab *scope) { _scope = scope; }
	static type_t type() { return &Type_Aggr; }
};

class obj_Aggr : public Object {
public:
	obj_Aggr(vStrTab::Id name, vSymTab *scope)
			: Object(Aggr::type(), Auto)
		{ _ref.ptr = new Aggr(name, scope); }
	size_t _sizeof() const;
	GEN_COPYREFCTOR(obj_Aggr, Aggr)
};

GEN_DCAST(Aggr)

////////////////////////////////////////////////////////

extern obj_Integral Type_char;
extern obj_Integral Type_int;
extern obj_Integral Type_short;
extern obj_Integral Type_long;
extern obj_Integral Type_long2;
extern obj_Integral Type_schar;
extern obj_Integral Type_uchar;
extern obj_Integral Type_uint;
extern obj_Integral Type_ushort;
extern obj_Integral Type_ulong;
extern obj_Integral Type_ulong2;

extern obj_t	Type_void;
extern obj_t	Type_bool;

extern obj_t	Type_ptr;
extern obj_t	Type_ref;

extern obj_t	Type_aggr;

extern obj_t	Type_List;
extern obj_t	Type_TokenList;

extern obj_t	Type_Str;

extern obj_t	Type_wchar;
extern obj_t	Type_float;
extern obj_t	Type_double;


// obj_t num2 = 15;

// obj_t num(Type_int, 12); 


/* obj_t num = Type_int(12); */


/*
Obj Type_int = Type_Integral(32, 5, true, false);
Obj Type_long = Type_Integral(32, 5, true, false);
Obj Type_longlong = Type_Integral(64, 6, true, false);
Obj Type_short = Type_Integral(16, 4, true, false);
Obj Type_char = Type_Integral(8, 3, true, false);
*/




/*
for each variable, only type matters for now (although data will
be needed later)
types could be, for instance:

	Type_bool -> ??
	Type_int -> obj(Type_Integral, 32)
	Type_char -> obj(Type_Integral, 8)
	Type_fvoid0 -> obj(Type_Func, void, 0 param)


	in a class:

		foo, obj(Type_fvoid0, ptr_to_foo)

	if function pointers are objects, then they can be remotely
	passed as well, and automatically create proxy objects for RPC
	in this manner...
	
	Every object still needs a way to flatten itself to a network
	standard and to reproduce appropriate proxies at the other side

	64-bits is enough for two pointers inlined in each object
	For a simple linked list, this allows 32-bit data and a pointer
	to the next.
	For larger data, it allows a data ptr and a next
	
	For trees, it allows either 32-bit data or a ptr to the data,
	and a ptr to the left/right pair.
	(or a single ptr to the data/left/right)


Assorted ideas:

	make 'obj_t' the new default type (instead of int)

	Thus, simple procedures like:

			max(x, y) { if(x >= y) return x; else return y; }

	are possible.
	This is more or less equivalent to a template,
	except that only one body of code exists at runtime,
	and dynamically adapts using the types of the objects

Shell idea:

	Use  @( shell expr )  for standard shell expressions

	Within this, any name can be substituted by {} or ()
	for standard C/C++ code

	Use ::{ global C++ code } for global scope expressions

	or scope::{  } for any other scope

	By default, the shell-expression mode is in effect, which
	means ::{} must be used for C++ code/defines
	
	`expr` is unnecessary... Hmmm....
	it could be replaced by @(expr)
	OR...
	`expr` could just be used instead...

	Of course, then backslashing of ` marks is necessary for nesting.

	Perhaps both should be supported, where `` actually executes a new
	shell to run the command (as it traditionally would)

	$ substitution must take place in "" and `` but not ''
*/


