/*  $Id: ccint.cc,v 1.32 1998/07/11 23:06:47 shaggy Exp $
 *      C++ internals (knowledge representation and manipulation)
 *      
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: ccint.cc,v $
//  Revision 1.32  1998/07/11 23:06:47  shaggy
//  fixed labels and goto
//
//  Revision 1.31  1998/07/08 13:42:42  shaggy
//  started adding proper function calls
//
//  Revision 1.30  1998/07/06 22:49:42  shaggy
//  reworked symtab, output ordering, etc
//
//  Revision 1.29  1998/07/05 02:47:37  shaggy
//  tons of bug fixes
//
//  Revision 1.28  1998/07/04 01:23:29  shaggy
//  anonymous unions work now
//
//  Revision 1.27  1998/07/03 19:06:25  shaggy
//  changed scope semantics; fn name is available in scope now
//
//  Revision 1.26  1998/07/03 16:53:51  shaggy
//  new Scope wrapper class, and dcast_XXX functions eliminate lots of casts
//
//  Revision 1.25  1998/06/30 13:23:32  shaggy
//  fcall fixes for d++ conv
//
//  Revision 1.24  1998/06/30 12:22:32  shaggy
//  many fixes for "this"
//
//  Revision 1.23  1998/06/29 01:36:54  shaggy
//  fixed inline attr, output of pure virtuals, etc...
//
//  Revision 1.22  1998/06/08 04:07:30  shaggy
//  more fixes
//
//  Revision 1.21  1998/06/07 18:09:03  shaggy
//  added types to Expr()'s, and fixed output of members (. and ->)
//
//  Revision 1.20  1998/05/26 13:40:13  shaggy
//  many more fixes
//
//  Revision 1.19  1998/05/26 05:02:54  shaggy
//  parameter scoping works correctly
//  most basic C/C++ statements now work
//
//  Revision 1.18  1998/05/25 20:08:40  shaggy
//  more bugfixes
//
//  Revision 1.17  1998/05/24 23:39:32  shaggy
//  various bug fixes and code cleanup
//
//  Revision 1.16  1998/05/20 14:41:23  shaggy
//  better code output
//
//  Revision 1.15  1998/05/19 14:55:35  shaggy
//  a bit closer to generating code (stmtlists work)
//
//  Revision 1.14  1998/05/17 23:14:57  shaggy
//  expressions, expression lists, etc
//
//  Revision 1.13  1998/05/16 23:30:43  shaggy
//  added CTOR_ID and some rules to help fix ctor ambiguity problem
//  output turned off for #include'ed files
//  better error checking, and ability to list locations
//  of symbol declaration and definition
//
//  Revision 1.12  1998/05/15 07:03:37  shaggy
//  dependency ordering fixed; constructor initializers; lots of bug fixes, etc
//
//  Revision 1.11  1998/05/14 05:08:41  shaggy
//  inheritence works and lots of bugs fixed
//
//  Revision 1.10  1998/05/13 21:02:48  shaggy
//  deinitions/initializations work better now
//
//  Revision 1.9  1998/05/12 15:53:21  shaggy
//  scoping and class tags almost work
//
//  Revision 1.8  1998/05/12 05:00:13  shaggy
//  more scope work
//
//  Revision 1.7  1998/05/08 13:03:23  shaggy
//  fixed const/volatile
//
//  Revision 1.6  1998/05/07 14:46:39  shaggy
//  const/volatile partially working
//
//  Revision 1.5  1998/05/05 13:00:42  shaggy
//  better function parameter matching
//
//  Revision 1.4  1998/05/04 03:09:20  shaggy
//  support for ctors, dtors, conversion operators, operator overloading, etc
//
//  Revision 1.3  1998/04/16 13:35:32  shaggy
//  access control
//
//  Revision 1.2  1998/04/14 03:57:08  shaggy
//  forward reference support
//
//  Revision 1.1  1998/02/08 23:41:58  shaggy
//  Initial revision
//

#include <defines>
#include <posix/string.h>
#include <posix/stdio.h>
#include <StrTab>
#include <SymTab>

#include "ccint.h"


typedef vStrTab::Id Id;

csz* TDecl::dtname[] = {"default", "int", "char", "wchar",
			"bool", "void", "float", "double",
			"enum", "struct", "class", "union",
			"namespace", "dtype", "dclass", "enumval"};

uint8 TDecl::dtflags[] = {
	DTF_SIGNOK|DTF_SHORTOK|DTF_LONGOK, DTF_SIGNOK|DTF_SHORTOK|DTF_LONGOK,
	DTF_SIGNOK, 0,
	0, 0,
	0, DTF_LONGOK,
	DTF_TAG, DTF_TAG|DTF_AGGR|DTF_SCOPE,
	DTF_TAG|DTF_AGGR|DTF_SCOPE, DTF_TAG|DTF_AGGR|DTF_SCOPE,
	DTF_TAG|DTF_SCOPE, DTF_TAG|DTF_AGGR|DTF_SCOPE,
	DTF_TAG|DTF_AGGR|DTF_SCOPE, 0,
};

csz* TDecl::acname[] = {"public", "protected", "private", "?"};

obj_t Type_DeclRef;
obj_t Type_Enum;


Decl*
symget(Id id, const Scope sc, bool tagged)
{
	for(Decl *d = sc.get(id); d; d = d->link) {
		if(!tagged || d->tagged())
			return d;
	}
	for(Decl *a = sc.get(ID_ANON); a; a = a->link) {
		if(DeclRef *dref = dcast_DeclRef(a->ptype)) 
			if(Aggr *aggr = dcast_Aggr(dref->decl()->ptype)) 
				if(Decl *d = symget(id, aggr->scope(), tagged))
					return d;
	}
	for(Decl *i = sc.get(ID_INHERIT); i; i = i->link) {
		if(Decl *d = symget(id, i->scope, tagged))
			return d;
	}
	return 0;
}

Decl*
symfind(Id id, const Scope sc, bool tagged, bool search)
{
	Decl *d;
	if(id == sc.pchid())
		sc = sc.parent();
	do {
		if((d = symget(id, sc, tagged))) 
			return d;
		sc = sc.parent();
	} while(search && sc);
	return 0;
}


Expr::Expr(type_t t)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(TYPE);
	p->rtype = t;
}

Expr::Expr(const Lit &lit)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(LITERAL);
	p->lit = lit;
	p->rtype = lit.type;
}

Expr::Expr(Decl *decl)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(VAL_ID);
	p->decl = decl;
	p->rtype = decl->rtype;
}

Expr::Expr(TokenList *tlist)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(TOKENLIST);
	p->tlist = tlist;
	p->rtype = 0;
}

Expr::Expr(Id id)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(ID);
	p->id = id;
	p->rtype = 0;
}

Expr::Expr(CAction::Token tok, Expr e)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->rtype = e.p ? e.p->rtype : 0;
	switch(tok) {
	case T(OP_INCP):
	case T(OP_DECP):
		p->e.left = e.p;
		p->e.mid = 0;
		p->e.right = 0;
		break;
	default:
		p->e.left = 0;
		p->e.mid = 0;
		p->e.right = e.p;
	}
}

Expr::Expr(CAction::Token tok, Expr e1, Expr e2)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = e1.p;
	p->e.mid = 0;
	p->e.right = e2.p;
	p->rtype = e1.p ? e1.p->rtype : 0;
}

Expr::Expr(CAction::Token tok, Expr e1, Expr e2, Expr e3)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = e1.p;
	p->e.mid = e2.p;
	p->e.right = e3.p;
	p->rtype = e3.p ? e3.p->rtype : 0;
}


ExprList::ExprList(CAction::Token tok)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = 0;
	p->e.mid = 0;
	p->e.right = 0;
	p->rtype = 0;
}

ExprList&
ExprList::append(Expr e)
{
	if(p->e.right)
		p->e.right->next = e.p;
	else
		p->e.left = e.p;
	p->e.right = e.p;
	return *this;
}

int
ExprList::len() const
{
	int n;
	Expr e = first();
	for(n = 0; e; e = e.next())
		n++;
	return n;
}

Stmt::Stmt(const Lit &lit)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(LITERAL);
	p->lit = lit;
	p->rtype = 0;
}


Stmt::Stmt(TokenList *tlist)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(TOKENLIST);
	p->tlist = tlist;
	p->rtype = 0;
}

Stmt::Stmt(CAction::Token tok, Expr e)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = 0;
	p->e.mid = 0;
	p->e.right = e.p;
	p->rtype = 0;
}

Stmt::Stmt(CAction::Token tok, Expr e1, Expr e2)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = e1.p;
	p->e.mid = 0;
	p->e.right = e2.p;
	p->rtype = 0;
}

Stmt::Stmt(CAction::Token tok, Expr e1, Expr e2, Expr e3)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = e1.p;
	p->e.mid = e2.p;
	p->e.right = e3.p;
	p->rtype = 0;
}

StmtList::StmtList(CAction::Token tok, Scope scope)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = 0;
	p->e.sc_mid = scope;
	p->e.right = 0;
	p->rtype = 0;
}

StmtList::StmtList(Scope scope, CAction::Token tok)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = 0;
	p->e.sc_mid = scope;
	p->e.right = 0;
	p->rtype = 0;
}

StmtList&
StmtList::append(Stmt s)
{
	if(p->e.right)
		p->e.right->next = s.p;
	else
		p->e.left = s.p;
	p->e.right = s.p;
	return *this;
}

int
tokenprec(CAction::Token tok, int assoc)
{
	switch(tok) {
	case T(TOKENLIST):
	case T(LITERAL):
	case T(VAL_ID):
		return 100;
	case T(OP_DOT): case T(OP_PTRTO): case T(OP_INDEX):
	case T(KW_SCAST): case T(KW_DCAST): case T(KW_CCAST): case T(KW_RCAST):
	case T(OP_FUNC):
		return 76;
	case T(OP_INCP): case T(OP_DECP):
		return 76+assoc;
	case T(OP_SIZEOF):
	case T(OP_INC): case T(OP_DEC):	
	case T(OP_NOT): case T(OP_LNOT):
	case T(OP_UMINUS): case T(OP_UPLUS):
	case T(OP_REF): case T(OP_PTR):
	case T(OP_NEW): case T(OP_DELETE):
	case T(OP_NEWA): case T(OP_DELETEA):
		return 72+assoc;
	case T(OP_MDOT): case T(OP_MPTRTO):
		return 68;
	case T(OP_MUL): case T(OP_DIV): case T(OP_MOD): 
		return 64;
	case T(OP_ADD): case T(OP_SUB):
		return 60;
	case T(OP_LSHIFT): case T(OP_RSHIFT):
		return 56;
	case T(OP_LSS): case T(OP_LEQ): case T(OP_GTR): case T(OP_GEQ):
		return 52;
	case T(OP_EQU): case T(OP_NEQ):
		return 48;
	case T(OP_AND):
		return 44;
	case T(OP_XOR):
		return 40;
	case T(OP_OR):
		return 36;
	case T(OP_LAND):
		return 32;
	case T(OP_LXOR):
		return 28;
	case T(OP_LOR):
		return 24;
	case T(OP_ASN):
	case T(OP_ASNMUL): case T(OP_ASNDIV): case T(OP_ASNMOD):
	case T(OP_ASNADD): case T(OP_ASNSUB): 
	case T(OP_ASNLSHIFT): case T(OP_ASNRSHIFT):
	case T(OP_ASNAND): case T(OP_ASNOR): case T(OP_ASNXOR):
		return 20;
	case T(QMARK):
		return 16;
	case T(KW_THROW):
		return 12+assoc;
	case T(OP_COMMA):
		return 8;
	default:
		return 0;
	}
}

bool
scope_empty(const Scope sc)
{
	Id id;
	for(int n=0; (id = sc.index(n)); n++) 
		if(id != ID_INHERIT)
			return false;
	return true;
}

bool
scope_derived(const Scope sc)
{
	return (sc.get(ID_INHERIT)) ? true : false;
}


void
dump_scope(const Scope scope)
{
	vStrTab *strtab = scope.strtab();
	printf("Scope %p: %s\n", (vSymTab*)(scope), scope.pchid()
			? strtab->string(scope.pchid()) : "?");
	Id id;
	for(int n=0; (id = scope.index(n)); n++) {
		csz *s = "?";
		if(id == ID_INHERIT)
			s = "<inherit>";
		else if(id == ID_USING)
			s = "<using>";
		else if(id == ID_ANON)
			s = "<anon>";
		else if(id)
			s = strtab->string(id);
		printf("\t%s: ", s);
		Decl *d = scope.get(id);
		while(d) {
			printf("%p(%s%s%s) ", d, Decl::dtname[d->dtype],
				d->defined ? ",D" : "",
				d->initialized ? ",I" : "");
			d = d->link;
		}
		printf("\n");
	}
}

Decl *
param_decl(Func *func, int i)
{
	Scope scope = func->scope();
	if(!scope)
		return 0;
	Id id;
	for(int n=0; (id = scope.index(n)); n++) {
		for(Decl *d = scope.get(id); d; d = d->link)
			if(d->where == i)
				return d;
	}
	return 0;
}


bool
subscope(Scope scope, Scope sub)
{
	if(scope) {
		if(sub == scope)
			return true;
		if(Scope parent = sub.parent())
			return subscope(scope, parent);
	}
	return false;
}


void
decladd(Decl *decl)
{
	Scope scope = decl->scope;
	Id id = decl->id;
	if(!id)
		id = ID_ANON;
	decl->link = 0;
	if(!scope.add(id, decl)) {
		Decl *d = scope.get(id);
		decl->link = d;
		scope.set(id, decl);
	}
}

void
declrename(Decl *decl, Id id)
{
	printf("rename\n");
	Id oldid = decl->id;
	if(!oldid)
		oldid = ID_ANON;
	Decl *d = decl->scope.get(oldid);
	if(d == decl) {
		decl->scope.del(oldid);
		if(decl->link)
			decl->scope.add(oldid, decl->link);
	} else {
		while(d->link != decl)
			d = d->link;
		d->link = decl->link;
	}
	decl->id = id;
	decladd(decl);
	printf("rename done\n");
}


Id
Scope::pchid() const
{
	Decl *d = decl();
	return d ? d->id : 0;
}

