/*  $Id: ccout.h,v 1.29 1998/07/12 03:58:48 shaggy Exp $
 *      C++ output
 *
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: ccout.h,v $
//  Revision 1.29  1998/07/12 03:58:48  shaggy
//  lots of const fixes, -frealtype option, -dT option; better debugging
//
//  Revision 1.28  1998/07/11 19:03:03  shaggy
//  declarations in for() work now (even multiple decls)
//
//  Revision 1.27  1998/07/10 13:45:10  shaggy
//  working on decls in conditions
//
//  Revision 1.26  1998/07/04 20:39:00  shaggy
//  lots of bug fixes; fn, aggr, etc improvements, new/delete
//
//  Revision 1.25  1998/07/03 16:53:53  shaggy
//  new Scope wrapper class, and dcast_XXX functions eliminate lots of casts
//
//  Revision 1.24  1998/06/07 15:45:13  shaggy
//  realtype() works, and . and -> now work
//
//  Revision 1.23  1998/05/26 21:34:16  shaggy
//  C/C++/DC++ header formats now supported
//  header output includes external requirements
//
//  Revision 1.22  1998/05/26 13:40:15  shaggy
//  many more fixes
//
//  Revision 1.21  1998/05/25 20:08:41  shaggy
//  more bugfixes
//
//  Revision 1.20  1998/05/25 04:22:14  shaggy
//  enough for tonight
//
//  Revision 1.19  1998/05/24 23:39:34  shaggy
//  various bug fixes and code cleanup
//
//  Revision 1.18  1998/05/24 18:42:00  shaggy
//  improved regression testing
//  proper scoping for initializers
//  decls available for their initializer
//
//  Revision 1.17  1998/05/24 01:41:43  shaggy
//  internal support for most statements
//
//  Revision 1.16  1998/05/22 14:42:40  shaggy
//  fixed output formatting
//
//  Revision 1.15  1998/05/21 13:27:23  shaggy
//  output formatting problems
//
//  Revision 1.14  1998/05/20 14:41:25  shaggy
//  better code output
//
//  Revision 1.13  1998/05/19 14:55:37  shaggy
//  a bit closer to generating code (stmtlists work)
//
//  Revision 1.12  1998/05/17 23:14:58  shaggy
//  expressions, expression lists, etc
//
//  Revision 1.11  1998/05/16 23:30:45  shaggy
//  added CTOR_ID and some rules to help fix ctor ambiguity problem
//  output turned off for #include'ed files
//  better error checking, and ability to list locations
//  of symbol declaration and definition
//
//  Revision 1.10  1998/05/14 05:08:42  shaggy
//  inheritence works and lots of bugs fixed
//
//  Revision 1.9  1998/05/13 21:02:49  shaggy
//  deinitions/initializations work better now
//
//  Revision 1.8  1998/05/12 22:40:25  shaggy
//  scoping FINALLY works right (or seems to)
//
//  Revision 1.7  1998/05/12 15:53:22  shaggy
//  scoping and class tags almost work
//
//  Revision 1.6  1998/05/12 05:00:14  shaggy
//  more scope work
//
//  Revision 1.5  1998/04/17 04:37:41  shaggy
//  public/privat/protected works now
//
//  Revision 1.4  1998/04/12 18:38:08  shaggy
//  improved code output
//
//  Revision 1.3  1998/04/12 03:19:35  shaggy
//  got function output working
//
//  Revision 1.2  1998/04/11 21:48:42  shaggy
//  added some simple regurgitation
//
//  Revision 1.1  1998/02/08 23:41:58  shaggy
//  Initial revision
//

#include <defines>
#include <CObj>
#include <CLex>
#include <CParse>
#include <CAction>
#include <StrTab>
#include <SymTab>
#include <TokenList>


class CCOut {
	const Scope _scope;
	const vStrTab *_strtab;
	TDecl::Access _access;
	ErrorReport &_err;
	char _buf[65536];
	CLex _clex;
	Decl *_curdecl;		// for error reporting
	int _nl;
	int _x;
	int _indent;
	int _tabwid;
	char *_obuf;
	bool _fullscope;
	bool _debugtypes;

	void printliteral(const type_t type, uint64 val);
	// void printlist(ExprListNode *n);
	// static bool has_dependencies(Decl *decl);
	// static bool has_dependencies(type_t type, bool needdef);
	// static bool is_type(Decl *decl);
	void print_storageclass(Decl *decl);
	void regurgitate(TokenList *t);
	void print_access(Decl *decl);
	void define_sym(Decl *decl, bool semi = true);
	void define_enum(Decl *decl);
	void define_aggr(Decl *decl);
	void declare_sym(Decl *decl);
	void print_types(const Scope);
	void print_fndecls(const Scope, bool exp, bool stat);
	void print_vardecls(const Scope, bool exp);
	void _assertfail(csz *cond, csz *filename, int line);
public:
	CCOut(ErrorReport &err, const Scope,
		TDecl::Access access = TDecl::ac_public);
	void debugtypes(bool d = true) { _debugtypes = d; }
	void printexpr(Expr, bool paren = false);
	void printstmt(Stmt);
	void printstmti(Stmt);
	void printexprlist(ExprList);
	void printstmtlist(StmtList);
	void printdecllist(StmtList);
	void printctorinit(StmtList);
	sz* literal_string(const type_t type, uint64 val, sz *s);
	sz* type_string(const type_t type, sz *s);
	sz* dbg_type_string(const type_t type, sz *s);
	void print_decls(const Scope, TDecl::Access acc = TDecl::ac_public);
	void print_defs(const Scope);
	void print_reqs(const Scope);
	sz *scoped_string(const Scope, sz *s);
	sz *scoped_string(Decl *decl, sz *s, bool needtag = false);
	sz *fullscoped_string(Decl *decl, sz *s);
	void fullscope(bool tf = true) { _fullscope = tf; }
	void onl(int y = 1);
	void onltab(int x = -1);
	void rshift()  { _indent += _tabwid; }
	void lshift()  { _indent -= _tabwid; }
	int oprintf(csz *fmt, ...);
};


sz *strpcat(sz *str, csz *prefix);
csz *simple_typename(const type_t type);

