/*  $Id: YLex.cc,v 1.5 1998/05/24 18:42:04 shaggy Exp $
 *      Yacc-style Lexical Analysis
 *      
 *      relative path: ./lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

  
//  $Log: YLex.cc,v $
//  Revision 1.5  1998/05/24 18:42:04  shaggy
//  improved regression testing
//  proper scoping for initializers
//  decls available for their initializer
//
//  Revision 1.4  1998/04/13 13:17:01  shaggy
//  added %uniq
//
//  Revision 1.3  1998/03/17 19:14:24  shaggy
//  fixing some damage from the move
//
//  Revision 1.2  1998/03/03 14:25:52  shaggy
//  *** empty log message ***
//
//  Revision 1.1  1998/02/23 14:33:51  shaggy
//  working on "shacc"
//
//


#define IMPLEMENTATION_YLex
#include "YLex.hh"
#include <stdio.h>

YLex::YLex(char *s, size_t ssize)
{
	str = s; str_size = ssize;
	buf = s; buf_size = ssize;
	name = s; name_size = ssize;
	state = S_NEXT;
}


YLex::Token
YLex::next_token()
{
// Macros for processing tokens and switching states:
//	RC: Return/Continue: return token
//	RBS: Return/Break/State: return token, eat char, state = s
//	RCS: Return/Continue/State: return token, state = s
//	RBNEXT: Return/Break/NEXT: return token, eat char, state = S_NEXT
//	RCNEXT: Return/Cont/NEXT: return token, state = S_NEXT
//	BS: Break/State: eat char, state = s
//	OBS: Old/Skip/State: old_state = state, eat char, state = s
//	CS: Continue/State: state = s
//	HANDLE_WS: Eat/Store whitespace, returning it if necessary
#define RC(t) { return (token = t); }
#define RBS(t,s) { old_state = s; state = S_SKIP; return (token = t); }
#define RCS(t,s) { state = s; return (token = t); }
#define RBNEXT(t) { state = S_SKIP_NEXT; return (token = t); }
#define RCNEXT(t) { state = S_NEXT; return (token = t); }
#define BS(s) { state = s; break; }
#define OBS(s) { old_state = state; state = s; break; }
#define CS(s) { state = s; continue; }
#define OCS(s) { old_state = state; state = s; continue; }
#define HANDLE_WS { if(is_whitespace()) { strp = str; OCS(S_SPACE) }}


    while(!is_empty()) {
	switch(state) {
	case S_NEXT:
		HANDLE_WS
		_startline = line_num();
		switch(get_char()) {
		case 'A': case 'B': case 'C': case 'D':
		case 'E': case 'F': case 'G': case 'H':
		case 'I': case 'J': case 'K': case 'L':
		case 'M': case 'N': case 'O': case 'P':
		case 'Q': case 'R': case 'S': case 'T':
		case 'U': case 'V': case 'W': case 'X':
		case 'Y': case 'Z': 
		case 'a': case 'b': case 'c': case 'd':
		case 'e': case 'f': case 'g': case 'h':
		case 'i': case 'j': case 'k': case 'l':
		case 'm': case 'n': case 'o': case 'p':
		case 'q': case 'r': case 's': case 't':
		case 'u': case 'v': case 'w': case 'x':
		case 'y': case 'z': 
		case '_':
			strp = str;
			*strp++ = get_char();
			BS(S_ID)
		case '0':
		case '1': case '2': case '3':
		case '4': case '5': case '6': case '7':
		case '8': case '9':
			num = get_digit();
			BS(S_NUM)
		case '%':
			 BS(S_PERCENT)
		case ':': RBNEXT(COLON)
		case '|': RBNEXT(OR)
		case ';': RBNEXT(EOS)
		case '/': BS(S_SCOMMENT)
		case '\'':
			strp = str; BS(S_SQUOTE)
		case '{':
			strp = str;
			*strp++ = get_char();
			_blevel = 1;
			_rtok = ACTION;
			BS(S_CODE)
		default:
			strp = str;
			*strp++ = get_char();
			*strp = 0;
			size = strp - str;
			RBNEXT(INVALID)
		}
		break;
	case S_SKIP:
		BS(old_state)
	case S_SKIP_NEXT:
		BS(S_NEXT)
	case S_SPACE:
		while(is_whitespace())
			if(!next_char())
				RC(INCOMPLETE)
		CS(old_state)
	case S_NUM:
		if(is_num()) num = num*10 + get_digit();
		else { size = 1; RCNEXT(NUM) }
		break;
	case S_ID:
		if(is_id()) {
			*strp++ = get_char();
			size++;
		} else {
			*strp = 0; size = strp-str;
			RCNEXT(ID)
		}
		break;
	case S_PERCENT:
		if(match('{')) {
			_rtok = CODE; _blevel = 1; BS(S_CODE)
		} else if(match('%')) {
			RBNEXT(SECTION)
		} else if(is_id()) {
			strp = str;
			CS(S_KW)
		} else
			RBNEXT(INVALID)
	case S_KW:
		if(is_id()) {
			*strp++ = get_char();
			size++;
		} else {
			*strp = 0; size = strp-str;
			for(int t = START_KW; t<= END_KW; t++)
				if(!strcmp(&token_table[t-ID][1], str)) 
					RCNEXT(Token(t))
			RBNEXT(INVALID)
		}
		break;
	case S_CODE:
		switch((*strp++ = get_char())) {
		case '{':
			if(_rtok == ACTION)
				_blevel++;
			break;
		case '}':
			if(_rtok == ACTION) 
				if(--_blevel == 0) {
					*strp = 0; size = strp-str;
					RBNEXT(ACTION)
				}
			break;
		case '$':
			if(_rtok == ACTION) 
				BS(S_CODE_DOLLAR)
			break;
		case '%':
			if(_rtok == CODE) {
				*strp = 0; size = strp-str;
				BS(S_CODE_PERCENT)
			}
			break;
		case '/':
			BS(S_CODE_SCOMMENT)
		case '\'':
			BS(S_CODE_SQUOTE)
		case '"':
			BS(S_CODE_DQUOTE)
		}
		break;
	case S_CODE_PERCENT:
		if(match('}')) {
			*(--strp) = 0; size = strp-str;
			RBNEXT(CODE)
		} else {
			*strp++ = get_char();
			BS(S_CODE)
		}
		break;
	case S_CODE_DOLLAR:
		if(match('$')) {
			--strp;
			strp += sprintf(strp, "%s", _resultname);
			BS(S_CODE)
		}
		if(match('-')) {
			--strp;
			num = 0;
			neg = true;
			BS(S_CODE_NUM);
		}
		if(is_num()) { 
			--strp;
			num = get_digit();
			neg = false;
			BS(S_CODE_NUM)
		}
		CS(S_CODE)
	case S_CODE_NUM:
		if(is_num()) num = num*10 + get_digit();
		else {
			if(neg) 
				num = -num;
			strp += sprintf(strp, "%s[%ld]", _stackname,
							(num-1)+_stackofs);
			CS(S_CODE)
		}
		break;
	case S_CODE_SCOMMENT:
		*strp++ = get_char();
		if(match('*')) BS(S_CODE_COMMENT)
		else if(match('/')) { old_state = S_CODE; BS(S_EAT) }
		else BS(S_CODE)
	case S_CODE_COMMENT:
		*strp++ = get_char();
		if(match('*')) BS(S_CODE_COMMENT2)
		break;
	case S_CODE_COMMENT2:
		*strp++ = get_char();
		if(match('/')) BS(S_CODE)
		else BS(S_CODE_COMMENT)
	case S_CODE_SQUOTE:
		*strp++ = get_char();
		if(match('\\')) OBS(S_CODE_QUOTE_ESC)
		if(match('\'')) BS(S_CODE)
		break;
	case S_CODE_DQUOTE:
		*strp++ = get_char();
		if(match('\\')) OBS(S_CODE_QUOTE_ESC)
		if(match('\"')) BS(S_CODE)
		break;
	case S_CODE_QUOTE_ESC:
		*strp++ = get_char();
		state = old_state;
		break;
	case S_SQUOTE:
		if(match('\\')) OBS(S_QUOTE_ESC)
		else if(match('\'')) {
			*strp = 0; size = strp-str; RBNEXT(CHAR)
		} else { *strp++ = get_char(); break; }
	case S_QUOTE_ESC:
		state = old_state;
		switch(get_char()) {
		case 'a': *strp++ = '\a'; break;
		case 'b': *strp++ = '\b'; break;
		case 'd': *strp++ = '\177'; break;
		case 'e': *strp++ = '\033'; break;
		case 'f': *strp++ = '\f'; break;
		case 'n': *strp++ = '\n'; break;
		case 'r': *strp++ = '\r'; break;
		case 't': *strp++ = '\t'; break;
		case 'v': *strp++ = '\v'; break;
		case '\n': break;
		case 'x': num = 0; BS(S_QUOTE_HEX)
		case '0' ... '9':
			num = get_digit(); BS(S_QUOTE_OCT)
		default:
			*strp++ = get_char();
		}
		break;
	case S_QUOTE_OCT:
		if(is_oct()) { num = num*8 + get_digit(); break; }
		else { *strp++ = num; CS(old_state) }
	case S_QUOTE_HEX:
		if(is_hex()) { num = num*16 + get_hex(); break; }
		else { *strp++ = num; CS(old_state) }
	case S_SCOMMENT:
		if(match('*')) BS(S_COMMENT)
		else if(match('/')) { old_state = S_NEXT; BS(S_EAT) }
		else RBNEXT(INVALID)
	case S_COMMENT:
		if(match('*')) BS(S_COMMENT2)
		break;
	case S_COMMENT2:
		if(match('/')) BS(S_NEXT)
		else BS(S_COMMENT)
	case S_EAT:
		for(;;) {
			*strp++ = get_char();
			if(eol())
				break;
			if(!next_char())
				RC(INCOMPLETE)
		}
		BS(old_state)
	}
	next_char();
    }
    if(state == S_NEXT || state == S_SPACE)
	if(eof()) RC(EOF_COMPLETE)
	else RC(NONE)
    else
	if(eof()) RC(EOF_INCOMPLETE)
	else RC(INCOMPLETE)

#undef RC
#undef RBS
#undef RCS
#undef RBNEXT
#undef RCNEXT
#undef BS
#undef OBS
#undef CS
#undef HANDLE_WS
}


uint
YLex::last_char() const
{
uint num = 0;
csz *cp = str;
	if(*cp) do {
		num += *cp;
	} while(*++cp && ((num <<= 8), true)); 
	return num;
}


csz*
YLex::token_string()
{
	switch(token) {
	case NONE: case INCOMPLETE:
		return "";
	case NUM:
		return token_string(token, last_num());
	case INVALID:
	case CHAR: case CODE: case ID:
		return token_string(token, last_str(), last_size());
	case EOS: return ";";
	default:
		return token_string(token);
	}
}


csz*
YLex::token_string(Token t) const
{
	if(t >= ID && t < NUM_TOKENS)
		return token_table[t-ID];
	else if(t == NONE)
		return "<NONE>";
	else
		return "<??>";
}


csz*
YLex::token_string(Token t, uint32 val)
{
	if(t != NUM)
		return token_string(t);
	bufp = buf;
	*bufp++ = '0';
	*bufp++ = 'x';
	register char *endp = bufp;
	for(uint32 n = val; *endp = '0'+(n%10), (n%10); n /= 10)
		endp++;
	register char *bp = endp++;
	while(bufp < bp) {
		char t = *bufp; *bufp++ = *bp; *bp-- = t;
	}
	bufp = endp;
	*bufp = 0;
	return buf;
}


csz*
YLex::token_string(Token t, register csz *s, size_t len)
{
	char quote_start, quote_end;
	switch(t) {
	case INVALID:
	case CHAR: quote_start = quote_end = '\''; break;
	case ID: quote_start = quote_end = 0; break;
	default:
		return token_string(t);
	}
	if(len == NPOS) {
		len = 0;
		for(register csz *sp = s; *sp; sp++)
			len++;
	}
	bufp = buf;
	if(quote_start)
		*bufp++ = quote_start;
	while(len--) {
		char ch = *s++;
		if(ch < 32 || ch > 126) {
			*bufp++ = '\\';
			switch(ch) {
			case '\a': *bufp++ = 'a'; break;
			case '\b': *bufp++ = 'b'; break;
			case '\f': *bufp++ = 'f'; break;
			case '\n': *bufp++ = 'n'; break;
			case '\r': *bufp++ = 'r'; break;
			case '\t': *bufp++ = 't'; break;
			case '\v': *bufp++ = 'v'; break;
			case '\0': *bufp++ = '0'; break;
			default:
				uint n = (uchar)ch;
				*bufp++ = '0'+(n>>6);
				*bufp++ = '0'+((n>>3)&7);
				*bufp++ = '0'+(n&7);
			}
		} else if(ch == quote_end || ch == '\\') {
			*bufp++ = '\\';
			*bufp++ = ch;
		} else {
			*bufp++ = ch;
		}
	}
	if(quote_end)
		*bufp++ = quote_end;
	*bufp = 0;
	return buf;
}

// Note that all of the tokens in this table
// must be in the exact same order as their
// corresponding token names in YLex.hh

csz *YLex::token_table[] = {
	"<ID>", "<NUM>", "<CHAR>", "<CODE>", "<ACTION>",

	"%%", ":", "|", ";",

	"%token", "%union", "%start",
	"%left", "%right", "%nonassoc", "%prec",
	"%rprec", "%uniq",
	"%pure_parser"
};

