#line 344 "CParse.y"
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#pragma Make Exclusive

#include "ActionStack.h"
#include <TokenList>

#define IMPLEMENTATION_CParse
#include "CParse.hh"


#define YYSTYPE CParse::Item


#define yyassert(c)	{ if(!(c)) INTERNAL_ERROR; }
#define yyerror 	_caction->error
#define yywarning 	_caction->warning

#define YYCLASS		CParse
#define YYPARSE		_do_parse
#define YYLEX		_peekahead_lex
#define YYDEBUGFLAG	_yydebug
#define YYEPRINTF	_caction->eprintf
#define YYPARSE_VARS	\
	ActionStack actionstack(_caction);

//#define YYPDEBUG


// XXX - eventually, all of the constant strings (for errors/warnings)
// should be removed and replaced with error ID's
// that can be passed to CAction where they can have defaults
// and overrides.  This will also aid internationalization.
// The position specifier needs to be added to printf for this
// to entirely work correctly.



#define swap_items(a,b)	{ Item tmp = a; a = b; b = tmp; }


#define eval_decl(r, a, t) \
	{ actionstack.pop_lifo(&(r), (a), decl_actions, (t)); }
#define eval_pdecl(r, a, t) \
	{ actionstack.pop_lifo(&(r), (a), pdecl_actions, (t)); }

#define declSym(r, s, d) \
	{ actionstack.push(&(r), none, ACTION_Decl, T(TBC), (d), (s)); }
#define defineSym(r, s, v) \
	{ actionstack.push(&(r), none, ACTION_Define, (s), (v), \
						none); }
#define declPtr(r, op, d) \
	{ actionstack.push(&(r), (d), ACTION_Ptr, (op), T(TBC)); }
#define declArray(r, d, i) \
	{ actionstack.push(&(r), (d), ACTION_Array, T(TBC), i, 0); }
#define declFunc(r, d, p, c) \
	{ actionstack.push(&(r), (d), ACTION_Func, T(TBC), (p), (c)); }
#define declInit(r, d, v) \
	{ actionstack.push(&(r), (d), ACTION_Init, T(TBC), (p)); }
#define buildParamList(r, l, i) \
	{ actionstack.push(&(r), none, ACTION_BuildList, (l), (i)); }
#define buildDeclList(r, l, i) \
	{ actionstack.push(&(r), none, ACTION_None, (l), (i)); }

#define declInherit(r, t, b) \
	{ actionstack.push(&(r), none, ACTION_Inherit, T(TBC), (t), (b)); }


#define UNSUPPORTED yyerror("unsupported syntax (%d)", __LINE__);
#define NOSCOPE(s) if(s.scope) yyerror("scope not allowed");



enum ActionIndex {
	ACTION_None = 0,
	ACTION_Decl,
	ACTION_Define,
	ACTION_Ptr,
	ACTION_Array,
	ACTION_Func,
	ACTION_BuildList,
	ACTION_Inherit,
};


static const CAction::Action decl_actions[] = {
	0,		/* none */
	A(DeclSym),	/* declarator */
	A(Define),	/* define initial value */
	A(DeclPtr),	/* pointer (*) or reference (&) declaration */
	A(DeclArray),	/* array declaration ([n]) */
	A(DeclFunc),	/* function declaration */
	A(BuildParamList), /* parameter list (,) */
	A(Inherit)	/* inherit from a base class */
};

static const CAction::Action pdecl_actions[] = {
	0,		/* none */
	A(DeclParam),	/* declarator */
	A(DefParam),	/* default value */
	A(DeclPtr),	/* pointer (*) or reference (&) declaration */
	A(DeclArray),	/* array declaration ([n]) */
	A(DeclFunc),	/* function declaration */
	A(BuildParamList), /* parameter list (,) */
	0,		/* none */
};


/* Note: when defining parameter lists, the current scope,
	which is used for each parameter, can either be some arbitrary
	undefined scope, for arbitrary declarations,
	or the scope of the function, for the header on the definition */

/* Need a 'kill_scope' to get rid of unneeded scopes later, and
   reference counts should probably be used on all items so
   that those items referenced in a scope which are still needed 
   will be available */



#define YYDEBUG 1

typedef signed char yys8;
typedef unsigned char yyu8;
typedef short yys16;
typedef unsigned short yyu16;
typedef int yys32;
typedef unsigned int yyu32;

#ifndef YYSTYPE
#	define	YYSTYPE	int
#endif

typedef yys16 YYTTYPE;
typedef yyu16 YYRTYPE;
/* Parser code for shacc
 *
 * possible #define's:
 *	YYLEX		- lexer function [yylex]
 *	YYPARSE		- parser function [yyparse]
 *	YYDEBUGFLAG	- debugging flag [yydebug]
 *	YYERROR		- error reporting function [yyerror]
 *	YYEPRINTF	- error printf [yyeprintf]
 *	YYCLASS		- class name to encapsulate parser in
 *	YYLEX_PARAM	- parameter for lexer function
 *	YYPARSE_PARAM	- parameter for parser function
 *	YYPARSE_VARS	- local variables for parsing
 */

/* Note:  I really hate excessive #ifdef's and #define's like the following,
 * so if anyone has a much cleaner solution to the problem of keeping this
 * one source file compatible with many different variations, and still
 * allowing default values, please let me know...
 * (keep in mind that this code needs to work for both C and C++ as well)
 */
#ifndef YYDEBUG
#	ifdef YYPDEBUG
#		error The '-t' flag is required if YYPDEBUG is defined
#	else
#		define YYDEBUG 0
#	endif
#endif
#ifndef YYDEBUGFLAG
#	define YYDEBUGFLAG	yydebug
#endif
#ifndef YYLEX
#	define	YYLEX	yylex
#endif
#ifndef	YYPARSE
#	define	YYPARSE	yyparse
#endif
#ifndef YYMALLOC
#	define	YYMALLOC	malloc
#	include <stdlib.h>
#endif
#ifndef YYFREE
#	define YYFREE		free
#endif
#ifndef YYEPRINTF
#	define	YYEPRINTF	printf
#	include <stdio.h>
#endif
#ifndef YYERROR
#	define	YYERROR		yyerror
#endif
#ifndef YYINTERR_
#	define	YYINTERR_(f,n)	YYERROR("internal parser error (" \
					## f ## ":" ## YYQUOTE(n) ## ")\n")
#endif
#ifndef YYINTERR
#	define	YYINTERR	YYINTERR_("shacc.c.parser", __LINE__)
#endif
#ifdef YYIMPURE
#	define YYLEXCALL YYLEX()
#else
#	ifdef YYLEX_PARAM
#		define YYLEXCALL YYLEX(&yylval, YYLEX_PARAM)
#	else
#		define YYLEXCALL YYLEX(&yylval)
#	endif
#endif
#ifdef YYCLASS
#	define YYYYCLASS	yy##YYCLASS
#	define YYPREFIX		YYYYCLASS::
#	define YYSTATIC
#else
#	define YYPREFIX
#	define YYSTATIC		static
#endif
#ifdef YYPARSE_PARAM
#	define YYPARSEPARAMS	void *YYPARSE_PARAM
#	define YYPARSEPARAM	YYPARSE_PARAM
#else
#	define YYPARSEPARAMS
#	define YYPARSEPARAM
#endif
#ifndef YYPARSE_VARS
#	define YYPARSE_VARS	
#endif

#ifdef YYYACC
#	if YYDEBUG
		int	YYDEBUGFLAG;
#	endif
#endif


#define YYQUOTE(x)	#x

/* temporary -- for debugging this parser code */
//#define YYPDEBUG
#ifdef YYPDEBUG
#	define DBG(x)	x
#else
#	define DBG(x)
#endif


enum {
	YYNRULES = 508,
	YYNNONTERMS = 182,
	YYNTOKENS = 453,
	YYNSYMS = 858,
	YYINITDEPTH = 1000,
#line 106 "/u4/shaggy/u2/src/sos/util/shacc/shacc.c.parser"
};

enum YYToken {
	/* $ = 0 */
	/* error = 1 */
	/* ? = 2 */
	_ID = 257,
	_NUM = 258,
	_CHAR = 259,
	_FLOAT = 260,
	_STR = 261,
	_CSTR = 262,
	_BOOL = 263,
	_BACKSTR = 264,
	_ANGSTR = 265,
	_LINE = 266,
	_SPACE = 267,
	_EOL = 268,
	_CPP = 269,
	_CPP_JOIN = 270,
	_CPP_DEFINED = 271,
	_CPP_DEFINE = 272,
	_CPP_UNDEF = 273,
	_CPP_IF = 274,
	_CPP_IFDEF = 275,
	_CPP_IFNDEF = 276,
	_CPP_ENDIF = 277,
	_CPP_ELIF = 278,
	_CPP_INCLUDE = 279,
	_CPP_PRAGMA = 280,
	_CPP_ASSERT = 281,
	_CPP_UNASSERT = 282,
	_CPP_LINE = 283,
	_CPP_ERROR = 284,
	_CPP_WARNING = 285,
	_CPP_IDENT = 286,
	_CPP__FILE = 287,
	_CPP__LINE = 288,
	_CPP__DATE = 289,
	_CPP__TIME = 290,
	_CPP__STDC = 291,
	_CPP__STDCVER = 292,
	_EOS = 293,
	_COLON = 294,
	_COMMA = 295,
	_QMARK = 296,
	_POUND = 297,
	_LBRACE = 298,
	_RBRACE = 299,
	_LBRACK = 300,
	_RBRACK = 301,
	_LPAREN = 302,
	_RPAREN = 303,
	_QLINE = 304,
	_DOTDOT = 305,
	_AT = 306,
	_BSLASH = 307,
	_OP_ADD = 308,
	_OP_SUB = 309,
	_OP_MUL = 310,
	_OP_DIV = 311,
	_OP_MOD = 312,
	_OP_NOT = 313,
	_OP_AND = 314,
	_OP_OR = 315,
	_OP_XOR = 316,
	_OP_LNOT = 317,
	_OP_LAND = 318,
	_OP_LOR = 319,
	_OP_LXOR = 320,
	_OP_LSS = 321,
	_OP_GTR = 322,
	_OP_LEQ = 323,
	_OP_GEQ = 324,
	_OP_EQU = 325,
	_OP_NEQ = 326,
	_OP_INC = 327,
	_OP_DEC = 328,
	_OP_LSHIFT = 329,
	_OP_RSHIFT = 330,
	_OP_ASN = 331,
	_OP_ASNADD = 332,
	_OP_ASNSUB = 333,
	_OP_ASNMUL = 334,
	_OP_ASNDIV = 335,
	_OP_ASNMOD = 336,
	_OP_ASNAND = 337,
	_OP_ASNOR = 338,
	_OP_ASNXOR = 339,
	_OP_ASNLSHIFT = 340,
	_OP_ASNRSHIFT = 341,
	_OP_DOT = 342,
	_OP_PTRTO = 343,
	_OP_SCOPE = 344,
	_ELIPSIS = 345,
	_OP_MDOT = 346,
	_OP_MPTRTO = 347,
	_OP_CONDASN = 348,
	_KW_ATTRIBUTE = 349,
	_KW_CONST = 350,
	_KW_VOLATILE = 351,
	_KW_AUTO = 352,
	_KW_STATIC = 353,
	_KW_EXTERN = 354,
	_KW_REGISTER = 355,
	_KW_LONG = 356,
	_KW_SHORT = 357,
	_KW_SIGNED = 358,
	_KW_UNSIGNED = 359,
	_KW_INT = 360,
	_KW_CHAR = 361,
	_KW_FLOAT = 362,
	_KW_DOUBLE = 363,
	_KW_VOID = 364,
	_KW_IF = 365,
	_KW_ELSE = 366,
	_KW_DO = 367,
	_KW_WHILE = 368,
	_KW_FOR = 369,
	_KW_SWITCH = 370,
	_KW_CASE = 371,
	_KW_DEFAULT = 372,
	_KW_BREAK = 373,
	_KW_CONTINUE = 374,
	_KW_GOTO = 375,
	_KW_RETURN = 376,
	_KW_ENUM = 377,
	_KW_UNION = 378,
	_KW_STRUCT = 379,
	_KW_TYPEDEF = 380,
	_OP_SIZEOF = 381,
	_KW__FUNC = 382,
	_KW__PFUNC = 383,
	_KW_OPERATOR = 384,
	_KW_CLASS = 385,
	_KW_THIS = 386,
	_KW_TYPEID = 387,
	_KW_TYPENAME = 388,
	_KW_NAMESPACE = 389,
	_KW_USING = 390,
	_KW_TEMPLATE = 391,
	_KW_PUBLIC = 392,
	_KW_PRIVATE = 393,
	_KW_PROTECTED = 394,
	_KW_FRIEND = 395,
	_KW_VIRTUAL = 396,
	_KW_MUTABLE = 397,
	_KW_EXPLICIT = 398,
	_KW_BOOL = 399,
	_KW_TRUE = 400,
	_KW_FALSE = 401,
	_KW_WCHAR = 402,
	_KW_THROW = 403,
	_KW_CATCH = 404,
	_KW_TRY = 405,
	_KW_SCAST = 406,
	_KW_DCAST = 407,
	_KW_CCAST = 408,
	_KW_RCAST = 409,
	_KW_INLINE = 410,
	_KW_ASM = 411,
	_OP_TYPEOF = 412,
	_OP_NEW = 413,
	_OP_DELETE = 414,
	_OP_NEWA = 415,
	_OP_DELETEA = 416,
	_KW_DYNAMIC = 417,
	_VAL_ID = 418,
	_AGGR_ID = 419,
	_ENUM_ID = 420,
	_TYPE_ID = 421,
	_CTOR_ID = 422,
	_DTOR_ID = 423,
	_OPER_ID = 424,
	_CONV_ID = 425,
	_NSPACE_ID = 426,
	_ASPACE_ID = 427,
	_TPLATE_ID = 428,
	_SCOPE = 429,
	_TOKENLIST = 430,
	_OP_REF = 431,
	_OP_PTR = 432,
	_OP_MPTR = 433,
	_OP_UMINUS = 434,
	_OP_UPLUS = 435,
	_OP_INCP = 436,
	_OP_DECP = 437,
	_OP_TSIZEOF = 438,
	_OP_COMMA = 439,
	_OP_FUNC = 440,
	_OP_INDEX = 441,
	_OP_CAST = 442,
	_KW_DCLASS = 443,
	_KW_DTYPE = 444,
	_NEXT_INSN = 445,
	_BACKPATCH = 446,
	_ACTION = 447,
	_TBC = 448,
	_LBRACE_SAVE = 449,
	_COLON_SAVE = 450,
	_PARSE_COMPOUND = 451,
	_DEFINE = 452,
	YYLASTTOKEN = 453,
#line 110 "/u4/shaggy/u2/src/sos/util/shacc/shacc.c.parser"
};

/* future: Maybe YYRedNode can be a union'ed case of YYNode?  use x = -1 
	to signal it's a reduction (for GC purposes) */

typedef struct YYNode YYNode;
struct YYNode {
	const YYNode *a; /* active/reduction link */
	const YYNode *s; /* substitution link */
	const YYNode *p; /* previous link */
	YYRTYPE r;	/* rule # */
	yys16 c;	/* reduction chain */
	yys16 c1;	/* 1st reduction */
	yyu8 x;		/* offset into rule */
	yyu8 rp;	/* reduction precedence */

	/* for GC */
	yyu8 rc;	/* reference count */
	yyu8 mark;
	YYNode *np;
	YYNode *pp;
};


typedef struct YYRedNode YYRedNode;
struct YYRedNode {
	YYRTYPE r;	/* rule # */
	yyu16 sp;	/* yydvsa stack pointer */
	yys16 p;	/* previous reduction */
#if YYDEBUG
	YYTTYPE	nt;	/* non-terminal to reduce to */
#endif
};


typedef struct YY YY;
struct YY {
	YYRedNode *rstk;
	YYSTYPE *vsa;		/* semantic value stack */
	YYSTYPE *dvsa;		/* deferred semantic value stack */
	int rsp;
	YYSTYPE *vsp;
	int dsp;		/* deferred stack pointer */
	int dsrp;		/* deferred stack read pointer */
	const YYNode *anp;	/* list of active states (non reduceable) */
	const YYNode *rnp;	/* list of reduceable active states */
	YYNode *free;		/* free list for node allocator */
	void *alloc;		/* allocated memory list */
	// char errstr[80];	/* error string */
	const YYNode *nts[YYNNONTERMS];	/* non-terminal substitutions */
};



typedef struct YYRule YYRule;
struct YYRule {
	yyu16	sym;		/* first symbol */
	yyu8	ns;		/* number of symbols */
	yyu8	skip;		/* number of rules to skip for next */
	yyu8	srp;		/* shift/reduce rule precedence */
	yyu8	rrp:6;		/* reduce/reduce rule precedence */
	yyu8	uniq:1;		/* reduction must be unique */
	yyu8	hasact:1;	/* has action */
};


enum {
	YYTOKEOF=0, YYTOKERROR=1, YYTOKUNDEF=2,
};


static YYRule yyrule[] = {
	/* 0 */ {	0,	2,	0,	0,	0,	0,0 },
	/* 1 */ {	2,	1,	1,	0,	0,	0,0 },
	/* 2 */ {	3,	2,	0,	0,	0,	0,1 },
	/* 3 */ {	5,	3,	0,	0,	0,	0,1 },
	/* 4 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 5 */ {	8,	0,	1,	0,	0,	0,1 },
	/* 6 */ {	8,	1,	0,	0,	0,	0,0 },
	/* 7 */ {	9,	0,	0,	0,	0,	0,1 },
	/* 8 */ {	9,	1,	1,	0,	0,	0,0 },
	/* 9 */ {	10,	3,	1,	0,	0,	0,0 },
	/* 10 */ {	13,	1,	1,	0,	0,	0,0 },
	/* 11 */ {	14,	1,	1,	0,	0,	0,0 },
	/* 12 */ {	15,	6,	0,	0,	0,	0,1 },
	/* 13 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 14 */ {	21,	1,	1,	0,	0,	0,0 },
	/* 15 */ {	22,	1,	0,	0,	0,	0,0 },
	/* 16 */ {	23,	1,	1,	0,	0,	0,0 },
	/* 17 */ {	24,	1,	0,	0,	0,	0,0 },
	/* 18 */ {	25,	3,	1,	0,	1,	1,0 },
	/* 19 */ {	28,	2,	0,	0,	2,	1,1 },
	/* 20 */ {	30,	1,	1,	0,	0,	0,0 },
	/* 21 */ {	31,	1,	1,	0,	0,	0,0 },
	/* 22 */ {	32,	1,	1,	0,	0,	0,0 },
	/* 23 */ {	33,	1,	1,	0,	0,	0,0 },
	/* 24 */ {	34,	1,	0,	0,	0,	0,0 },
	/* 25 */ {	35,	1,	1,	0,	0,	0,1 },
	/* 26 */ {	36,	2,	0,	0,	0,	0,1 },
	/* 27 */ {	38,	1,	1,	0,	0,	0,0 },
	/* 28 */ {	39,	1,	1,	0,	0,	0,1 },
	/* 29 */ {	40,	2,	1,	0,	0,	0,1 },
	/* 30 */ {	42,	2,	1,	0,	0,	0,1 },
	/* 31 */ {	44,	3,	0,	0,	0,	0,1 },
	/* 32 */ {	47,	0,	1,	0,	0,	0,1 },
	/* 33 */ {	47,	1,	0,	0,	0,	0,0 },
	/* 34 */ {	48,	1,	1,	0,	0,	0,0 },
	/* 35 */ {	49,	1,	1,	0,	0,	0,0 },
	/* 36 */ {	50,	1,	1,	0,	0,	0,0 },
	/* 37 */ {	51,	1,	1,	0,	0,	0,0 },
	/* 38 */ {	52,	1,	0,	0,	0,	0,0 },
	/* 39 */ {	53,	1,	1,	0,	0,	0,0 },
	/* 40 */ {	54,	1,	1,	0,	0,	0,0 },
	/* 41 */ {	55,	1,	0,	0,	0,	0,0 },
	/* 42 */ {	56,	1,	1,	0,	0,	0,0 },
	/* 43 */ {	57,	1,	1,	0,	0,	0,0 },
	/* 44 */ {	58,	1,	1,	0,	0,	0,0 },
	/* 45 */ {	59,	1,	1,	0,	0,	0,0 },
	/* 46 */ {	60,	1,	0,	0,	0,	0,0 },
	/* 47 */ {	61,	1,	1,	0,	0,	0,0 },
	/* 48 */ {	62,	1,	1,	0,	0,	0,0 },
	/* 49 */ {	63,	1,	1,	0,	0,	0,0 },
	/* 50 */ {	64,	1,	1,	0,	0,	0,0 },
	/* 51 */ {	65,	1,	1,	0,	0,	0,0 },
	/* 52 */ {	66,	1,	1,	0,	0,	0,0 },
	/* 53 */ {	67,	1,	1,	0,	0,	0,0 },
	/* 54 */ {	68,	1,	1,	0,	0,	0,0 },
	/* 55 */ {	69,	1,	1,	0,	0,	0,0 },
	/* 56 */ {	70,	1,	1,	0,	0,	0,0 },
	/* 57 */ {	71,	1,	0,	0,	0,	0,0 },
	/* 58 */ {	72,	1,	1,	0,	0,	0,0 },
	/* 59 */ {	73,	1,	1,	0,	0,	0,0 },
	/* 60 */ {	74,	1,	0,	0,	0,	0,0 },
	/* 61 */ {	75,	2,	1,	0,	0,	0,1 },
	/* 62 */ {	77,	2,	0,	0,	0,	0,1 },
	/* 63 */ {	79,	7,	0,	0,	0,	1,1 },
	/* 64 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 65 */ {	86,	1,	1,	0,	0,	0,0 },
	/* 66 */ {	87,	3,	0,	0,	0,	0,0 },
	/* 67 */ {	90,	0,	1,	0,	0,	0,1 },
	/* 68 */ {	90,	1,	0,	0,	0,	0,0 },
	/* 69 */ {	91,	1,	1,	0,	0,	0,1 },
	/* 70 */ {	92,	3,	0,	0,	0,	0,1 },
	/* 71 */ {	95,	1,	1,	0,	0,	0,0 },
	/* 72 */ {	96,	1,	0,	0,	0,	0,0 },
	/* 73 */ {	97,	1,	1,	0,	0,	0,0 },
	/* 74 */ {	98,	1,	0,	0,	0,	0,0 },
	/* 75 */ {	99,	6,	0,	0,	0,	1,1 },
	/* 76 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 77 */ {	105,	6,	0,	0,	0,	1,1 },
	/* 78 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 79 */ {	111,	5,	0,	0,	0,	1,1 },
	/* 80 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 81 */ {	116,	1,	0,	0,	0,	0,0 },
	/* 82 */ {	117,	5,	0,	0,	0,	1,1 },
	/* 83 */ {	122,	4,	1,	0,	0,	1,1 },
	/* 84 */ {	126,	3,	1,	0,	0,	1,1 },
	/* 85 */ {	129,	4,	0,	0,	0,	1,1 },
	/* 86 */ {	133,	4,	0,	0,	0,	1,1 },
	/* 87 */ {	137,	5,	0,	0,	0,	1,1 },
	/* 88 */ {	142,	5,	1,	0,	0,	0,0 },
	/* 89 */ {	147,	3,	0,	0,	0,	0,0 },
	/* 90 */ {	150,	0,	0,	0,	0,	0,1 },
	/* 91 */ {	150,	4,	1,	0,	0,	0,1 },
	/* 92 */ {	154,	4,	0,	0,	0,	0,1 },
	/* 93 */ {	158,	2,	1,	0,	0,	0,1 },
	/* 94 */ {	160,	2,	0,	0,	0,	0,1 },
	/* 95 */ {	162,	0,	1,	0,	0,	0,1 },
	/* 96 */ {	162,	1,	0,	0,	0,	0,0 },
	/* 97 */ {	163,	2,	1,	0,	0,	0,1 },
	/* 98 */ {	165,	4,	0,	0,	0,	0,1 },
	/* 99 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 100 */ {	169,	1,	1,	0,	0,	0,0 },
	/* 101 */ {	170,	2,	0,	0,	0,	0,1 },
	/* 102 */ {	172,	1,	1,	0,	0,	0,0 },
	/* 103 */ {	173,	5,	1,	0,	0,	0,1 },
	/* 104 */ {	178,	4,	1,	0,	0,	0,1 },
	/* 105 */ {	182,	3,	1,	0,	0,	0,1 },
	/* 106 */ {	185,	3,	0,	0,	0,	0,1 },
	/* 107 */ {	188,	1,	0,	0,	0,	0,1 },
	/* 108 */ {	189,	5,	0,	0,	0,	0,1 },
	/* 109 */ {	194,	1,	1,	0,	1,	0,0 },
	/* 110 */ {	195,	1,	0,	0,	2,	0,0 },
	/* 111 */ {	196,	2,	1,	0,	0,	0,1 },
	/* 112 */ {	198,	1,	1,	0,	0,	0,1 },
	/* 113 */ {	199,	3,	0,	0,	0,	0,1 },
	/* 114 */ {	202,	2,	0,	0,	0,	0,1 },
	/* 115 */ {	204,	0,	1,	0,	0,	0,1 },
	/* 116 */ {	204,	1,	0,	0,	0,	0,0 },
	/* 117 */ {	205,	1,	1,	0,	0,	0,0 },
	/* 118 */ {	206,	1,	0,	0,	0,	0,0 },
	/* 119 */ {	207,	1,	0,	0,	0,	0,1 },
	/* 120 */ {	208,	2,	0,	0,	0,	0,1 },
	/* 121 */ {	210,	1,	1,	0,	0,	0,1 },
	/* 122 */ {	211,	2,	0,	0,	0,	0,1 },
	/* 123 */ {	213,	1,	1,	0,	0,	0,1 },
	/* 124 */ {	214,	1,	0,	0,	0,	0,0 },
	/* 125 */ {	215,	2,	1,	0,	0,	0,1 },
	/* 126 */ {	217,	1,	0,	0,	0,	0,0 },
	/* 127 */ {	218,	0,	1,	0,	0,	0,1 },
	/* 128 */ {	218,	1,	0,	0,	0,	0,0 },
	/* 129 */ {	219,	4,	1,	0,	0,	0,1 },
	/* 130 */ {	223,	5,	1,	0,	0,	0,1 },
	/* 131 */ {	228,	3,	0,	0,	0,	0,1 },
	/* 132 */ {	231,	0,	1,	0,	0,	0,1 },
	/* 133 */ {	231,	1,	1,	0,	0,	0,0 },
	/* 134 */ {	232,	3,	1,	0,	0,	0,1 },
	/* 135 */ {	235,	1,	0,	0,	0,	0,1 },
	/* 136 */ {	236,	1,	1,	0,	0,	0,1 },
	/* 137 */ {	237,	3,	0,	0,	0,	0,1 },
	/* 138 */ {	240,	4,	1,	0,	1,	0,1 },
	/* 139 */ {	244,	2,	1,	0,	1,	0,1 },
	/* 140 */ {	246,	4,	1,	0,	2,	0,1 },
	/* 141 */ {	250,	2,	0,	0,	2,	0,1 },
	/* 142 */ {	252,	1,	0,	0,	0,	0,1 },
	/* 143 */ {	253,	6,	0,	0,	0,	0,1 },
	/* 144 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 145 */ {	259,	2,	2,	0,	0,	0,1 },
	/* 146 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 147 */ {	261,	1,	0,	0,	0,	0,0 },
	/* 148 */ {	262,	2,	1,	0,	0,	0,1 },
	/* 149 */ {	264,	3,	0,	0,	0,	0,1 },
	/* 150 */ {	267,	1,	1,	0,	0,	0,0 },
	/* 151 */ {	268,	4,	1,	0,	0,	1,1 },
	/* 152 */ {	272,	2,	0,	0,	0,	1,1 },
	/* 153 */ {	274,	0,	1,	0,	0,	0,0 },
	/* 154 */ {	274,	1,	0,	0,	0,	0,0 },
	/* 155 */ {	275,	1,	1,	0,	0,	0,1 },
	/* 156 */ {	276,	3,	0,	0,	0,	0,1 },
	/* 157 */ {	279,	7,	0,	0,	0,	1,1 },
	/* 158 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 159 */ {	286,	1,	1,	0,	0,	0,0 },
	/* 160 */ {	287,	1,	1,	0,	0,	0,0 },
	/* 161 */ {	288,	1,	1,	0,	0,	0,0 },
	/* 162 */ {	289,	2,	1,	0,	0,	0,1 },
	/* 163 */ {	291,	3,	1,	0,	0,	0,1 },
	/* 164 */ {	294,	1,	1,	0,	0,	0,0 },
	/* 165 */ {	295,	1,	0,	0,	0,	0,0 },
	/* 166 */ {	296,	2,	1,	0,	0,	0,0 },
	/* 167 */ {	298,	4,	2,	0,	0,	0,0 },
	/* 168 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 169 */ {	302,	4,	2,	0,	0,	0,0 },
	/* 170 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 171 */ {	306,	5,	0,	0,	0,	0,0 },
	/* 172 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 173 */ {	311,	0,	1,	0,	0,	0,1 },
	/* 174 */ {	311,	1,	0,	0,	0,	0,0 },
	/* 175 */ {	312,	1,	1,	0,	0,	1,0 },
	/* 176 */ {	313,	2,	1,	0,	0,	1,1 },
	/* 177 */ {	315,	3,	1,	0,	0,	1,1 },
	/* 178 */ {	318,	2,	1,	0,	0,	1,1 },
	/* 179 */ {	320,	1,	0,	0,	0,	0,1 },
	/* 180 */ {	321,	1,	1,	0,	0,	0,0 },
	/* 181 */ {	322,	3,	0,	0,	0,	0,1 },
	/* 182 */ {	325,	1,	1,	0,	0,	0,0 },
	/* 183 */ {	326,	3,	1,	0,	0,	0,1 },
	/* 184 */ {	329,	2,	1,	0,	0,	0,1 },
	/* 185 */ {	331,	3,	0,	0,	0,	0,1 },
	/* 186 */ {	334,	0,	1,	0,	0,	0,1 },
	/* 187 */ {	334,	2,	0,	0,	0,	0,1 },
	/* 188 */ {	336,	1,	1,	0,	0,	0,0 },
	/* 189 */ {	337,	3,	0,	0,	0,	0,1 },
	/* 190 */ {	340,	1,	1,	0,	0,	0,1 },
	/* 191 */ {	341,	2,	1,	0,	0,	0,1 },
	/* 192 */ {	343,	3,	1,	0,	0,	0,1 },
	/* 193 */ {	346,	2,	1,	0,	0,	0,1 },
	/* 194 */ {	348,	3,	0,	0,	0,	0,1 },
	/* 195 */ {	351,	1,	1,	0,	0,	0,0 },
	/* 196 */ {	352,	1,	1,	0,	0,	0,0 },
	/* 197 */ {	353,	1,	0,	0,	0,	0,0 },
	/* 198 */ {	354,	2,	0,	0,	0,	0,1 },
	/* 199 */ {	356,	2,	0,	0,	0,	0,1 },
	/* 200 */ {	358,	2,	0,	0,	0,	0,1 },
	/* 201 */ {	360,	0,	1,	0,	0,	0,1 },
	/* 202 */ {	360,	1,	0,	0,	0,	0,0 },
	/* 203 */ {	361,	3,	2,	0,	0,	0,1 },
	/* 204 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 205 */ {	364,	3,	0,	0,	0,	0,1 },
	/* 206 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 207 */ {	367,	0,	1,	0,	0,	0,1 },
	/* 208 */ {	367,	1,	0,	0,	0,	0,0 },
	/* 209 */ {	368,	1,	1,	0,	0,	0,0 },
	/* 210 */ {	369,	3,	0,	0,	0,	0,1 },
	/* 211 */ {	372,	1,	1,	0,	0,	0,1 },
	/* 212 */ {	373,	3,	0,	0,	0,	0,1 },
	/* 213 */ {	376,	4,	0,	0,	0,	0,1 },
	/* 214 */ {	380,	4,	0,	0,	0,	0,1 },
	/* 215 */ {	384,	2,	0,	0,	0,	0,1 },
	/* 216 */ {	386,	1,	1,	0,	0,	0,0 },
	/* 217 */ {	387,	1,	1,	0,	0,	0,0 },
	/* 218 */ {	388,	1,	1,	0,	0,	0,0 },
	/* 219 */ {	389,	1,	1,	0,	0,	0,0 },
	/* 220 */ {	390,	1,	1,	0,	0,	0,0 },
	/* 221 */ {	391,	1,	1,	0,	0,	0,0 },
	/* 222 */ {	392,	1,	1,	0,	0,	0,0 },
	/* 223 */ {	393,	1,	1,	0,	0,	0,0 },
	/* 224 */ {	394,	1,	1,	0,	0,	0,0 },
	/* 225 */ {	395,	1,	1,	0,	0,	0,0 },
	/* 226 */ {	396,	1,	1,	0,	0,	0,0 },
	/* 227 */ {	397,	1,	1,	0,	0,	0,0 },
	/* 228 */ {	398,	1,	1,	0,	0,	0,0 },
	/* 229 */ {	399,	1,	1,	0,	0,	0,0 },
	/* 230 */ {	400,	1,	1,	0,	0,	0,0 },
	/* 231 */ {	401,	1,	1,	0,	0,	0,0 },
	/* 232 */ {	402,	1,	1,	0,	0,	0,0 },
	/* 233 */ {	403,	1,	1,	0,	0,	0,0 },
	/* 234 */ {	404,	1,	1,	0,	0,	0,0 },
	/* 235 */ {	405,	1,	1,	0,	0,	0,0 },
	/* 236 */ {	406,	1,	1,	0,	0,	0,0 },
	/* 237 */ {	407,	1,	1,	0,	0,	0,0 },
	/* 238 */ {	408,	1,	1,	0,	0,	0,0 },
	/* 239 */ {	409,	1,	1,	0,	0,	0,0 },
	/* 240 */ {	410,	1,	1,	0,	0,	0,0 },
	/* 241 */ {	411,	1,	1,	0,	0,	0,0 },
	/* 242 */ {	412,	1,	1,	0,	0,	0,0 },
	/* 243 */ {	413,	1,	1,	0,	0,	0,0 },
	/* 244 */ {	414,	1,	1,	0,	0,	0,0 },
	/* 245 */ {	415,	1,	1,	0,	0,	0,0 },
	/* 246 */ {	416,	1,	1,	0,	0,	0,0 },
	/* 247 */ {	417,	1,	1,	0,	0,	0,0 },
	/* 248 */ {	418,	1,	1,	0,	0,	0,0 },
	/* 249 */ {	419,	1,	1,	0,	0,	0,0 },
	/* 250 */ {	420,	1,	1,	0,	0,	0,0 },
	/* 251 */ {	421,	1,	1,	0,	0,	0,0 },
	/* 252 */ {	422,	1,	1,	0,	0,	0,0 },
	/* 253 */ {	423,	1,	1,	0,	0,	0,0 },
	/* 254 */ {	424,	1,	1,	0,	0,	0,0 },
	/* 255 */ {	425,	1,	0,	0,	0,	0,0 },
	/* 256 */ {	426,	1,	1,	0,	0,	0,0 },
	/* 257 */ {	427,	2,	1,	0,	0,	0,1 },
	/* 258 */ {	429,	2,	0,	0,	0,	0,1 },
	/* 259 */ {	431,	1,	1,	0,	0,	0,0 },
	/* 260 */ {	432,	1,	1,	0,	0,	0,0 },
	/* 261 */ {	433,	1,	1,	0,	0,	0,0 },
	/* 262 */ {	434,	1,	1,	0,	0,	0,0 },
	/* 263 */ {	435,	1,	0,	0,	0,	0,0 },
	/* 264 */ {	436,	1,	1,	0,	0,	0,1 },
	/* 265 */ {	437,	1,	1,	0,	0,	0,1 },
	/* 266 */ {	438,	3,	1,	0,	0,	0,1 },
	/* 267 */ {	441,	1,	0,	0,	0,	0,0 },
	/* 268 */ {	442,	1,	0,	0,	0,	0,0 },
	/* 269 */ {	443,	1,	0,	0,	0,	0,0 },
	/* 270 */ {	444,	1,	1,	0,	0,	0,0 },
	/* 271 */ {	445,	1,	1,	0,	0,	0,0 },
	/* 272 */ {	446,	1,	1,	0,	0,	0,0 },
	/* 273 */ {	447,	1,	0,	0,	0,	0,1 },
	/* 274 */ {	448,	1,	1,	0,	0,	0,0 },
	/* 275 */ {	449,	1,	1,	0,	0,	0,1 },
	/* 276 */ {	450,	1,	1,	0,	0,	0,0 },
	/* 277 */ {	451,	1,	1,	0,	0,	0,0 },
	/* 278 */ {	452,	1,	0,	0,	0,	0,1 },
	/* 279 */ {	453,	1,	1,	0,	0,	0,0 },
	/* 280 */ {	454,	4,	1,	0,	0,	0,1 },
	/* 281 */ {	458,	4,	1,	0,	0,	0,1 },
	/* 282 */ {	462,	4,	1,	0,	0,	0,1 },
	/* 283 */ {	466,	4,	1,	0,	0,	0,1 },
	/* 284 */ {	470,	3,	1,	0,	0,	0,1 },
	/* 285 */ {	473,	3,	1,	0,	0,	0,1 },
	/* 286 */ {	476,	3,	1,	0,	0,	0,1 },
	/* 287 */ {	479,	3,	1,	0,	0,	0,1 },
	/* 288 */ {	482,	2,	1,	0,	0,	0,1 },
	/* 289 */ {	484,	2,	1,	0,	0,	0,1 },
	/* 290 */ {	486,	7,	1,	0,	0,	0,1 },
	/* 291 */ {	493,	4,	1,	0,	0,	0,1 },
	/* 292 */ {	497,	4,	0,	0,	0,	0,1 },
	/* 293 */ {	501,	1,	1,	0,	0,	0,0 },
	/* 294 */ {	502,	1,	1,	0,	0,	0,0 },
	/* 295 */ {	503,	1,	1,	0,	0,	0,0 },
	/* 296 */ {	504,	1,	0,	0,	0,	0,0 },
	/* 297 */ {	505,	1,	1,	0,	0,	0,1 },
	/* 298 */ {	506,	3,	0,	0,	0,	0,1 },
	/* 299 */ {	509,	0,	1,	0,	0,	0,1 },
	/* 300 */ {	509,	1,	0,	0,	0,	0,0 },
	/* 301 */ {	510,	1,	0,	0,	0,	0,1 },
	/* 302 */ {	511,	1,	1,	0,	0,	0,0 },
	/* 303 */ {	512,	2,	1,	0,	0,	0,1 },
	/* 304 */ {	514,	2,	1,	0,	0,	0,1 },
	/* 305 */ {	516,	2,	1,	0,	0,	0,1 },
	/* 306 */ {	518,	2,	1,	0,	0,	0,1 },
	/* 307 */ {	520,	4,	1,	0,	0,	0,1 },
	/* 308 */ {	524,	1,	1,	0,	0,	0,0 },
	/* 309 */ {	525,	1,	0,	0,	0,	0,0 },
	/* 310 */ {	526,	1,	1,	0,	0,	0,1 },
	/* 311 */ {	527,	1,	1,	0,	0,	0,1 },
	/* 312 */ {	528,	1,	1,	0,	0,	0,1 },
	/* 313 */ {	529,	1,	1,	0,	0,	0,1 },
	/* 314 */ {	530,	1,	1,	0,	0,	0,0 },
	/* 315 */ {	531,	1,	0,	0,	0,	0,0 },
	/* 316 */ {	532,	4,	1,	0,	0,	0,1 },
	/* 317 */ {	536,	6,	0,	0,	0,	0,1 },
	/* 318 */ {	542,	3,	0,	0,	0,	0,1 },
	/* 319 */ {	545,	0,	1,	0,	0,	0,1 },
	/* 320 */ {	545,	1,	0,	0,	0,	0,0 },
	/* 321 */ {	546,	2,	0,	0,	0,	0,1 },
	/* 322 */ {	548,	2,	1,	0,	0,	0,1 },
	/* 323 */ {	550,	1,	0,	0,	0,	0,0 },
	/* 324 */ {	551,	0,	1,	0,	0,	0,1 },
	/* 325 */ {	551,	1,	0,	0,	0,	0,0 },
	/* 326 */ {	552,	3,	1,	0,	0,	0,1 },
	/* 327 */ {	555,	4,	0,	0,	0,	0,1 },
	/* 328 */ {	559,	3,	0,	0,	0,	0,1 },
	/* 329 */ {	562,	0,	1,	0,	0,	0,1 },
	/* 330 */ {	562,	1,	0,	0,	0,	0,0 },
	/* 331 */ {	563,	2,	1,	0,	0,	0,1 },
	/* 332 */ {	565,	2,	0,	0,	0,	0,1 },
	/* 333 */ {	567,	1,	1,	0,	0,	0,0 },
	/* 334 */ {	568,	4,	0,	0,	0,	0,1 },
	/* 335 */ {	572,	1,	1,	0,	0,	0,0 },
	/* 336 */ {	573,	3,	1,	0,	0,	0,1 },
	/* 337 */ {	576,	3,	0,	0,	0,	0,1 },
	/* 338 */ {	579,	1,	1,	0,	0,	0,0 },
	/* 339 */ {	580,	3,	1,	0,	0,	0,1 },
	/* 340 */ {	583,	3,	1,	0,	0,	0,1 },
	/* 341 */ {	586,	3,	0,	0,	0,	0,1 },
	/* 342 */ {	589,	1,	1,	0,	0,	0,0 },
	/* 343 */ {	590,	3,	1,	0,	0,	0,1 },
	/* 344 */ {	593,	3,	0,	0,	0,	0,1 },
	/* 345 */ {	596,	1,	1,	0,	0,	0,0 },
	/* 346 */ {	597,	3,	1,	0,	0,	0,1 },
	/* 347 */ {	600,	3,	0,	0,	0,	0,1 },
	/* 348 */ {	603,	1,	1,	0,	0,	0,0 },
	/* 349 */ {	604,	3,	1,	0,	0,	0,1 },
	/* 350 */ {	607,	3,	1,	0,	0,	0,1 },
	/* 351 */ {	610,	3,	1,	0,	0,	0,1 },
	/* 352 */ {	613,	3,	0,	0,	0,	0,1 },
	/* 353 */ {	616,	1,	1,	0,	0,	0,0 },
	/* 354 */ {	617,	3,	1,	0,	0,	0,1 },
	/* 355 */ {	620,	3,	0,	0,	0,	0,1 },
	/* 356 */ {	623,	1,	1,	0,	0,	0,0 },
	/* 357 */ {	624,	3,	0,	0,	0,	0,1 },
	/* 358 */ {	627,	1,	1,	0,	0,	0,0 },
	/* 359 */ {	628,	3,	0,	0,	0,	0,1 },
	/* 360 */ {	631,	1,	1,	0,	0,	0,0 },
	/* 361 */ {	632,	3,	0,	0,	0,	0,1 },
	/* 362 */ {	635,	1,	1,	0,	0,	0,0 },
	/* 363 */ {	636,	3,	0,	0,	0,	0,1 },
	/* 364 */ {	639,	1,	1,	0,	0,	0,0 },
	/* 365 */ {	640,	3,	0,	0,	0,	0,1 },
	/* 366 */ {	643,	1,	1,	0,	0,	0,0 },
	/* 367 */ {	644,	5,	0,	0,	0,	0,1 },
	/* 368 */ {	649,	1,	1,	0,	0,	0,0 },
	/* 369 */ {	650,	3,	0,	0,	0,	0,1 },
	/* 370 */ {	653,	1,	1,	0,	0,	0,0 },
	/* 371 */ {	654,	1,	1,	0,	0,	0,0 },
	/* 372 */ {	655,	1,	1,	0,	0,	0,0 },
	/* 373 */ {	656,	1,	1,	0,	0,	0,0 },
	/* 374 */ {	657,	1,	1,	0,	0,	0,0 },
	/* 375 */ {	658,	1,	1,	0,	0,	0,0 },
	/* 376 */ {	659,	1,	1,	0,	0,	0,0 },
	/* 377 */ {	660,	1,	1,	0,	0,	0,0 },
	/* 378 */ {	661,	1,	1,	0,	0,	0,0 },
	/* 379 */ {	662,	1,	1,	0,	0,	0,0 },
	/* 380 */ {	663,	1,	0,	0,	0,	0,0 },
	/* 381 */ {	664,	1,	1,	0,	0,	0,0 },
	/* 382 */ {	665,	3,	0,	0,	0,	0,1 },
	/* 383 */ {	668,	0,	1,	0,	0,	0,1 },
	/* 384 */ {	668,	1,	0,	0,	0,	0,0 },
	/* 385 */ {	669,	1,	0,	0,	0,	0,1 },
	/* 386 */ {	670,	1,	1,	0,	0,	0,0 },
	/* 387 */ {	671,	1,	0,	0,	0,	0,0 },
	/* 388 */ {	672,	1,	1,	0,	0,	0,0 },
	/* 389 */ {	673,	1,	1,	0,	0,	0,0 },
	/* 390 */ {	674,	1,	1,	0,	0,	0,0 },
	/* 391 */ {	675,	1,	1,	0,	0,	0,0 },
	/* 392 */ {	676,	1,	1,	0,	0,	0,0 },
	/* 393 */ {	677,	1,	1,	0,	0,	0,0 },
	/* 394 */ {	678,	1,	0,	0,	0,	0,0 },
	/* 395 */ {	679,	1,	1,	0,	0,	0,0 },
	/* 396 */ {	680,	7,	0,	0,	0,	0,1 },
	/* 397 */ {	687,	5,	1,	0,	0,	0,1 },
	/* 398 */ {	692,	7,	0,	0,	0,	0,1 },
	/* 399 */ {	699,	5,	0,	0,	0,	0,1 },
	/* 400 */ {	704,	3,	1,	0,	0,	0,1 },
	/* 401 */ {	707,	4,	1,	0,	0,	0,1 },
	/* 402 */ {	711,	3,	0,	0,	0,	0,1 },
	/* 403 */ {	714,	2,	0,	0,	0,	1,0 },
	/* 404 */ {	716,	4,	0,	0,	0,	1,1 },
	/* 405 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 406 */ {	720,	0,	1,	0,	0,	0,0 },
	/* 407 */ {	720,	2,	0,	0,	0,	0,1 },
	/* 408 */ {	722,	1,	1,	0,	0,	0,0 },
	/* 409 */ {	723,	4,	0,	0,	0,	0,1 },
	/* 410 */ {	727,	0,	1,	0,	0,	0,1 },
	/* 411 */ {	727,	1,	0,	0,	0,	0,0 },
	/* 412 */ {	728,	5,	1,	0,	0,	0,1 },
	/* 413 */ {	733,	7,	1,	0,	0,	1,1 },
	/* 414 */ {	740,	8,	0,	0,	0,	0,1 },
	/* 415 */ {	748,	1,	1,	0,	0,	0,0 },
	/* 416 */ {	749,	1,	0,	0,	0,	0,0 },
	/* 417 */ {	750,	2,	1,	0,	0,	1,1 },
	/* 418 */ {	752,	2,	1,	0,	0,	1,1 },
	/* 419 */ {	754,	3,	1,	0,	0,	1,1 },
	/* 420 */ {	757,	3,	0,	0,	0,	1,1 },
	/* 421 */ {	760,	1,	0,	0,	0,	0,1 },
	/* 422 */ {	761,	1,	0,	0,	0,	0,0 },
	/* 423 */ {	762,	1,	1,	0,	0,	0,0 },
	/* 424 */ {	763,	1,	0,	0,	0,	0,0 },
	/* 425 */ {	764,	1,	0,	0,	0,	0,0 },
	/* 426 */ {	765,	1,	0,	0,	0,	0,0 },
	/* 427 */ {	766,	1,	0,	0,	0,	0,0 },
	/* 428 */ {	767,	1,	0,	0,	0,	0,0 },
	/* 429 */ {	768,	1,	0,	0,	0,	0,0 },
	/* 430 */ {	769,	1,	1,	0,	0,	0,0 },
	/* 431 */ {	770,	1,	1,	0,	0,	0,0 },
	/* 432 */ {	771,	1,	1,	0,	0,	0,0 },
	/* 433 */ {	772,	1,	1,	0,	0,	0,0 },
	/* 434 */ {	773,	1,	1,	0,	0,	0,1 },
	/* 435 */ {	774,	1,	1,	0,	0,	0,0 },
	/* 436 */ {	775,	1,	1,	0,	0,	0,0 },
	/* 437 */ {	776,	1,	0,	0,	0,	0,0 },
	/* 438 */ {	777,	0,	1,	0,	0,	0,1 },
	/* 439 */ {	777,	1,	0,	0,	0,	0,0 },
	/* 440 */ {	778,	1,	1,	0,	0,	0,0 },
	/* 441 */ {	779,	1,	0,	0,	0,	0,1 },
	/* 442 */ {	780,	1,	1,	0,	0,	0,0 },
	/* 443 */ {	781,	1,	0,	0,	0,	0,1 },
	/* 444 */ {	782,	0,	1,	0,	0,	0,0 },
	/* 445 */ {	782,	1,	0,	0,	0,	0,0 },
	/* 446 */ {	783,	4,	0,	0,	0,	1,1 },
	/* 447 */ {	0,	0,	0,	0,	0,	0,1 },
	/* 448 */ {	787,	2,	1,	0,	0,	0,1 },
	/* 449 */ {	789,	2,	0,	0,	0,	0,1 },
	/* 450 */ {	791,	0,	1,	0,	0,	0,1 },
	/* 451 */ {	791,	1,	0,	0,	0,	0,0 },
	/* 452 */ {	792,	3,	1,	0,	0,	1,1 },
	/* 453 */ {	795,	3,	1,	0,	0,	0,1 },
	/* 454 */ {	798,	3,	0,	0,	0,	0,1 },
	/* 455 */ {	801,	1,	1,	0,	0,	0,0 },
	/* 456 */ {	802,	1,	1,	0,	0,	0,0 },
	/* 457 */ {	803,	1,	1,	0,	0,	0,0 },
	/* 458 */ {	804,	1,	1,	0,	0,	0,0 },
	/* 459 */ {	805,	1,	1,	0,	0,	0,0 },
	/* 460 */ {	806,	1,	1,	0,	0,	0,0 },
	/* 461 */ {	807,	1,	1,	0,	0,	0,0 },
	/* 462 */ {	808,	1,	1,	0,	0,	0,0 },
	/* 463 */ {	809,	1,	1,	0,	0,	0,0 },
	/* 464 */ {	810,	1,	1,	0,	0,	0,0 },
	/* 465 */ {	811,	1,	1,	0,	0,	0,0 },
	/* 466 */ {	812,	1,	1,	0,	0,	0,0 },
	/* 467 */ {	813,	1,	1,	0,	0,	0,0 },
	/* 468 */ {	814,	1,	1,	0,	0,	0,0 },
	/* 469 */ {	815,	1,	1,	0,	0,	0,0 },
	/* 470 */ {	816,	1,	1,	0,	0,	0,0 },
	/* 471 */ {	817,	1,	1,	0,	0,	0,0 },
	/* 472 */ {	818,	1,	1,	0,	0,	0,0 },
	/* 473 */ {	819,	1,	1,	0,	0,	0,0 },
	/* 474 */ {	820,	1,	1,	0,	0,	0,0 },
	/* 475 */ {	821,	1,	1,	0,	0,	0,0 },
	/* 476 */ {	822,	1,	1,	0,	0,	0,0 },
	/* 477 */ {	823,	1,	1,	0,	0,	0,0 },
	/* 478 */ {	824,	1,	1,	0,	0,	0,0 },
	/* 479 */ {	825,	1,	1,	0,	0,	0,0 },
	/* 480 */ {	826,	1,	1,	0,	0,	0,0 },
	/* 481 */ {	827,	1,	1,	0,	0,	0,0 },
	/* 482 */ {	828,	1,	1,	0,	0,	0,0 },
	/* 483 */ {	829,	1,	1,	0,	0,	0,0 },
	/* 484 */ {	830,	1,	1,	0,	0,	0,0 },
	/* 485 */ {	831,	1,	1,	0,	0,	0,0 },
	/* 486 */ {	832,	1,	1,	0,	0,	0,0 },
	/* 487 */ {	833,	1,	1,	0,	0,	0,0 },
	/* 488 */ {	834,	1,	1,	0,	0,	0,0 },
	/* 489 */ {	835,	1,	1,	0,	0,	0,0 },
	/* 490 */ {	836,	1,	1,	0,	0,	0,0 },
	/* 491 */ {	837,	1,	1,	0,	0,	0,0 },
	/* 492 */ {	838,	1,	1,	0,	0,	0,0 },
	/* 493 */ {	839,	1,	1,	0,	0,	0,0 },
	/* 494 */ {	840,	1,	1,	0,	0,	0,0 },
	/* 495 */ {	841,	1,	1,	0,	0,	0,0 },
	/* 496 */ {	842,	1,	1,	0,	0,	0,0 },
	/* 497 */ {	843,	1,	1,	0,	0,	0,0 },
	/* 498 */ {	844,	1,	1,	0,	0,	0,0 },
	/* 499 */ {	845,	1,	1,	0,	0,	0,0 },
	/* 500 */ {	846,	1,	1,	0,	0,	0,0 },
	/* 501 */ {	847,	1,	1,	0,	0,	0,0 },
	/* 502 */ {	848,	1,	1,	0,	0,	0,0 },
	/* 503 */ {	849,	1,	1,	0,	0,	0,1 },
	/* 504 */ {	850,	1,	0,	0,	0,	0,0 },
	/* 505 */ {	851,	6,	0,	0,	0,	0,1 },
	/* 506 */ {	857,	0,	1,	0,	0,	0,1 },
	/* 507 */ {	857,	1,	0,	0,	0,	0,0 },
#line 183 "/u4/shaggy/u2/src/sos/util/shacc/shacc.c.parser"
};

static YYTTYPE yysym[] = {
	/* 0 */ 	-1,	0,	-2,	451,	-3,	-2,
	/* 6 */ 	-5,	-6,	-4,	-8,	-7,	-9,
	/* 12 */ 	293,	-10,	-11,	452,	-12,	-13,
	/* 18 */ 	-14,	-15,	-16,	-17,	422,	-18,
	/* 24 */ 	-19,	-20,	-9,	293,	-20,	293,
	/* 30 */ 	-22,	-23,	-24,	395,	380,	-21,
	/* 36 */ 	-25,	-21,	-25,	-26,	-25,	-26,
	/* 42 */ 	-26,	-25,	-25,	-26,	-25,	-20,
	/* 48 */ 	352,	355,	353,	354,	397,	410,
	/* 54 */ 	396,	398,	-28,	-29,	-30,	-31,
	/* 60 */ 	-32,	361,	402,	399,	357,	360,
	/* 66 */ 	356,	358,	359,	362,	363,	364,
	/* 72 */ 	-33,	-34,	-35,	-36,	-17,	377,
	/* 78 */ 	-17,	377,	-37,	-38,	-39,	-40,
	/* 84 */ 	-41,	299,	-43,	-42,	295,	-43,
	/* 90 */ 	-42,	-17,	-17,	331,	-44,	-46,
	/* 96 */ 	-47,	-48,	-49,	389,	-17,	-38,
	/* 102 */ 	-50,	-51,	299,	389,	-52,	-38,
	/* 108 */ 	-53,	-51,	299,	389,	-38,	-54,
	/* 114 */ 	-51,	299,	-2,	389,	-17,	331,
	/* 120 */ 	-56,	293,	390,	388,	-58,	293,
	/* 126 */ 	390,	-58,	293,	390,	429,	-58,
	/* 132 */ 	293,	390,	389,	-60,	293,	411,
	/* 138 */ 	302,	261,	303,	293,	354,	261,
	/* 144 */ 	-38,	-2,	299,	354,	261,	-6,
	/* 150 */ 	-12,	-62,	295,	-12,	-61,	-62,
	/* 156 */ 	295,	-12,	-12,	-62,	-61,	-62,
	/* 162 */ 	-9,	-64,	-65,	-64,	-65,	-66,
	/* 168 */ 	-67,	-68,	-69,	-64,	-70,	-68,
	/* 174 */ 	302,	-71,	303,	-72,	-68,	300,
	/* 180 */ 	-44,	301,	-68,	300,	301,	302,
	/* 186 */ 	-64,	303,	422,	-73,	302,	-71,
	/* 192 */ 	303,	-72,	-74,	-64,	310,	-72,
	/* 198 */ 	314,	429,	310,	-72,	-32,	-72,
	/* 204 */ 	-76,	350,	351,	-77,	-79,	-80,
	/* 210 */ 	-23,	-81,	-23,	-26,	-81,	-69,
	/* 216 */ 	-80,	-83,	-82,	302,	-71,	303,
	/* 222 */ 	-72,	-83,	302,	-71,	303,	-72,
	/* 228 */ 	302,	-82,	303,	-84,	-84,	-41,
	/* 234 */ 	345,	345,	-85,	-84,	295,	-85,
	/* 240 */ 	-20,	-64,	331,	-86,	-20,	-64,
	/* 246 */ 	-20,	-80,	331,	-86,	-20,	-80,
	/* 252 */ 	-75,	-27,	-12,	-75,	-88,	-15,
	/* 258 */ 	-16,	-89,	-3,	-90,	331,	-91,
	/* 264 */ 	302,	-92,	303,	-86,	-38,	-93,
	/* 270 */ 	-41,	299,	-38,	299,	295,	-91,
	/* 276 */ 	-93,	295,	-91,	-36,	-37,	-94,
	/* 282 */ 	-38,	-95,	-96,	299,	385,	379,
	/* 288 */ 	378,	417,	385,	417,	396,	385,
	/* 294 */ 	443,	444,	-98,	-96,	-99,	-100,
	/* 300 */ 	-101,	-96,	-33,	-100,	-102,	-96,
	/* 306 */ 	417,	-33,	-100,	-103,	-96,	-97,
	/* 312 */ 	293,	-20,	293,	-20,	-104,	293,
	/* 318 */ 	-104,	293,	-10,	-105,	-104,	295,
	/* 324 */ 	-105,	-75,	-64,	331,	-44,	-100,
	/* 330 */ 	-44,	-17,	-100,	-44,	-100,	-106,
	/* 336 */ 	-107,	-106,	295,	-107,	-33,	396,
	/* 342 */ 	-33,	396,	-99,	-33,	-99,	-33,
	/* 348 */ 	-99,	396,	-33,	393,	394,	392,
	/* 354 */ 	384,	-109,	-79,	-110,	-69,	-110,
	/* 360 */ 	-111,	450,	-113,	-114,	294,	-115,
	/* 366 */ 	-116,	-112,	-117,	-114,	295,	-117,
	/* 372 */ 	-118,	-116,	295,	-118,	-17,	302,
	/* 378 */ 	-119,	303,	-17,	302,	-120,	303,
	/* 384 */ 	384,	-122,	413,	414,	415,	416,
	/* 390 */ 	308,	309,	310,	311,	312,	316,
	/* 396 */ 	314,	315,	313,	317,	331,	321,
	/* 402 */ 	322,	332,	333,	334,	335,	336,
	/* 408 */ 	339,	337,	338,	329,	330,	341,
	/* 414 */ 	340,	325,	326,	323,	324,	318,
	/* 420 */ 	319,	327,	328,	295,	347,	343,
	/* 426 */ 	-123,	302,	303,	300,	301,	258,
	/* 432 */ 	259,	261,	260,	263,	-124,	386,
	/* 438 */ 	302,	-126,	303,	-127,	-58,	-128,
	/* 444 */ 	-17,	-121,	-108,	423,	418,	257,
	/* 450 */ 	-121,	-108,	423,	-125,	-129,	300,
	/* 456 */ 	-126,	301,	-129,	302,	-120,	303,
	/* 462 */ 	-28,	302,	-120,	303,	-26,	302,
	/* 468 */ 	-120,	303,	-129,	342,	-77,	-129,
	/* 474 */ 	343,	-77,	-129,	342,	-130,	-129,
	/* 480 */ 	343,	-130,	-129,	327,	-129,	328,
	/* 486 */ 	-131,	321,	-78,	322,	302,	-126,
	/* 492 */ 	303,	387,	302,	-126,	303,	387,
	/* 498 */ 	302,	-78,	303,	407,	406,	409,
	/* 504 */ 	408,	-86,	-92,	295,	-86,	-92,
	/* 510 */ 	423,	-129,	327,	-133,	328,	-133,
	/* 516 */ 	-134,	-133,	381,	-132,	381,	302,
	/* 522 */ 	-78,	303,	-135,	-136,	310,	314,
	/* 528 */ 	308,	309,	317,	313,	413,	-137,
	/* 534 */ 	-138,	-139,	413,	-137,	302,	-78,
	/* 540 */ 	303,	-139,	302,	-92,	303,	-140,
	/* 546 */ 	-79,	-141,	-69,	-141,	-143,	-142,
	/* 552 */ 	300,	-126,	301,	-143,	300,	-44,
	/* 558 */ 	301,	302,	-120,	303,	-144,	414,
	/* 564 */ 	-133,	416,	-133,	-132,	302,	-78,
	/* 570 */ 	303,	-133,	-133,	-145,	346,	-133,
	/* 576 */ 	-145,	347,	-133,	-145,	-146,	310,
	/* 582 */ 	-145,	-146,	311,	-145,	-146,	312,
	/* 588 */ 	-145,	-146,	-147,	308,	-146,	-147,
	/* 594 */ 	309,	-146,	-147,	-148,	329,	-147,
	/* 600 */ 	-148,	330,	-147,	-148,	-149,	321,
	/* 606 */ 	-148,	-149,	322,	-148,	-149,	323,
	/* 612 */ 	-148,	-149,	324,	-148,	-149,	-150,
	/* 618 */ 	325,	-149,	-150,	326,	-149,	-150,
	/* 624 */ 	-151,	314,	-150,	-151,	-152,	316,
	/* 630 */ 	-151,	-152,	-153,	315,	-152,	-153,
	/* 636 */ 	-154,	318,	-153,	-154,	-155,	319,
	/* 642 */ 	-154,	-155,	-155,	296,	-126,	-100,
	/* 648 */ 	-86,	-156,	-155,	-157,	-86,	331,
	/* 654 */ 	334,	335,	336,	332,	333,	341,
	/* 660 */ 	340,	337,	339,	338,	-86,	-126,
	/* 666 */ 	295,	-86,	-126,	-156,	-160,	-161,
	/* 672 */ 	-163,	-164,	-3,	-165,	-166,	-167,
	/* 678 */ 	-168,	-162,	365,	302,	-169,	303,
	/* 684 */ 	-160,	366,	-160,	365,	302,	-169,
	/* 690 */ 	303,	-159,	365,	302,	-169,	303,
	/* 696 */ 	-160,	366,	-161,	370,	302,	-169,
	/* 702 */ 	303,	-159,	-17,	-100,	-159,	371,
	/* 708 */ 	-44,	-100,	-159,	372,	-100,	-159,
	/* 714 */ 	-158,	293,	298,	-170,	-171,	299,
	/* 720 */ 	-171,	-159,	-126,	-79,	-64,	331,
	/* 726 */ 	-86,	-169,	368,	302,	-169,	303,
	/* 732 */ 	-159,	367,	-159,	368,	302,	-126,
	/* 738 */ 	303,	293,	369,	302,	-173,	-172,
	/* 744 */ 	293,	-158,	303,	-159,	-164,	-18,
	/* 750 */ 	373,	293,	374,	293,	376,	-158,
	/* 756 */ 	293,	375,	-17,	293,	-8,	421,
	/* 762 */ 	-52,	-174,	426,	427,	419,	420,
	/* 768 */ 	428,	257,	418,	421,	420,	419,
	/* 774 */ 	426,	427,	428,	-17,	298,	449,
	/* 780 */ 	294,	450,	430,	-177,	449,	-119,
	/* 786 */ 	299,	-119,	-179,	-119,	-180,	-178,
	/* 792 */ 	-38,	-119,	299,	302,	-119,	303,
	/* 798 */ 	300,	-119,	301,	258,	259,	261,
	/* 804 */ 	260,	263,	293,	-100,	296,	-123,
	/* 810 */ 	342,	346,	344,	345,	-32,	-28,
	/* 816 */ 	-22,	365,	366,	367,	368,	369,
	/* 822 */ 	370,	371,	372,	373,	374,	375,
	/* 828 */ 	376,	377,	380,	-36,	-24,	381,
	/* 834 */ 	412,	411,	384,	386,	387,	388,
	/* 840 */ 	389,	390,	391,	-99,	395,	403,
	/* 846 */ 	404,	405,	-131,	-17,	423,	349,
	/* 852 */ 	302,	302,	-178,	303,	303,	-181,
#line 187 "/u4/shaggy/u2/src/sos/util/shacc/shacc.c.parser"
};

static YYRTYPE yynt[] = {
	/* 0 */ 	0,	1,	5,	404,	3,	4,
	/* 6 */ 	8,	7,	16,	93,	143,	88,
	/* 12 */ 	90,	14,	13,	207,	145,	430,
	/* 18 */ 	18,	87,	27,	20,	34,	42,
	/* 24 */ 	39,	25,	58,	32,	47,	157,
	/* 30 */ 	63,	61,	117,	427,	428,	422,
	/* 36 */ 	159,	438,	440,	64,	67,	153,
	/* 42 */ 	65,	69,	385,	71,	73,	79,
	/* 48 */ 	75,	77,	76,	81,	425,	78,
	/* 54 */ 	80,	82,	0,	83,	270,	86,
	/* 60 */ 	423,	91,	97,	95,	100,	506,
	/* 66 */ 	99,	148,	102,	111,	119,	132,
	/* 72 */ 	115,	107,	108,	109,	114,	268,
	/* 78 */ 	120,	123,	127,	121,	125,	129,
	/* 84 */ 	136,	138,	368,	142,	144,	146,
	/* 90 */ 	446,	150,	297,	155,	186,	158,
	/* 96 */ 	173,	166,	175,	195,	442,	168,
	/* 102 */ 	170,	172,	180,	182,	188,	190,
	/* 108 */ 	198,	199,	201,	200,	203,	204,
	/* 114 */ 	209,	206,	211,	213,	214,	450,
	/* 120 */ 	299,	215,	256,	216,	259,	264,
	/* 126 */ 	381,	269,	274,	279,	301,	293,
	/* 132 */ 	302,	333,	310,	316,	331,	319,
	/* 138 */ 	321,	329,	318,	324,	322,	326,
	/* 144 */ 	328,	335,	338,	342,	345,	348,
	/* 150 */ 	353,	356,	358,	360,	362,	364,
	/* 156 */ 	366,	370,	383,	386,	395,	397,
	/* 162 */ 	388,	400,	403,	399,	412,	417,
	/* 168 */ 	421,	408,	405,	406,	410,	415,
	/* 174 */ 	426,	429,	444,	447,	448,	455,
	/* 180 */ 	452,	505,
#line 191 "/u4/shaggy/u2/src/sos/util/shacc/shacc.c.parser"
};

#ifdef ndef
static yyu8 yytp[] = {
	/* 0 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 10 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 20 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 30 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 40 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 50 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 60 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 70 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 80 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 90 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 100 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 110 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 120 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 130 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 140 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 150 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 160 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 170 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 180 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 190 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 200 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 210 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 220 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 230 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 240 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 250 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 260 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 270 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 280 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 290 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 300 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 310 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 320 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 330 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 340 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 350 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 360 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 370 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 380 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 390 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 400 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 410 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 420 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 430 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 440 */   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	/* 450 */   0,  0,  0,
#line 196 "/u4/shaggy/u2/src/sos/util/shacc/shacc.c.parser"
};
#endif

#if YYDEBUG
static const char * yysname[] = {
	/* -181 (1) */	"attribute",
	/* -180 (1) */	"save_statement_list",
	/* -179 (1) */	"any_statement_token",
	/* -178 (2) */	"save_statement_seq",
	/* -177 (1) */	"<447>",
	/* -176 (0) */	"ignore_savecode",
	/* -175 (0) */	"template_name",
	/* -174 (0) */	"namespace_alias",
	/* -173 (1) */	"for_init_statement",
	/* -172 (1) */	"condition_opt",
	/* -171 (2) */	"statement_seq_opt",
	/* -170 (1) */	"<405>",
	/* -169 (6) */	"condition",
	/* -168 (1) */	"declaration_statement",
	/* -167 (1) */	"jump_statement",
	/* -166 (1) */	"iteration_statement",
	/* -165 (1) */	"selection_statement",
	/* -164 (2) */	"expression_statement",
	/* -163 (1) */	"labeled_statement",
	/* -162 (1) */	"other_statement",
	/* -161 (2) */	"unmatched_statement",
	/* -160 (4) */	"matched_statement",
	/* -159 (9) */	"statement",
	/* -158 (3) */	"expression_opt",
	/* -157 (1) */	"assignment_operator",
	/* -156 (2) */	"conditional_expression",
	/* -155 (4) */	"logical_or_expression",
	/* -154 (3) */	"logical_and_expression",
	/* -153 (3) */	"inclusive_or_expression",
	/* -152 (3) */	"exclusive_or_expression",
	/* -151 (3) */	"and_expression",
	/* -150 (4) */	"equality_expression",
	/* -149 (7) */	"relational_expression",
	/* -148 (7) */	"shift_expression",
	/* -147 (5) */	"additive_expression",
	/* -146 (6) */	"multiplicative_expression",
	/* -145 (6) */	"pm_expression",
	/* -144 (1) */	"new_initializer",
	/* -143 (2) */	"direct_new_declarator",
	/* -142 (1) */	"new_declarator",
	/* -141 (2) */	"new_declarator_opt",
	/* -140 (1) */	"new_placement",
	/* -139 (2) */	"new_initializer_opt",
	/* -138 (1) */	"new_type_id",
	/* -137 (2) */	"new_placement_opt",
	/* -136 (1) */	"delete_expression",
	/* -135 (1) */	"new_expression",
	/* -134 (1) */	"unary_operator",
	/* -133 (9) */	"cast_expression",
	/* -132 (2) */	"unary_expression",
	/* -131 (2) */	"newcast",
	/* -130 (2) */	"pseudo_destructor_name",
	/* -129 (9) */	"postfix_expression",
	/* -128 (1) */	"unqualified_id_val",
	/* -127 (1) */	"id_expression_val",
	/* -126 (10) */	"expression",
	/* -125 (1) */	"primary_expression",
	/* -124 (1) */	"literal",
	/* -123 (2) */	"simple_operator",
	/* -122 (1) */	"operator",
	/* -121 (2) */	"operator_function_id",
	/* -120 (5) */	"expression_list_opt",
	/* -119 (7) */	"save_statement_seq_opt",
	/* -118 (2) */	"mem_initializer",
	/* -117 (2) */	"save_mem_initializer",
	/* -116 (2) */	"mem_initializer_list",
	/* -115 (1) */	"<206>",
	/* -114 (2) */	"save_mem_initializer_list",
	/* -113 (1) */	"<204>",
	/* -112 (1) */	"ctor_initializer",
	/* -111 (1) */	"conversion_declarator",
	/* -110 (2) */	"conversion_declarator_opt",
	/* -109 (1) */	"conversion_type_id",
	/* -108 (2) */	"conversion_function_id",
	/* -107 (2) */	"base_specifier",
	/* -106 (2) */	"base_specifier_list",
	/* -105 (2) */	"member_declarator",
	/* -104 (3) */	"member_declarator_list",
	/* -103 (1) */	"<172>",
	/* -102 (1) */	"<170>",
	/* -101 (1) */	"<168>",
	/* -100 (11) */	"colon",
	/* -99 (5) */	"access_specifier",
	/* -98 (1) */	"member_declaration",
	/* -97 (1) */	"member_specification",
	/* -96 (5) */	"member_specification_opt",
	/* -95 (1) */	"<158>",
	/* -94 (1) */	"base_clause_opt",
	/* -93 (2) */	"initializer_list",
	/* -92 (4) */	"expression_list",
	/* -91 (3) */	"initializer_clause",
	/* -90 (1) */	"save_compound_statement",
	/* -89 (1) */	"<146>",
	/* -88 (1) */	"<144>",
	/* -87 (0) */	"eval_declarator_or_ctor",
	/* -86 (10) */	"assignment_expression",
	/* -85 (2) */	"parameter_declaration",
	/* -84 (3) */	"parameter_declaration_list",
	/* -83 (2) */	"direct_abstract_declarator",
	/* -82 (2) */	"abstract_declarator",
	/* -81 (2) */	"type_specifier_notypename_seq",
	/* -80 (4) */	"abstract_declarator_opt",
	/* -79 (4) */	"type_specifier_seq",
	/* -78 (5) */	"type_id",
	/* -77 (3) */	"id_expression",
	/* -76 (1) */	"cv_qualifier_seq",
	/* -75 (2) */	"declarator_or_ctor",
	/* -74 (1) */	"ctor_declarator",
	/* -73 (1) */	"ctor_declarator_id",
	/* -72 (7) */	"cv_qualifier_seq_opt",
	/* -71 (4) */	"parameter_declaration_clause",
	/* -70 (1) */	"declarator_id",
	/* -69 (4) */	"ptr_operator",
	/* -68 (4) */	"direct_declarator",
	/* -67 (1) */	"initializer",
	/* -66 (1) */	"<99>",
	/* -65 (2) */	"attribute_opt",
	/* -64 (9) */	"declarator",
	/* -63 (0) */	"init_declarator_list_opt",
	/* -62 (4) */	"init_declarator",
	/* -61 (2) */	"init_declarator_list1",
	/* -60 (0) */	"namespace_name",
	/* -59 (0) */	"using_directive",
	/* -58 (1) */	"unqualified_id",
	/* -57 (0) */	"using_declaration",
	/* -56 (0) */	"qualifier_namespace_specifier",
	/* -55 (0) */	"namespace_alias_definition",
	/* -54 (0) */	"<80>",
	/* -53 (0) */	"<78>",
	/* -52 (0) */	"original_namespace_name",
	/* -51 (0) */	"namespace_body",
	/* -50 (0) */	"<76>",
	/* -49 (0) */	"extension_namespace_definition",
	/* -48 (0) */	"original_namespace_definition",
	/* -47 (0) */	"unnamed_namespace_definition",
	/* -46 (0) */	"named_namespace_definition",
	/* -45 (0) */	"namespace_definition",
	/* -44 (7) */	"constant_expression",
	/* -43 (2) */	"enumerator_definition",
	/* -42 (2) */	"enumerator_list",
	/* -41 (3) */	"comma_opt",
	/* -40 (1) */	"enumerator_list_opt",
	/* -39 (1) */	"<64>",
	/* -38 (6) */	"lbrace",
	/* -37 (2) */	"identifier_opt",
	/* -36 (3) */	"class_key",
	/* -35 (1) */	"typedef_name",
	/* -34 (1) */	"enum_name",
	/* -33 (8) */	"class_name",
	/* -32 (3) */	"cv_qualifier",
	/* -31 (1) */	"elaborated_type_specifier",
	/* -30 (1) */	"enum_specifier",
	/* -29 (1) */	"class_specifier",
	/* -28 (3) */	"simple_type_specifier_notypename",
	/* -27 (1) */	"decl_specifier_seq_opt",
	/* -26 (6) */	"type_name",
	/* -25 (6) */	"decl_specifier_notypename_seq",
	/* -24 (2) */	"function_specifier",
	/* -23 (3) */	"type_specifier_notypename",
	/* -22 (2) */	"storage_class_specifier",
	/* -21 (2) */	"decl_specifier_notypename",
	/* -20 (9) */	"decl_specifier_seq",
	/* -19 (1) */	"asm_directive",
	/* -18 (2) */	"simple_declaration",
	/* -17 (13) */	"identifier",
	/* -16 (2) */	"function_body",
	/* -15 (2) */	"ctor_initializer_opt",
	/* -14 (1) */	"<13>",
	/* -13 (1) */	"id_or_ctor",
	/* -12 (6) */	"save_scope",
	/* -11 (1) */	"linkage_specification",
	/* -10 (2) */	"function_definition",
	/* -9 (2) */	"init_declarator_list",
	/* -8 (2) */	"block_declaration",
	/* -7 (1) */	"none",
	/* -6 (2) */	"declaration",
	/* -5 (1) */	"<4>",
	/* -4 (1) */	"declaration_seq",
	/* -3 (3) */	"compound_statement",
	/* -2 (3) */	"declaration_seq_opt",
	/* -1 (1) */	"translation_unit",
	/* 0 (0) */	"$",
	/* 1 (0) */	"error",
	/* 2 (0) */	"?",
	/* 3 (0) */	"",
	/* 4 (0) */	"",
	/* 5 (0) */	"",
	/* 6 (0) */	"",
	/* 7 (0) */	"",
	/* 8 (0) */	"",
	/* 9 (0) */	"",
	/* 10 (0) */	"",
	/* 11 (0) */	"",
	/* 12 (0) */	"",
	/* 13 (0) */	"",
	/* 14 (0) */	"",
	/* 15 (0) */	"",
	/* 16 (0) */	"",
	/* 17 (0) */	"",
	/* 18 (0) */	"",
	/* 19 (0) */	"",
	/* 20 (0) */	"",
	/* 21 (0) */	"",
	/* 22 (0) */	"",
	/* 23 (0) */	"",
	/* 24 (0) */	"",
	/* 25 (0) */	"",
	/* 26 (0) */	"",
	/* 27 (0) */	"",
	/* 28 (0) */	"",
	/* 29 (0) */	"",
	/* 30 (0) */	"",
	/* 31 (0) */	"",
	/* 32 (0) */	"",
	/* 33 (0) */	"",
	/* 34 (0) */	"",
	/* 35 (0) */	"",
	/* 36 (0) */	"",
	/* 37 (0) */	"",
	/* 38 (0) */	"",
	/* 39 (0) */	"",
	/* 40 (0) */	"",
	/* 41 (0) */	"",
	/* 42 (0) */	"",
	/* 43 (0) */	"",
	/* 44 (0) */	"",
	/* 45 (0) */	"",
	/* 46 (0) */	"",
	/* 47 (0) */	"",
	/* 48 (0) */	"",
	/* 49 (0) */	"",
	/* 50 (0) */	"",
	/* 51 (0) */	"",
	/* 52 (0) */	"",
	/* 53 (0) */	"",
	/* 54 (0) */	"",
	/* 55 (0) */	"",
	/* 56 (0) */	"",
	/* 57 (0) */	"",
	/* 58 (0) */	"",
	/* 59 (0) */	"",
	/* 60 (0) */	"",
	/* 61 (0) */	"",
	/* 62 (0) */	"",
	/* 63 (0) */	"",
	/* 64 (0) */	"",
	/* 65 (0) */	"",
	/* 66 (0) */	"",
	/* 67 (0) */	"",
	/* 68 (0) */	"",
	/* 69 (0) */	"",
	/* 70 (0) */	"",
	/* 71 (0) */	"",
	/* 72 (0) */	"",
	/* 73 (0) */	"",
	/* 74 (0) */	"",
	/* 75 (0) */	"",
	/* 76 (0) */	"",
	/* 77 (0) */	"",
	/* 78 (0) */	"",
	/* 79 (0) */	"",
	/* 80 (0) */	"",
	/* 81 (0) */	"",
	/* 82 (0) */	"",
	/* 83 (0) */	"",
	/* 84 (0) */	"",
	/* 85 (0) */	"",
	/* 86 (0) */	"",
	/* 87 (0) */	"",
	/* 88 (0) */	"",
	/* 89 (0) */	"",
	/* 90 (0) */	"",
	/* 91 (0) */	"",
	/* 92 (0) */	"",
	/* 93 (0) */	"",
	/* 94 (0) */	"",
	/* 95 (0) */	"",
	/* 96 (0) */	"",
	/* 97 (0) */	"",
	/* 98 (0) */	"",
	/* 99 (0) */	"",
	/* 100 (0) */	"",
	/* 101 (0) */	"",
	/* 102 (0) */	"",
	/* 103 (0) */	"",
	/* 104 (0) */	"",
	/* 105 (0) */	"",
	/* 106 (0) */	"",
	/* 107 (0) */	"",
	/* 108 (0) */	"",
	/* 109 (0) */	"",
	/* 110 (0) */	"",
	/* 111 (0) */	"",
	/* 112 (0) */	"",
	/* 113 (0) */	"",
	/* 114 (0) */	"",
	/* 115 (0) */	"",
	/* 116 (0) */	"",
	/* 117 (0) */	"",
	/* 118 (0) */	"",
	/* 119 (0) */	"",
	/* 120 (0) */	"",
	/* 121 (0) */	"",
	/* 122 (0) */	"",
	/* 123 (0) */	"",
	/* 124 (0) */	"",
	/* 125 (0) */	"",
	/* 126 (0) */	"",
	/* 127 (0) */	"",
	/* 128 (0) */	"",
	/* 129 (0) */	"",
	/* 130 (0) */	"",
	/* 131 (0) */	"",
	/* 132 (0) */	"",
	/* 133 (0) */	"",
	/* 134 (0) */	"",
	/* 135 (0) */	"",
	/* 136 (0) */	"",
	/* 137 (0) */	"",
	/* 138 (0) */	"",
	/* 139 (0) */	"",
	/* 140 (0) */	"",
	/* 141 (0) */	"",
	/* 142 (0) */	"",
	/* 143 (0) */	"",
	/* 144 (0) */	"",
	/* 145 (0) */	"",
	/* 146 (0) */	"",
	/* 147 (0) */	"",
	/* 148 (0) */	"",
	/* 149 (0) */	"",
	/* 150 (0) */	"",
	/* 151 (0) */	"",
	/* 152 (0) */	"",
	/* 153 (0) */	"",
	/* 154 (0) */	"",
	/* 155 (0) */	"",
	/* 156 (0) */	"",
	/* 157 (0) */	"",
	/* 158 (0) */	"",
	/* 159 (0) */	"",
	/* 160 (0) */	"",
	/* 161 (0) */	"",
	/* 162 (0) */	"",
	/* 163 (0) */	"",
	/* 164 (0) */	"",
	/* 165 (0) */	"",
	/* 166 (0) */	"",
	/* 167 (0) */	"",
	/* 168 (0) */	"",
	/* 169 (0) */	"",
	/* 170 (0) */	"",
	/* 171 (0) */	"",
	/* 172 (0) */	"",
	/* 173 (0) */	"",
	/* 174 (0) */	"",
	/* 175 (0) */	"",
	/* 176 (0) */	"",
	/* 177 (0) */	"",
	/* 178 (0) */	"",
	/* 179 (0) */	"",
	/* 180 (0) */	"",
	/* 181 (0) */	"",
	/* 182 (0) */	"",
	/* 183 (0) */	"",
	/* 184 (0) */	"",
	/* 185 (0) */	"",
	/* 186 (0) */	"",
	/* 187 (0) */	"",
	/* 188 (0) */	"",
	/* 189 (0) */	"",
	/* 190 (0) */	"",
	/* 191 (0) */	"",
	/* 192 (0) */	"",
	/* 193 (0) */	"",
	/* 194 (0) */	"",
	/* 195 (0) */	"",
	/* 196 (0) */	"",
	/* 197 (0) */	"",
	/* 198 (0) */	"",
	/* 199 (0) */	"",
	/* 200 (0) */	"",
	/* 201 (0) */	"",
	/* 202 (0) */	"",
	/* 203 (0) */	"",
	/* 204 (0) */	"",
	/* 205 (0) */	"",
	/* 206 (0) */	"",
	/* 207 (0) */	"",
	/* 208 (0) */	"",
	/* 209 (0) */	"",
	/* 210 (0) */	"",
	/* 211 (0) */	"",
	/* 212 (0) */	"",
	/* 213 (0) */	"",
	/* 214 (0) */	"",
	/* 215 (0) */	"",
	/* 216 (0) */	"",
	/* 217 (0) */	"",
	/* 218 (0) */	"",
	/* 219 (0) */	"",
	/* 220 (0) */	"",
	/* 221 (0) */	"",
	/* 222 (0) */	"",
	/* 223 (0) */	"",
	/* 224 (0) */	"",
	/* 225 (0) */	"",
	/* 226 (0) */	"",
	/* 227 (0) */	"",
	/* 228 (0) */	"",
	/* 229 (0) */	"",
	/* 230 (0) */	"",
	/* 231 (0) */	"",
	/* 232 (0) */	"",
	/* 233 (0) */	"",
	/* 234 (0) */	"",
	/* 235 (0) */	"",
	/* 236 (0) */	"",
	/* 237 (0) */	"",
	/* 238 (0) */	"",
	/* 239 (0) */	"",
	/* 240 (0) */	"",
	/* 241 (0) */	"",
	/* 242 (0) */	"",
	/* 243 (0) */	"",
	/* 244 (0) */	"",
	/* 245 (0) */	"",
	/* 246 (0) */	"",
	/* 247 (0) */	"",
	/* 248 (0) */	"",
	/* 249 (0) */	"",
	/* 250 (0) */	"",
	/* 251 (0) */	"",
	/* 252 (0) */	"",
	/* 253 (0) */	"",
	/* 254 (0) */	"",
	/* 255 (0) */	"",
	/* 256 (0) */	"",
	/* 257 (2) */	"_ID",
	/* 258 (2) */	"_NUM",
	/* 259 (2) */	"_CHAR",
	/* 260 (2) */	"_FLOAT",
	/* 261 (5) */	"_STR",
	/* 262 (0) */	"_CSTR",
	/* 263 (2) */	"_BOOL",
	/* 264 (0) */	"_BACKSTR",
	/* 265 (0) */	"_ANGSTR",
	/* 266 (0) */	"_LINE",
	/* 267 (0) */	"_SPACE",
	/* 268 (0) */	"_EOL",
	/* 269 (0) */	"_CPP",
	/* 270 (0) */	"_CPP_JOIN",
	/* 271 (0) */	"_CPP_DEFINED",
	/* 272 (0) */	"_CPP_DEFINE",
	/* 273 (0) */	"_CPP_UNDEF",
	/* 274 (0) */	"_CPP_IF",
	/* 275 (0) */	"_CPP_IFDEF",
	/* 276 (0) */	"_CPP_IFNDEF",
	/* 277 (0) */	"_CPP_ENDIF",
	/* 278 (0) */	"_CPP_ELIF",
	/* 279 (0) */	"_CPP_INCLUDE",
	/* 280 (0) */	"_CPP_PRAGMA",
	/* 281 (0) */	"_CPP_ASSERT",
	/* 282 (0) */	"_CPP_UNASSERT",
	/* 283 (0) */	"_CPP_LINE",
	/* 284 (0) */	"_CPP_ERROR",
	/* 285 (0) */	"_CPP_WARNING",
	/* 286 (0) */	"_CPP_IDENT",
	/* 287 (0) */	"_CPP__FILE",
	/* 288 (0) */	"_CPP__LINE",
	/* 289 (0) */	"_CPP__DATE",
	/* 290 (0) */	"_CPP__TIME",
	/* 291 (0) */	"_CPP__STDC",
	/* 292 (0) */	"_CPP__STDCVER",
	/* 293 (16) */	"_EOS",
	/* 294 (2) */	"_COLON",
	/* 295 (13) */	"_COMMA",
	/* 296 (2) */	"_QMARK",
	/* 297 (0) */	"_POUND",
	/* 298 (2) */	"_LBRACE",
	/* 299 (8) */	"_RBRACE",
	/* 300 (7) */	"_LBRACK",
	/* 301 (7) */	"_RBRACK",
	/* 302 (33) */	"_LPAREN",
	/* 303 (33) */	"_RPAREN",
	/* 304 (0) */	"_QLINE",
	/* 305 (0) */	"_DOTDOT",
	/* 306 (0) */	"_AT",
	/* 307 (0) */	"_BSLASH",
	/* 308 (3) */	"_OP_ADD",
	/* 309 (3) */	"_OP_SUB",
	/* 310 (5) */	"_OP_MUL",
	/* 311 (2) */	"_OP_DIV",
	/* 312 (2) */	"_OP_MOD",
	/* 313 (2) */	"_OP_NOT",
	/* 314 (4) */	"_OP_AND",
	/* 315 (2) */	"_OP_OR",
	/* 316 (2) */	"_OP_XOR",
	/* 317 (2) */	"_OP_LNOT",
	/* 318 (2) */	"_OP_LAND",
	/* 319 (2) */	"_OP_LOR",
	/* 320 (0) */	"_OP_LXOR",
	/* 321 (3) */	"_OP_LSS",
	/* 322 (3) */	"_OP_GTR",
	/* 323 (2) */	"_OP_LEQ",
	/* 324 (2) */	"_OP_GEQ",
	/* 325 (2) */	"_OP_EQU",
	/* 326 (2) */	"_OP_NEQ",
	/* 327 (3) */	"_OP_INC",
	/* 328 (3) */	"_OP_DEC",
	/* 329 (2) */	"_OP_LSHIFT",
	/* 330 (2) */	"_OP_RSHIFT",
	/* 331 (8) */	"_OP_ASN",
	/* 332 (2) */	"_OP_ASNADD",
	/* 333 (2) */	"_OP_ASNSUB",
	/* 334 (2) */	"_OP_ASNMUL",
	/* 335 (2) */	"_OP_ASNDIV",
	/* 336 (2) */	"_OP_ASNMOD",
	/* 337 (2) */	"_OP_ASNAND",
	/* 338 (2) */	"_OP_ASNOR",
	/* 339 (2) */	"_OP_ASNXOR",
	/* 340 (2) */	"_OP_ASNLSHIFT",
	/* 341 (2) */	"_OP_ASNRSHIFT",
	/* 342 (3) */	"_OP_DOT",
	/* 343 (3) */	"_OP_PTRTO",
	/* 344 (1) */	"_OP_SCOPE",
	/* 345 (3) */	"_ELIPSIS",
	/* 346 (2) */	"_OP_MDOT",
	/* 347 (2) */	"_OP_MPTRTO",
	/* 348 (0) */	"_OP_CONDASN",
	/* 349 (1) */	"_KW_ATTRIBUTE",
	/* 350 (1) */	"_KW_CONST",
	/* 351 (1) */	"_KW_VOLATILE",
	/* 352 (1) */	"_KW_AUTO",
	/* 353 (1) */	"_KW_STATIC",
	/* 354 (3) */	"_KW_EXTERN",
	/* 355 (1) */	"_KW_REGISTER",
	/* 356 (1) */	"_KW_LONG",
	/* 357 (1) */	"_KW_SHORT",
	/* 358 (1) */	"_KW_SIGNED",
	/* 359 (1) */	"_KW_UNSIGNED",
	/* 360 (1) */	"_KW_INT",
	/* 361 (1) */	"_KW_CHAR",
	/* 362 (1) */	"_KW_FLOAT",
	/* 363 (1) */	"_KW_DOUBLE",
	/* 364 (1) */	"_KW_VOID",
	/* 365 (4) */	"_KW_IF",
	/* 366 (3) */	"_KW_ELSE",
	/* 367 (2) */	"_KW_DO",
	/* 368 (3) */	"_KW_WHILE",
	/* 369 (2) */	"_KW_FOR",
	/* 370 (2) */	"_KW_SWITCH",
	/* 371 (2) */	"_KW_CASE",
	/* 372 (2) */	"_KW_DEFAULT",
	/* 373 (2) */	"_KW_BREAK",
	/* 374 (2) */	"_KW_CONTINUE",
	/* 375 (2) */	"_KW_GOTO",
	/* 376 (2) */	"_KW_RETURN",
	/* 377 (3) */	"_KW_ENUM",
	/* 378 (1) */	"_KW_UNION",
	/* 379 (1) */	"_KW_STRUCT",
	/* 380 (2) */	"_KW_TYPEDEF",
	/* 381 (3) */	"_OP_SIZEOF",
	/* 382 (0) */	"_KW__FUNC",
	/* 383 (0) */	"_KW__PFUNC",
	/* 384 (3) */	"_KW_OPERATOR",
	/* 385 (3) */	"_KW_CLASS",
	/* 386 (2) */	"_KW_THIS",
	/* 387 (3) */	"_KW_TYPEID",
	/* 388 (1) */	"_KW_TYPENAME",
	/* 389 (1) */	"_KW_NAMESPACE",
	/* 390 (1) */	"_KW_USING",
	/* 391 (1) */	"_KW_TEMPLATE",
	/* 392 (1) */	"_KW_PUBLIC",
	/* 393 (1) */	"_KW_PRIVATE",
	/* 394 (1) */	"_KW_PROTECTED",
	/* 395 (2) */	"_KW_FRIEND",
	/* 396 (5) */	"_KW_VIRTUAL",
	/* 397 (1) */	"_KW_MUTABLE",
	/* 398 (1) */	"_KW_EXPLICIT",
	/* 399 (1) */	"_KW_BOOL",
	/* 400 (0) */	"_KW_TRUE",
	/* 401 (0) */	"_KW_FALSE",
	/* 402 (1) */	"_KW_WCHAR",
	/* 403 (1) */	"_KW_THROW",
	/* 404 (1) */	"_KW_CATCH",
	/* 405 (1) */	"_KW_TRY",
	/* 406 (1) */	"_KW_SCAST",
	/* 407 (1) */	"_KW_DCAST",
	/* 408 (1) */	"_KW_CCAST",
	/* 409 (1) */	"_KW_RCAST",
	/* 410 (1) */	"_KW_INLINE",
	/* 411 (2) */	"_KW_ASM",
	/* 412 (1) */	"_OP_TYPEOF",
	/* 413 (3) */	"_OP_NEW",
	/* 414 (2) */	"_OP_DELETE",
	/* 415 (1) */	"_OP_NEWA",
	/* 416 (2) */	"_OP_DELETEA",
	/* 417 (3) */	"_KW_DYNAMIC",
	/* 418 (2) */	"_VAL_ID",
	/* 419 (2) */	"_AGGR_ID",
	/* 420 (2) */	"_ENUM_ID",
	/* 421 (2) */	"_TYPE_ID",
	/* 422 (2) */	"_CTOR_ID",
	/* 423 (4) */	"_DTOR_ID",
	/* 424 (0) */	"_OPER_ID",
	/* 425 (0) */	"_CONV_ID",
	/* 426 (1) */	"_NSPACE_ID",
	/* 427 (1) */	"_ASPACE_ID",
	/* 428 (1) */	"_TPLATE_ID",
	/* 429 (1) */	"_SCOPE",
	/* 430 (0) */	"_TOKENLIST",
	/* 431 (0) */	"_OP_REF",
	/* 432 (0) */	"_OP_PTR",
	/* 433 (0) */	"_OP_MPTR",
	/* 434 (0) */	"_OP_UMINUS",
	/* 435 (0) */	"_OP_UPLUS",
	/* 436 (0) */	"_OP_INCP",
	/* 437 (0) */	"_OP_DECP",
	/* 438 (0) */	"_OP_TSIZEOF",
	/* 439 (0) */	"_OP_COMMA",
	/* 440 (0) */	"_OP_FUNC",
	/* 441 (0) */	"_OP_INDEX",
	/* 442 (0) */	"_OP_CAST",
	/* 443 (1) */	"_KW_DCLASS",
	/* 444 (1) */	"_KW_DTYPE",
	/* 445 (0) */	"_NEXT_INSN",
	/* 446 (0) */	"_BACKPATCH",
	/* 447 (0) */	"_ACTION",
	/* 448 (0) */	"_TBC",
	/* 449 (2) */	"_LBRACE_SAVE",
	/* 450 (2) */	"_COLON_SAVE",
	/* 451 (1) */	"_PARSE_COMPOUND",
	/* 452 (1) */	"_DEFINE",
#line 202 "/u4/shaggy/u2/src/sos/util/shacc/shacc.c.parser"
};
#endif

#ifdef YYIMPURE
YYSTYPE yylval;
#endif


#ifdef YYPDEBUG
static int nalloc = 0, nfree = 0, nnew = 0, ndelete = 0, ncopy = 0, nwrite = 0;
#endif

static YYNode *aln = 0;

#ifdef YYCLASS
class YYYYCLASS : YYCLASS {
#if YYDEBUG
	void yyprintnode(const YYNode *np);
#endif
#ifdef YYPDEBUG
	void yyprintfullnode(YY *yy, const YYNode *np);
	void yyprintactive(YY *yy, const YYNode *np, const char *str);
#endif
#if YYDEBUG
	void yyprintrule(int rn, int nt);
	void yyprintactiverules(YY *yy, const YYNode *np);
#endif
	int yyfollow(YY *yy, const YYNode *np, int s);
	void yygc(YY *yy, const YYNode *np);
	YYYYCLASS();
public:
	int YYPARSE(YYPARSEPARAMS);
};
#endif

#ifndef CCAST
#	ifdef _cplusplus
#		define	CCAST(t)	const_cast<t>	
#	else
#		define CCAST(t)	(t)
#	endif
#endif

#ifndef RCAST
#	ifdef _cplusplus
#		define	RCAST(t)	reinterpret_cast<t>	
#	else
#		define	RCAST(t)	(t)
#	endif
#endif

#define YYNODE_BAD	(RCAST(const YYNode*)(-1))

#define yynode_addact(yy, np) \
{ YYNode *_np = CCAST(YYNode*)(np); _np->a = (yy)->anp; (yy)->anp = _np; }
#define yynode_addred(yy, np) \
{ YYNode *_np = CCAST(YYNode*)(np); _np->a = (yy)->rnp; (yy)->rnp = _np; }
#define yynode_add(yy, np) \
{ const YYNode *_cnp = (np); if(_cnp->x >= yyrule[_cnp->r].ns) \
	yynode_addred((yy), _cnp) else yynode_addact((yy), _cnp); }


static YYNode *
yynode_alloc(YY *yy)
{
	YYNode *np;
	if(!yy->free) {
		int n = 255;
		void **m = RCAST(void**)(YYMALLOC(sizeof(void*)
							+n*sizeof(YYNode)));
		*m = yy->alloc;
		yy->alloc = RCAST(void*)(m);
		np = CCAST(YYNode*)(m+1);
		while(n--) {
			np->a = yy->free;
			yy->free = np;
			np++;
		}
	}
	np = yy->free;
	yy->free = RCAST(YYNode*)(yy->free->a);
	DBG(++nalloc);
	if(aln)
		aln->pp = np;
	np->np = aln;
	np->pp = 0;
	aln = np;
	return np;
}

void
yynode_free(YY *yy, YYNode *np)
{
	np->a = yy->free;
	yy->free = np;
	if(np->pp)
		np->pp->np = np->np;
	else 
		aln = np->np;
	if(np->np)
		np->np->pp = np->pp;
	DBG(++nfree);
}

void
yydogc(YYNode *np)
{
	if(np && np != YYNODE_BAD && !np->mark++) {
		yydogc(CCAST(YYNode*)(np->p));
		yydogc(CCAST(YYNode*)(np->s));
		yydogc(CCAST(YYNode*)(np->a));
	}
}

/* garbage collection - to get rid of cyclic structures that aren't
 * caught by reference counting alone
 */
void
YYPREFIX yygc(YY *yy, const YYNode *np)
{	
	YYNode *a = aln;
	int i;
	yy = yy;
	while(a) {
		a->mark = 0;
		a = a->np;
	}
	yydogc(CCAST(YYNode*)(np));
	yydogc(CCAST(YYNode*)(yy->anp));
	yydogc(CCAST(YYNode*)(yy->rnp));
	for(i=0; i<YYNNONTERMS; i++)
		yydogc(CCAST(YYNode*)(yy->nts[i]));
	i = 0;
	a = aln;
	while(a) {
		YYNode *n = a->np;
		if(!a->mark) {
			yynode_free(yy, a);
			if(a->s) {
				--(CCAST(YYNode*)(a->s))->rc;
				DBG(ndelete++);
			}
			if(a->p) {
				--(CCAST(YYNode*)(a->p))->rc;
				DBG(ndelete++);
			}
			i++;
		}
		a = n;
	}
#ifdef YYPDEBUG
	a = aln;
	while(a) {
		if(a->mark != a->rc)
			YYEPRINTF("GC RC: rc=%d, %d [%p]\n",
				a->rc, a->mark, a);
		a = a->np;
	}
	YYEPRINTF("GC: %d (%d)\n", i, i*sizeof(YYNode));
#endif
}

static YYNode *
yynode_new(YY *yy, int r, int x, const YYNode *p)
{
	YYNode *np = yynode_alloc(yy);
	np->rc = 1;
	np->c = -1;
	np->c1 = -1;
	np->s = 0;
	np->r = r;
	np->x = x;
	np->p = p;
	np->a = 0;
	np->rp = 0;
	DBG(++nnew;)
	return np;
}

static int
yynode_delete(YY *yy, const YYNode *np)
{
	YYNode *wnp = CCAST(YYNode*)(np);
	if(!wnp || wnp == YYNODE_BAD)
		return 0;
	DBG(++ndelete;)
	if(wnp->rc <= 0) {
		// sprintf(yy->errstr, "internal error: ref count rc=%d r=%d x=%d",
		//		wnp->rc, wnp->r, wnp->x);
		return -1;
	}
	if(!--wnp->rc) {
		yynode_delete(yy, wnp->s);
		yynode_delete(yy, wnp->p);
		yynode_free(yy, wnp);
	}
	/* Crazy cyclic checks... GC takes care of this now  
	else if(wnp->rc == 1 && wnp->p && wnp->p->rc == 1
			&& wnp->p->s == wnp) {
		np = wnp->p;
		wnp->p = 0;
		yynode_delete(yy, np);
	} else if(wnp->rc == 1 && wnp->s && wnp->s->rc == 1
			&& wnp->s->p == wnp) {
		np = wnp->s;
		wnp->s = 0;
		yynode_delete(yy, np);
	}
	*/
	return 0;
}

static const YYNode *
yynode_copy(YY *yy, const YYNode *np)
{
	yy = yy;
	if(!np || np == YYNODE_BAD)
		return np;
	if(!++(CCAST(YYNode*)(np))->rc) {
		// sprintf(yy->errstr, "internal error: ref count wraparound");
	}
	DBG(++ncopy;)
	return np;
}

void
yynode_assign(YY *yy, const YYNode **vnp, const YYNode *np)
{
	if(*vnp != np) {
		yynode_delete(yy, *vnp);
		*vnp = yynode_copy(yy, np);
	}
}

static YYNode *
yynode_canwrite(YY *yy, const YYNode *np)
{
	YYNode *wnp;
	if(np->rc == 1)
		wnp = CCAST(YYNode*)(np);
	else {
		YYNode *cnp, *cpp;
		wnp = yynode_alloc(yy);
		cnp = wnp->np;
		cpp = wnp->pp;
		*wnp = *np;
		wnp->np = cnp;
		wnp->pp = cpp;
		// memcpy(wnp, np, sizeof(*np)-8);
		wnp->rc = 1;
		wnp->a = 0;
		--(CCAST(YYNode*)(np))->rc;
		wnp->s = yynode_copy(yy, wnp->s);
		wnp->p = yynode_copy(yy, wnp->p);
	}
	DBG(++nwrite;)
	return wnp;
}

/* Notes:
 *	- w.r.t. ref counting and COW, YYNode.a is not considered to
 *	  be part of YYNode.  This is because it is simply used as 
 *	  a convenient place to keep the active and active reduction
 *	  lists, and since only one of each list exists at a time,
 *	  and there is no overlap, it is effectively equivalent to keeping
 *	  two separate lists for these and doesn't affect the COW semantics.
 *
 *	- All functions from this point on that take a YYNode* assume
 *	  that an explicit copy has been made if necessary, since they
 *	  can potentially modify the node.  These functions will
 *	  delete the copy when they are done (unless they add another
 *	  reference to the node).  If the node is going to be deleted
 *	  afterwards anyway, no copy is necessary, and this results
 *	  in an efficiency gain, since no extra copies are created.
 *
 *	- The substitution chain (YYNode.s) is another unusual case,
 *	  since the link needs to be added to nodes that already exist
 *
 *	- Possible bug: is it possible for an active node to get
 *	  made active again before reaching it while traversing
 *	  the active list?  (in which case, the YYNode.a pointer would
 *	  get changed before that part of the list is reached)
 *	  For now, a safety check will catch this and report an
 *	  internal error.
 *	
 *	- cyclic arrangements of nodes will never get deleted via
 *	  reference counting.  A garbage collector deals with this.
 *	
 *	- Recursive rule sets can cause themselves to exist,
 *	  since a rule needs to be activated before it can be tested
 */


#if YYDEBUG
YYSTATIC void
YYPREFIX yyprintnode(const YYNode *np)
{
	int i;
	YYRule *r;
	r = np->p ? &yyrule[np->p->r] : 0;
	YYEPRINTF("%s <-", r ? yysname[yysym[r->sym+np->p->x]
				+YYNNONTERMS-1] : "$");
	r = &yyrule[np->r];
	for(i=0; i<r->ns; i++) {
		if(i == np->x)
			YYEPRINTF(" .");
		YYEPRINTF(" %s",
			yysname[yysym[r->sym+i]+YYNNONTERMS-1]);
	}
	YYEPRINTF("\n");
}
#endif

#ifdef YYPDEBUG
YYSTATIC void
YYPREFIX yyprintfullnode(YY *yy, const YYNode *np)
{
	YYEPRINTF("\t%p: r=%d x=%d c=%d c1=%d p=%p s=%p rc=%d [r=%d]\n",
		np, np->r, np->x, np->c, np->c1, np->p, np->s, np->rc,
		np->c >= 0 ? yy->rstk[np->c].r : -1);
	YYEPRINTF("\t");
	yyprintnode(np);
}

YYSTATIC void
YYPREFIX yyprintactive(YY *yy, const YYNode *np, const char *str)
{
	YYEPRINTF("%s:\n", str);
	while(np) {
		yyprintfullnode(yy, np);
	 	np = np->a;
	}
}
#endif


YYSTATIC int
YYPREFIX yyfollow(YY *yy, const YYNode *np, int s)
{
	int m;
	if(np->x >= yyrule[np->r].ns) {
		yynode_addred(yy, np);
		return 1;
	}
	m = yysym[yyrule[np->r].sym + np->x];
	if(m < 0) {	/* it's a nonterminal */
		const YYNode **ntsp;
		YYNode *p;
		int rv=0;
		int y = 0;
		m = -m;
		ntsp = &yy->nts[m];
		if(*ntsp == YYNODE_BAD) {
			yynode_delete(yy, np);
			return 0;
		}
		while(*ntsp && (*ntsp)->c != np->c) 
			ntsp = &(*CCAST(YYNode**)(ntsp))->a;
		if(*ntsp) {
			YYNode *wnp;
			while((*ntsp)->s)
				ntsp = &(*CCAST(YYNode**)(ntsp))->s;
			wnp = CCAST(YYNode*)(*ntsp);
			p = yynode_canwrite(yy, np);
			DBG(YYEPRINTF("linking %p to %p\n", p, wnp));
			DBG(yyprintfullnode(yy, p));
			yynode_assign(yy, &p->s, 0);
			yynode_assign(yy, &wnp->s, p);
			yynode_delete(yy, p);
			return 1;
		}
		/* for recursive rules, need to make sure they don't
		 * simply cause themselves to exist...  need some valid
		 * reduction as well (will gc clean these up anyway,
		 * or are they cyclic?)
		 */
		p = yynode_canwrite(yy, np);
		yynode_assign(yy, ntsp, p);
		yynode_assign(yy, &p->s, 0);
		y = yynt[m];
		while(y > 0) {
			YYNode *wnp = yynode_new(yy, y, 0, yynode_copy(yy, p));
			wnp->c = p->c;
			wnp->c1 = p->c1;
			if(yyfollow(yy, wnp, s))
				rv=1;
			if(yyrule[y].skip)
				y += yyrule[y].skip;
			else	
				y = -1;
		}
		if(!rv) {
			// if(yy->nts[m])
			//	YYINTERR;
			yynode_assign(yy, &yy->nts[m], YYNODE_BAD);
		}
		yynode_delete(yy, p);
		return rv;
	} else {	/* it's a terminal */
		if(s == -1) {
			yynode_addact(yy, np);
			return 1;
		} else if(m == s) {
			YYNode *wnp = yynode_canwrite(yy, np);
			++wnp->x;
			yynode_add(yy, wnp);
			return 1;
		} else {
			yynode_delete(yy, np);
			return 0;
		}
	}
}



void
yyclearrc(YYNode *np)
{
	if(np->mark) {
		np->mark = 0;
		np->c = -1;
		if(np->s)
			yyclearrc(CCAST(YYNode*)(np->s));
		if(np->p)
			yyclearrc(CCAST(YYNode*)(np->p));
	}
}

#if YYDEBUG
void
YYPREFIX yyprintrule(int rn, int nt)
{
	int i;
	YYRule *r = &yyrule[rn];
	YYEPRINTF("%s <-", yysname[nt+YYNNONTERMS-1]);
	for(i=0; i<r->ns; i++)
		YYEPRINTF(" %s", yysname[yysym[r->sym+i] +YYNNONTERMS-1]);
	YYEPRINTF("\n");
}

void
YYPREFIX yyprintactiverules(YY *yy, const YYNode *np)
{
	while(np) {
		int c = np->c;
		YYEPRINTF("yydebug: ");
		yyprintnode(np);
		while(c >= 0) {	
			int nc = yy->rstk[c].p;
			yy->rstk[c].p = -1;
			YYEPRINTF("yydebug: %d: ", c);
			yyprintrule(yy->rstk[c].r, yy->rstk[c].nt);
			c = nc;
		}
		np = np->a;
	}
}
#endif


/* General overview of main loop: (needs updating)
 *	- if there's only one active rule or reduction
 *	 	- make sure any substitutions have equivalent reductions
 *		- if so, perform all reductions
 *	- go through all active reductions
 *		- substitute back into previous rules,
 *		  adding to reduction chain
 *		- check for any substitution links, and follow them,
 *		  also adding to reduction chain for each
 *	- if we don't currently have a token
 *		- get a new token
 *	- follow all active rules (advance to next token, eliminating
 *		the rule if it doesn't match)
 *	- if there are no active reductions
 *		- save yylval on stack
 *		- eat the current token	
 */

int
YYPREFIX YYPARSE(YYPARSEPARAMS)
{
	YYSTYPE yylval;
	int success = 0;
	int s, i;
	int eof = 0;
	YY yy;
	int maxrsp = 0;
	int rstksize = YYINITDEPTH;
	int ambig = 0;

	YYPARSE_VARS

	yy.rstk = RCAST(YYRedNode*)(YYMALLOC(rstksize*sizeof(YYRedNode)));
	yy.vsa = RCAST(YYSTYPE*)(YYMALLOC(YYINITDEPTH*sizeof(YYSTYPE)));
	yy.dvsa = RCAST(YYSTYPE*)(YYMALLOC(YYINITDEPTH*sizeof(YYSTYPE)));
	yy.rsp = 0;
	yy.dsp = 0;
	yy.dsrp = 0;
	yy.anp = 0;
	yy.rnp = 0;
	yy.vsp = yy.vsa;
	yy.alloc = 0;
	yy.free = 0;

	for(i=0; i<YYNNONTERMS; i++)
		yy.nts[i] = 0;
	yy.anp = 0;
	yynode_addact(&yy, yynode_new(&yy, 0,0,0));
	s = -1;
	while(yy.anp || yy.rnp) {
		const YYNode *np;
		int pr = 0;
		int cf = -1;
		int ct = -1;

#ifdef YYPDEBUG
		YYEPRINTF("----------------------------------------------\n");
		yyprintactive(&yy, yy.anp, "active");
		yyprintactive(&yy, yy.rnp, "reduce");
#endif

		for(;;) {
			int c1;
			const YYNode *snp;
			np = yy.anp ? yy.anp : yy.rnp;
			snp = yy.anp ? yy.rnp : 0;
			c1 = np->c1;
			while(np) {
				if(np->c1 != c1) {
					c1 = -2;
					break;
				}
				if(!(np = np->a)) {
					np = snp;
					snp = 0;
				}
			}
		
			if(c1 == -1) {
				DBG(YYEPRINTF("no reductions\n"));
				break;
			} else if(c1 == -2) {
				DBG(YYEPRINTF("reduction ambiguity\n"));
				break;
			}
			DBG(YYEPRINTF("could reduce %d (cf=%d,ct=%d)\n",
						c1, cf, ct));
			yy.rstk[c1].p = -1;
			if(cf < 0)
				cf = c1;
			else
				yy.rstk[ct].p = c1;
			ct = c1;
			pr = 0;
			np = yy.anp ? yy.anp : yy.rnp;
			snp = yy.anp ? yy.rnp : 0;
			while(np) {
				int nc = np->c;
				int nnc;
				if(nc == c1) {
					nc = -1;
					(CCAST(YYNode*)(np))->c = -1;
				} else {
					pr = 1;
					while((nnc = yy.rstk[nc].p) != c1) {	
						if(nnc == -1)
							break;
						nc = nnc;
					}
					yy.rstk[nc].p = -1;
				}
				(CCAST(YYNode*)(np))->c1 = nc;
				if(!(np = np->a)) {
					np = snp;
					snp = 0;
				}
			}
		}

		/* perform any reduction actions that can be done */
		DBG(YYEPRINTF("cf=%d pr=%d\n", cf, pr));
		if(cf >= 0) {
			int sp = cf;
			YYSTYPE *yyvsp = yy.vsp;
			if(yy.rnp && !yy.rnp->a) {
				DBG(YYEPRINTF("Only one active reduction!\n"));
				yygc(&yy, yy.rnp);
			}
			while(sp >= 0) {
				YYRedNode *rnp = &yy.rstk[sp];
				YYSTYPE yyval;
				YYSTYPE *nvsp;

				while(yy.dsrp < rnp->sp)
					*yyvsp++ = yy.dvsa[yy.dsrp++];
				nvsp = yyvsp - yyrule[rnp->r].ns;
				yyval = *nvsp;
#if YYDEBUG
				if(YYDEBUGFLAG) {
					YYEPRINTF("yydebug: ");
					yyprintrule(rnp->r, rnp->nt);
				}
#endif
				switch(rnp->r) {
				/* 0: <EOF>: translation_unit <EOF> */
				/* 1: translation_unit: declaration_seq_opt */
				case 2: /* translation_unit: _PARSE_COMPOUND compound_statement */
#line 562 "CParse.y"
					{ /* not used yet */ }
					break;
				case 3: /* declaration_seq: declaration_seq_opt <4> declaration */
#line 574 "CParse.y"
					{ _caction->complete(true);
		  if(_deftok && !_savecode) {
			inject(_deftok);
			_deftok = 0;
		  } }
					break;
				case 4: /* <4>: */
#line 2472 "CParse.y"
					{ _caction->complete(false); }
					break;
				case 5: /* declaration_seq_opt: */
				yyrule5: /* 19 */
#line 581 "CParse.y"
					{ yyval = none; }
					break;
				/* 6: declaration_seq_opt: declaration_seq */
				case 7: /* none: */
#line 592 "CParse.y"
					goto yyrule5;
					break;
				/* 8: declaration: block_declaration */
				/* 9: declaration: none init_declarator_list _EOS */
				/* 10: declaration: function_definition */
				/* 11: declaration: linkage_specification */
				case 12: /* declaration: _DEFINE save_scope id_or_ctor <13> ctor_initializer_opt function_body */
#line 611 "CParse.y"
					{ if(yyvsp[-2].token != T(NONE))
			do2(yyvsp[-1], CtorInit, yyvsp[-2], yyvsp[-1]);
		  do2(yyval, Define, yyvsp[-4], yyvsp[-1]);
		  _current_scope = yyvsp[-5]; }
					break;
				case 13: /* <13>: */
#line 2472 "CParse.y"
					{ if(yyvsp[-1].scope) {
			/* XXX - hack */
			_current_scope.token = T(SCOPE);
			_current_scope.tuple.right = yyvsp[-1].scope;
		  }
		  yyval = yyvsp[-1]; }
					break;
				/* 14: id_or_ctor: identifier */
				/* 15: id_or_ctor: _CTOR_ID */
				/* 16: block_declaration: simple_declaration */
				/* 17: block_declaration: asm_directive */
				/* 18: simple_declaration: decl_specifier_seq init_declarator_list _EOS */
				case 19: /* simple_declaration: decl_specifier_seq _EOS */
				yyrule19: /* 1 */
#line 640 "CParse.y"
					{ do3(yyval, DeclSym, yyvsp[-2], none, _current_scope); }
					break;
				/* 20: decl_specifier_notypename: storage_class_specifier */
				/* 21: decl_specifier_notypename: type_specifier_notypename */
				/* 22: decl_specifier_notypename: function_specifier */
				/* 23: decl_specifier_notypename: _KW_FRIEND */
				/* 24: decl_specifier_notypename: _KW_TYPEDEF */
				case 25: /* decl_specifier_notypename_seq: decl_specifier_notypename */
				yyrule25: /* 3 */
#line 663 "CParse.y"
					{ do2(yyval, BuildTypeList, none, yyvsp[-1]); }
					break;
				case 26: /* decl_specifier_notypename_seq: decl_specifier_notypename_seq decl_specifier_notypename */
				yyrule26: /* 2 */
#line 665 "CParse.y"
					{ do2(yyval, BuildTypeList, yyvsp[-2], yyvsp[-1]); }
					break;
				/* 27: decl_specifier_seq: decl_specifier_notypename_seq */
				case 28: /* decl_specifier_seq: type_name */
#line 671 "CParse.y"
					goto yyrule25;
					break;
				case 29: /* decl_specifier_seq: decl_specifier_notypename_seq type_name */
#line 673 "CParse.y"
					goto yyrule26;
					break;
				case 30: /* decl_specifier_seq: type_name decl_specifier_notypename_seq */
				yyrule30: /* 1 */
#line 675 "CParse.y"
					{ do2(yyval, BuildTypeList, yyvsp[-1], yyvsp[-2]); }
					break;
				case 31: /* decl_specifier_seq: decl_specifier_notypename_seq type_name decl_specifier_notypename_seq */
#line 677 "CParse.y"
					{ Item i;
	          do2(i, BuildTypeList, yyvsp[-3], yyvsp[-2]); 
		  do2(yyval, BuildTypeList, i, yyvsp[-1]); }
					break;
				case 32: /* decl_specifier_seq_opt: */
#line 682 "CParse.y"
					goto yyrule5;
					break;
				/* 33: decl_specifier_seq_opt: decl_specifier_seq */
				/* 34: storage_class_specifier: _KW_AUTO */
				/* 35: storage_class_specifier: _KW_REGISTER */
				/* 36: storage_class_specifier: _KW_STATIC */
				/* 37: storage_class_specifier: _KW_EXTERN */
				/* 38: storage_class_specifier: _KW_MUTABLE */
				/* 39: function_specifier: _KW_INLINE */
				/* 40: function_specifier: _KW_VIRTUAL */
				/* 41: function_specifier: _KW_EXPLICIT */
				/* 42: type_specifier_notypename: simple_type_specifier_notypename */
				/* 43: type_specifier_notypename: class_specifier */
				/* 44: type_specifier_notypename: enum_specifier */
				/* 45: type_specifier_notypename: elaborated_type_specifier */
				/* 46: type_specifier_notypename: cv_qualifier */
				/* 47: simple_type_specifier_notypename: _KW_CHAR */
				/* 48: simple_type_specifier_notypename: _KW_WCHAR */
				/* 49: simple_type_specifier_notypename: _KW_BOOL */
				/* 50: simple_type_specifier_notypename: _KW_SHORT */
				/* 51: simple_type_specifier_notypename: _KW_INT */
				/* 52: simple_type_specifier_notypename: _KW_LONG */
				/* 53: simple_type_specifier_notypename: _KW_SIGNED */
				/* 54: simple_type_specifier_notypename: _KW_UNSIGNED */
				/* 55: simple_type_specifier_notypename: _KW_FLOAT */
				/* 56: simple_type_specifier_notypename: _KW_DOUBLE */
				/* 57: simple_type_specifier_notypename: _KW_VOID */
				/* 58: type_name: class_name */
				/* 59: type_name: enum_name */
				/* 60: type_name: typedef_name */
				case 61: /* elaborated_type_specifier: class_key identifier */
#line 773 "CParse.y"
					{ do3(yyval, AggrRef, yyvsp[-2], yyvsp[-1], _current_scope); }
					break;
				case 62: /* elaborated_type_specifier: _KW_ENUM identifier */
#line 775 "CParse.y"
					{ do3(yyval, EnumRef, yyvsp[-2], yyvsp[-1], _current_scope); }
					break;
				case 63: /* enum_specifier: _KW_ENUM identifier_opt lbrace <64> enumerator_list_opt comma_opt _RBRACE */
#line 788 "CParse.y"
					{ yyval = _enum_type; }
					break;
				case 64: /* <64>: */
#line 2472 "CParse.y"
					{ do3(_enum_type, NewEnum, yyvsp[-3], yyvsp[-2], _current_scope);
		  _enum_val = none; }
					break;
				/* 65: enumerator_list: enumerator_definition */
				/* 66: enumerator_list: enumerator_list _COMMA enumerator_definition */
				case 67: /* enumerator_list_opt: */
#line 800 "CParse.y"
					goto yyrule5;
					break;
				/* 68: enumerator_list_opt: enumerator_list */
				case 69: /* enumerator_definition: identifier */
#line 811 "CParse.y"
					{ Item s, v;
		  do3(s, DeclSym, _enum_type, yyvsp[-1], _current_scope);
		  do1(v, EnumNext, _enum_val);
		  _enum_val = v;
		  do2(yyval, Define, s, v); }
					break;
				case 70: /* enumerator_definition: identifier _OP_ASN constant_expression */
#line 817 "CParse.y"
					{ do3(yyvsp[-2], DeclSym, _enum_type, yyvsp[-3], _current_scope);
		  _enum_val = yyvsp[-1];
		  do2(yyval, Define, yyvsp[-2], yyvsp[-1]); }
					break;
				/* 71: namespace_definition: named_namespace_definition */
				/* 72: namespace_definition: unnamed_namespace_definition */
				/* 73: named_namespace_definition: original_namespace_definition */
				/* 74: named_namespace_definition: extension_namespace_definition */
				case 75: /* original_namespace_definition: _KW_NAMESPACE identifier lbrace <76> namespace_body _RBRACE */
#line 848 "CParse.y"
					{ yyval = _current_scope; _current_scope = yyvsp[-4]; }
					break;
				case 76: /* <76>: */
#line 2472 "CParse.y"
					{ yyval = _current_scope;
		  do3(_current_scope, NewNspace, yyvsp[-3], yyvsp[-2], yyval); }
					break;
				case 77: /* extension_namespace_definition: _KW_NAMESPACE original_namespace_name lbrace <78> namespace_body _RBRACE */
				yyrule77: /* 1 */
#line 858 "CParse.y"
					{ yyval = _current_scope; _current_scope = yyvsp[-3]; }
					break;
				case 78: /* <78>: */
#line 2472 "CParse.y"
					{ yyval = _current_scope; _current_scope = yyvsp[-2]; }
					break;
				case 79: /* unnamed_namespace_definition: _KW_NAMESPACE lbrace <80> namespace_body _RBRACE */
#line 869 "CParse.y"
					goto yyrule77;
					break;
				case 80: /* <80>: */
#line 2472 "CParse.y"
					{ yyval = _current_scope;
		  do3(_current_scope, NewNspace, yyvsp[-2], T(NONE), yyval); }
					break;
				/* 81: namespace_body: declaration_seq_opt */
				case 82: /* namespace_alias_definition: _KW_NAMESPACE identifier _OP_ASN qualifier_namespace_specifier _EOS */
#line 884 "CParse.y"
					{ UNSUPPORTED /* 1 */ }
					break;
				case 83: /* using_declaration: _KW_USING _KW_TYPENAME unqualified_id _EOS */
#line 897 "CParse.y"
					{ UNSUPPORTED /* 2 */ }
					break;
				case 84: /* using_declaration: _KW_USING unqualified_id _EOS */
#line 899 "CParse.y"
					{ UNSUPPORTED /* 3 */ }
					break;
				case 85: /* using_declaration: _KW_USING _SCOPE unqualified_id _EOS */
#line 901 "CParse.y"
					{ UNSUPPORTED /* 4 */ }
					break;
				case 86: /* using_directive: _KW_USING _KW_NAMESPACE namespace_name _EOS */
#line 909 "CParse.y"
					{ UNSUPPORTED /* 5 */ }
					break;
				case 87: /* asm_directive: _KW_ASM _LPAREN _STR _RPAREN _EOS */
#line 917 "CParse.y"
					{ UNSUPPORTED /* 6 */ }
					break;
				/* 88: linkage_specification: _KW_EXTERN _STR lbrace declaration_seq_opt _RBRACE */
				/* 89: linkage_specification: _KW_EXTERN _STR declaration */
				case 90: /* save_scope: */
#line 940 "CParse.y"
					{ yyval = _current_scope; }
					break;
				case 91: /* init_declarator_list1: save_scope init_declarator _COMMA save_scope */
				yyrule91: /* 3 */
#line 945 "CParse.y"
					{ yyval = yyvsp[-1]; }
					break;
				case 92: /* init_declarator_list1: init_declarator_list1 init_declarator _COMMA save_scope */
#line 947 "CParse.y"
					goto yyrule91;
					break;
				case 93: /* init_declarator_list: save_scope init_declarator */
				yyrule93: /* 7 */
#line 952 "CParse.y"
					{ yyval = yyvsp[-2]; }
					break;
				case 94: /* init_declarator_list: init_declarator_list1 init_declarator */
#line 954 "CParse.y"
					goto yyrule93;
					break;
				case 95: /* init_declarator_list_opt: */
#line 957 "CParse.y"
					goto yyrule5;
					break;
				/* 96: init_declarator_list_opt: init_declarator_list */
				case 97: /* init_declarator: declarator attribute_opt */
#line 962 "CParse.y"
					{ eval_decl(yyval, yyvsp[-2], yyvsp[-4]);
		  if(_stmtlist.token != T(NONE)) {
			yyvsp[-2] = yyval; yyvsp[-2].token = T(DECL);
			do3(_stmtlist, BuildStmtList, _stmtlist, yyvsp[-2],
							_current_scope);
		  }
		  _current_scope = yyvsp[-3]; }
					break;
				case 98: /* init_declarator: declarator attribute_opt <99> initializer */
#line 977 "CParse.y"
					{ do2(yyval, Define, yyvsp[-2], yyvsp[-1]); 
		  _current_scope = yyvsp[-5]; 
		  yyval = yyvsp[-2]; }
					break;
				case 99: /* <99>: */
#line 2472 "CParse.y"
					{ eval_decl(yyval, yyvsp[-2], yyvsp[-4]); 
		  if(_stmtlist.token != T(NONE)) {
			yyvsp[-2] = yyval; yyvsp[-2].token = T(DECL);
			do3(_stmtlist, BuildStmtList, _stmtlist, yyvsp[-2],
							_current_scope);
		  } }
					break;
				/* 100: declarator: direct_declarator */
				case 101: /* declarator: ptr_operator declarator */
				yyrule101: /* 3 */
#line 989 "CParse.y"
					{ declPtr(yyval, yyvsp[-2], yyvsp[-1]); }
					break;
				/* 102: direct_declarator: declarator_id */
				case 103: /* direct_declarator: direct_declarator _LPAREN parameter_declaration_clause _RPAREN cv_qualifier_seq_opt */
				yyrule103: /* 2 */
#line 1003 "CParse.y"
					{ declFunc(yyval, yyvsp[-5], yyvsp[-3], yyvsp[-1]); }
					break;
				case 104: /* direct_declarator: direct_declarator _LBRACK constant_expression _RBRACK */
				yyrule104: /* 1 */
#line 1005 "CParse.y"
					{ declArray(yyval, yyvsp[-4], yyvsp[-2]); }
					break;
				case 105: /* direct_declarator: direct_declarator _LBRACK _RBRACK */
#line 1007 "CParse.y"
					{ declArray(yyval, yyvsp[-3], none); }
					break;
				case 106: /* direct_declarator: _LPAREN declarator _RPAREN */
#line 1009 "CParse.y"
					goto yyrule93;
					break;
				case 107: /* ctor_declarator_id: _CTOR_ID */
#line 1014 "CParse.y"
					{ do3(yyvsp[-1], SpecialId, T(CTOR_ID), yyvsp[-1], _current_scope);
		  declSym(yyval, _current_scope, yyvsp[-1]); 
		  if(yyvsp[-1].scope) {
			if(_current_scope.token == T(NONE)) {
				/* XXX - hack */
				_current_scope.token = T(SCOPE);
				_current_scope.tuple.right = yyvsp[-1].scope;
			} else 
				yyerror("scope cannot be specified for"
					" non-global constructor");
		  }
		}
					break;
				case 108: /* ctor_declarator: ctor_declarator_id _LPAREN parameter_declaration_clause _RPAREN cv_qualifier_seq_opt */
#line 1031 "CParse.y"
					goto yyrule103;
					break;
				/* 109: declarator_or_ctor: ctor_declarator */
				/* 110: declarator_or_ctor: declarator */
				case 111: /* ptr_operator: _OP_MUL cv_qualifier_seq_opt */
#line 1046 "CParse.y"
					{ yyval = T(OP_PTR); yyval.val = yyvsp[-1].val; }
					break;
				case 112: /* ptr_operator: _OP_AND */
				yyrule112: /* 1 */
#line 1048 "CParse.y"
					{ yyval = T(OP_REF); }
					break;
				case 113: /* ptr_operator: _SCOPE _OP_MUL cv_qualifier_seq_opt */
#line 1050 "CParse.y"
					{ yyval = T(OP_MPTR); yyval.scope = yyvsp[-3].scope; yyval.val = yyvsp[-1].val; }
					break;
				case 114: /* cv_qualifier_seq: cv_qualifier cv_qualifier_seq_opt */
#line 1058 "CParse.y"
					goto yyrule30;
					break;
				case 115: /* cv_qualifier_seq_opt: */
#line 1061 "CParse.y"
					goto yyrule5;
					break;
				/* 116: cv_qualifier_seq_opt: cv_qualifier_seq */
				/* 117: cv_qualifier: _KW_CONST */
				/* 118: cv_qualifier: _KW_VOLATILE */
				case 119: /* declarator_id: id_expression */
#line 1078 "CParse.y"
					{ 
		  declSym(yyval, _current_scope, yyvsp[-1]); 
		  if(yyvsp[-1].scope) {
			if(_current_scope.token == T(NONE)) {
				/* XXX - hack */
				_current_scope.token = T(SCOPE);
				_current_scope.tuple.right = yyvsp[-1].scope;
			} else 
				yyerror("scope cannot be specified for"
					" non-global declarator");
		  }
		}
					break;
				case 120: /* type_id: type_specifier_seq abstract_declarator_opt */
				yyrule120: /* 2 */
#line 1097 "CParse.y"
					{ eval_decl(yyval, yyvsp[-1], yyvsp[-2]); }
					break;
				case 121: /* type_specifier_notypename_seq: type_specifier_notypename */
#line 1105 "CParse.y"
					goto yyrule25;
					break;
				case 122: /* type_specifier_notypename_seq: type_specifier_notypename_seq type_specifier_notypename */
#line 1107 "CParse.y"
					goto yyrule26;
					break;
				case 123: /* type_specifier_seq: type_name */
#line 1112 "CParse.y"
					goto yyrule25;
					break;
				/* 124: type_specifier_seq: type_specifier_notypename_seq */
				case 125: /* abstract_declarator: ptr_operator abstract_declarator_opt */
#line 1122 "CParse.y"
					goto yyrule101;
					break;
				/* 126: abstract_declarator: direct_abstract_declarator */
				case 127: /* abstract_declarator_opt: */
#line 1126 "CParse.y"
					goto yyrule5;
					break;
				/* 128: abstract_declarator_opt: abstract_declarator */
				case 129: /* direct_abstract_declarator: _LPAREN parameter_declaration_clause _RPAREN cv_qualifier_seq_opt */
#line 1136 "CParse.y"
					{ declFunc(yyval, none, yyvsp[-3], yyvsp[-1]); }
					break;
				case 130: /* direct_abstract_declarator: direct_abstract_declarator _LPAREN parameter_declaration_clause _RPAREN cv_qualifier_seq_opt */
#line 1139 "CParse.y"
					goto yyrule103;
					break;
				case 131: /* direct_abstract_declarator: _LPAREN abstract_declarator _RPAREN */
#line 1141 "CParse.y"
					goto yyrule93;
					break;
				case 132: /* parameter_declaration_clause: */
#line 1149 "CParse.y"
					{ do3(yyval, BuildParamList, none, none, _current_scope); }
					break;
				/* 133: parameter_declaration_clause: parameter_declaration_list */
				case 134: /* parameter_declaration_clause: parameter_declaration_list comma_opt _ELIPSIS */
#line 1152 "CParse.y"
					{ do3(yyval, BuildParamList, yyvsp[-3], T(ELIPSIS), _current_scope); }
					break;
				case 135: /* parameter_declaration_clause: _ELIPSIS */
#line 1154 "CParse.y"
					{ do3(yyval, BuildParamList, none, T(ELIPSIS), _current_scope); }
					break;
				case 136: /* parameter_declaration_list: parameter_declaration */
#line 1163 "CParse.y"
					{ do3(yyval, BuildParamList, none, yyvsp[-1], _current_scope); }
					break;
				case 137: /* parameter_declaration_list: parameter_declaration_list _COMMA parameter_declaration */
#line 1165 "CParse.y"
					{ do3(yyval, BuildParamList, yyvsp[-3], yyvsp[-1], _current_scope); }
					break;
				case 138: /* parameter_declaration: decl_specifier_seq declarator _OP_ASN assignment_expression */
#line 1176 "CParse.y"
					{ eval_pdecl(yyvsp[-2], yyvsp[-3], yyvsp[-4]);
		  do2(yyval, DefParam, yyvsp[-2], yyvsp[-1]); }
					break;
				case 139: /* parameter_declaration: decl_specifier_seq declarator */
#line 1179 "CParse.y"
					{ eval_pdecl(yyval, yyvsp[-1], yyvsp[-2]); }
					break;
				case 140: /* parameter_declaration: decl_specifier_seq abstract_declarator_opt _OP_ASN assignment_expression */
#line 1182 "CParse.y"
					{ eval_pdecl(yyvsp[-2], yyvsp[-3], yyvsp[-4]);
		  do2(yyvsp[-4], DeclParam, yyvsp[-2], none);
		  do2(yyval, DefParam, yyvsp[-4], yyvsp[-1]); }
					break;
				case 141: /* parameter_declaration: decl_specifier_seq abstract_declarator_opt */
#line 1186 "CParse.y"
					{ eval_pdecl(yyval, yyvsp[-1], yyvsp[-2]); yyvsp[-1] = yyval;
		  do2(yyval, DeclParam, yyvsp[-1], none); }
					break;
				case 142: /* eval_declarator_or_ctor: declarator_or_ctor */
#line 1196 "CParse.y"
					{ eval_decl(yyval, yyvsp[-1], yyvsp[-3]); 
			yywarning("%p,%p,%p\n", &yyvsp[-3], &yyvsp[-2], &yyvsp[-1]); }
					break;
				case 143: /* function_definition: decl_specifier_seq_opt save_scope declarator_or_ctor <144> ctor_initializer_opt function_body */
#line 1204 "CParse.y"
					{ if(yyvsp[-2].token == T(TOKENLIST)) {
		  	((TokenList*)yyvsp[-2].ptr)->append_delete((TokenList*)yyvsp[-1].ptr);
			yyvsp[-1] = yyvsp[-2];
		  } else if(yyvsp[-2].token != T(NONE))
			do2(yyvsp[-1], CtorInit, yyvsp[-2], yyvsp[-1]);
		  if(_savecode && !_gsavecode) {
			if(!_deftok)
				_deftok = new TokenList;
			_deftok->append(T(DEFINE));
			yyvsp[-3].scope = _current_scope.scope;
			_deftok->append(yyvsp[-3]);
			_deftok->append_delete((TokenList*)yyvsp[-1].ptr);
			yyval = yyvsp[-3];
		  } else 
		  	do2(yyval, Define, yyvsp[-3], yyvsp[-1]);
		  _current_scope = yyvsp[-5]; }
					break;
				case 144: /* <144>: */
#line 2472 "CParse.y"
					{ eval_decl(yyval, yyvsp[-1], yyvsp[-3]); }
					break;
				case 145: /* function_body: <146> compound_statement */
#line 1232 "CParse.y"
					{ _current_scope = yyvsp[-2]; yyval = yyvsp[-1]; }
					break;
				case 146: /* <146>: */
#line 2472 "CParse.y"
					{ yyval = _current_scope;
		  do2(_current_scope, FuncScope, yyvsp[-2], _current_scope); }
					break;
				/* 147: function_body: save_compound_statement */
				case 148: /* initializer: _OP_ASN initializer_clause */
#line 1242 "CParse.y"
					goto yyrule91;
					break;
				case 149: /* initializer: _LPAREN expression_list _RPAREN */
#line 1244 "CParse.y"
					goto yyrule93;
					break;
				/* 150: initializer_clause: assignment_expression */
				case 151: /* initializer_clause: lbrace initializer_list comma_opt _RBRACE */
#line 1255 "CParse.y"
					{ yyval = yyvsp[-3]; }
					break;
				case 152: /* initializer_clause: lbrace _RBRACE */
#line 1257 "CParse.y"
					{ do2(yyval, BuildArray, none, none); }
					break;
				/* 153: comma_opt: */
				/* 154: comma_opt: _COMMA */
				case 155: /* initializer_list: initializer_clause */
#line 1268 "CParse.y"
					{ do2(yyval, BuildArray, none, yyvsp[-1]); }
					break;
				case 156: /* initializer_list: initializer_list _COMMA initializer_clause */
#line 1270 "CParse.y"
					{ do2(yyval, BuildArray, yyvsp[-3], yyvsp[-1]); }
					break;
				case 157: /* class_specifier: class_key identifier_opt base_clause_opt lbrace <158> member_specification_opt _RBRACE */
#line 1294 "CParse.y"
					{ do2(yyval, EndAggr, _current_scope, yyvsp[-3]);
		  _current_scope = yyvsp[-3];
		  _savecode = yyvsp[-4].bval; }
					break;
				case 158: /* <158>: */
#line 2472 "CParse.y"
					{ yyval = _current_scope;
		  Item s; do3(s, NewAggr, yyvsp[-4], yyvsp[-3], yyval); 
		  if(yyvsp[-2].token != T(NONE))
		  	eval_decl(_current_scope, yyvsp[-2], s)
		  else
			_current_scope = s;
		  yyvsp[-1].bval = _savecode; _savecode = true; }
					break;
				/* 159: class_key: _KW_CLASS */
				/* 160: class_key: _KW_STRUCT */
				/* 161: class_key: _KW_UNION */
				case 162: /* class_key: _KW_DYNAMIC _KW_CLASS */
#line 1315 "CParse.y"
					{ yyval = T(KW_DCLASS); }
					break;
				case 163: /* class_key: _KW_DYNAMIC _KW_VIRTUAL _KW_CLASS */
#line 1317 "CParse.y"
					{ yyval = T(KW_DTYPE); }
					break;
				/* 164: class_key: _KW_DCLASS */
				/* 165: class_key: _KW_DTYPE */
				/* 166: member_specification: member_declaration member_specification_opt */
				/* 167: member_specification: access_specifier colon <168> member_specification_opt */
				case 168: /* <168>: */
#line 2472 "CParse.y"
					{ do2(yyval, SetAccess, _current_scope, yyvsp[-2]); }
					break;
				/* 169: member_specification: class_name colon <170> member_specification_opt */
				case 170: /* <170>: */
#line 2472 "CParse.y"
					{ do3(yyvsp[-2], FindSym, T(AGGR_ID), yyvsp[-2], _current_scope); 
		  if(yyvsp[-2].token != T(NONE)) do2(yyval, SetAccess, _current_scope, yyvsp[-2]); }
					break;
				/* 171: member_specification: _KW_DYNAMIC class_name colon <172> member_specification_opt */
				case 172: /* <172>: */
#line 2472 "CParse.y"
					{ do3(yyvsp[-2], FindSym, T(AGGR_ID), yyvsp[-2], _current_scope);
		  if(yyvsp[-3].token != T(NONE)) do2(yyval, SetAccess, _current_scope, yyvsp[-2]); }
					break;
				case 173: /* member_specification_opt: */
#line 1345 "CParse.y"
					goto yyrule5;
					break;
				/* 174: member_specification_opt: member_specification */
				/* 175: member_declaration: _EOS */
				case 176: /* member_declaration: decl_specifier_seq _EOS */
#line 1357 "CParse.y"
					goto yyrule19;
					break;
				case 177: /* member_declaration: decl_specifier_seq member_declarator_list _EOS */
#line 1359 "CParse.y"
					{ eval_decl(yyval, yyvsp[-2], yyvsp[-3]); }
					break;
				case 178: /* member_declaration: member_declarator_list _EOS */
#line 1361 "CParse.y"
					{ eval_decl(yyval, yyvsp[-2], none); }
					break;
				case 179: /* member_declaration: function_definition */
#line 1363 "CParse.y"
					{ do3(yyval, SetAttr, yyvsp[-1], T(KW_INLINE), none); }
					break;
				/* 180: member_declarator_list: member_declarator */
				case 181: /* member_declarator_list: member_declarator_list _COMMA member_declarator */
				yyrule181: /* 1 */
#line 1376 "CParse.y"
					{ buildDeclList(yyval, yyvsp[-3], yyvsp[-1]); }
					break;
				/* 182: member_declarator: declarator_or_ctor */
				case 183: /* member_declarator: declarator _OP_ASN constant_expression */
#line 1393 "CParse.y"
					{ defineSym(yyval, yyvsp[-3], yyvsp[-1]); }
					break;
				case 184: /* member_declarator: colon constant_expression */
#line 1395 "CParse.y"
					goto yyrule5;
					break;
				case 185: /* member_declarator: identifier colon constant_expression */
#line 1397 "CParse.y"
					{ declSym(yyval, _current_scope, yyvsp[-3]); }
					break;
				case 186: /* base_clause_opt: */
#line 1408 "CParse.y"
					goto yyrule5;
					break;
				case 187: /* base_clause_opt: colon base_specifier_list */
#line 1410 "CParse.y"
					goto yyrule91;
					break;
				/* 188: base_specifier_list: base_specifier */
				case 189: /* base_specifier_list: base_specifier_list _COMMA base_specifier */
#line 1420 "CParse.y"
					goto yyrule181;
					break;
				case 190: /* base_specifier: class_name */
#line 1432 "CParse.y"
					{ declInherit(yyval, none, yyvsp[-1]); }
					break;
				case 191: /* base_specifier: _KW_VIRTUAL class_name */
#line 1434 "CParse.y"
					{ declInherit(yyval, yyvsp[-2], yyvsp[-1]); }
					break;
				case 192: /* base_specifier: _KW_VIRTUAL access_specifier class_name */
#line 1436 "CParse.y"
					{ yyvsp[-2].tval = T(KW_VIRTUAL); declInherit(yyval, yyvsp[-2], yyvsp[-1]); }
					break;
				case 193: /* base_specifier: access_specifier class_name */
#line 1438 "CParse.y"
					{ Item n = none; n.tval = T(KW_VIRTUAL);
		  declInherit(yyval, yyvsp[-2], yyvsp[-1]); }
					break;
				case 194: /* base_specifier: access_specifier _KW_VIRTUAL class_name */
#line 1441 "CParse.y"
					{ yyvsp[-3].tval = T(KW_VIRTUAL); declInherit(yyval, yyvsp[-3], yyvsp[-1]); }
					break;
				/* 195: access_specifier: _KW_PRIVATE */
				/* 196: access_specifier: _KW_PROTECTED */
				/* 197: access_specifier: _KW_PUBLIC */
				case 198: /* conversion_function_id: _KW_OPERATOR conversion_type_id */
#line 1463 "CParse.y"
					{ do3(yyval, SpecialId, T(CONV_ID), yyvsp[-1], _current_scope); }
					break;
				case 199: /* conversion_type_id: type_specifier_seq conversion_declarator_opt */
#line 1471 "CParse.y"
					goto yyrule120;
					break;
				case 200: /* conversion_declarator: ptr_operator conversion_declarator_opt */
#line 1479 "CParse.y"
					goto yyrule101;
					break;
				case 201: /* conversion_declarator_opt: */
#line 1482 "CParse.y"
					goto yyrule5;
					break;
				/* 202: conversion_declarator_opt: conversion_declarator */
				case 203: /* ctor_initializer: _COLON_SAVE <204> save_mem_initializer_list */
#line 1491 "CParse.y"
					{ _rawid = yyvsp[-2].bval;
		  yyvsp[-3].token = T(COLON);
		  ((TokenList*)yyvsp[-1].ptr)->prepend(yyvsp[-3]); 
		  yyval = yyvsp[-1]; }
					break;
				case 204: /* <204>: */
				yyrule204: /* 1 */
#line 2472 "CParse.y"
					{ yyval.bval = _rawid; _rawid = true; }
					break;
				case 205: /* ctor_initializer: _COLON <206> mem_initializer_list */
#line 1499 "CParse.y"
					{ yyval = yyvsp[-1]; _current_scope = yyvsp[-2]; }
					break;
				case 206: /* <206>: */
#line 2472 "CParse.y"
					{ yyval = _current_scope;
		  do2(_current_scope, FuncScope, yyvsp[-2], yyval); }
					break;
				case 207: /* ctor_initializer_opt: */
#line 1502 "CParse.y"
					goto yyrule5;
					break;
				/* 208: ctor_initializer_opt: ctor_initializer */
				/* 209: save_mem_initializer_list: save_mem_initializer */
				case 210: /* save_mem_initializer_list: save_mem_initializer_list _COMMA save_mem_initializer */
#line 1511 "CParse.y"
					{ ((TokenList*)yyvsp[-3].ptr)->append(yyvsp[-2]); 
		  ((TokenList*)yyvsp[-3].ptr)->append_delete((TokenList*)yyvsp[-1].ptr); 
		  yyval = yyvsp[-3]; }
					break;
				case 211: /* mem_initializer_list: mem_initializer */
#line 1518 "CParse.y"
					{ do3(yyval, BuildStmtList, none, yyvsp[-1], _current_scope); }
					break;
				case 212: /* mem_initializer_list: mem_initializer_list _COMMA mem_initializer */
#line 1520 "CParse.y"
					{ do3(yyval, BuildStmtList, yyvsp[-3], yyvsp[-1], _current_scope); }
					break;
				case 213: /* save_mem_initializer: identifier _LPAREN save_statement_seq_opt _RPAREN */
#line 1532 "CParse.y"
					{ yyvsp[-4].token = T(ID); 
		  ((TokenList*)yyvsp[-2].ptr)->prepend(yyvsp[-3]);
		  ((TokenList*)yyvsp[-2].ptr)->prepend(yyvsp[-4]);
		  ((TokenList*)yyvsp[-2].ptr)->append(yyvsp[-1]);
		  yyval = yyvsp[-2]; }
					break;
				case 214: /* mem_initializer: identifier _LPAREN expression_list_opt _RPAREN */
				yyrule214: /* 2 */
#line 1541 "CParse.y"
					{ do2(yyval, CtorCall, yyvsp[-4], yyvsp[-2]); }
					break;
				case 215: /* operator_function_id: _KW_OPERATOR operator */
#line 1553 "CParse.y"
					{ do3(yyval, SpecialId, T(OPER_ID), yyvsp[-1], _current_scope); }
					break;
				/* 216: simple_operator: _OP_NEW */
				/* 217: simple_operator: _OP_DELETE */
				/* 218: simple_operator: _OP_NEWA */
				/* 219: simple_operator: _OP_DELETEA */
				/* 220: simple_operator: _OP_ADD */
				/* 221: simple_operator: _OP_SUB */
				/* 222: simple_operator: _OP_MUL */
				/* 223: simple_operator: _OP_DIV */
				/* 224: simple_operator: _OP_MOD */
				/* 225: simple_operator: _OP_XOR */
				/* 226: simple_operator: _OP_AND */
				/* 227: simple_operator: _OP_OR */
				/* 228: simple_operator: _OP_NOT */
				/* 229: simple_operator: _OP_LNOT */
				/* 230: simple_operator: _OP_ASN */
				/* 231: simple_operator: _OP_LSS */
				/* 232: simple_operator: _OP_GTR */
				/* 233: simple_operator: _OP_ASNADD */
				/* 234: simple_operator: _OP_ASNSUB */
				/* 235: simple_operator: _OP_ASNMUL */
				/* 236: simple_operator: _OP_ASNDIV */
				/* 237: simple_operator: _OP_ASNMOD */
				/* 238: simple_operator: _OP_ASNXOR */
				/* 239: simple_operator: _OP_ASNAND */
				/* 240: simple_operator: _OP_ASNOR */
				/* 241: simple_operator: _OP_LSHIFT */
				/* 242: simple_operator: _OP_RSHIFT */
				/* 243: simple_operator: _OP_ASNRSHIFT */
				/* 244: simple_operator: _OP_ASNLSHIFT */
				/* 245: simple_operator: _OP_EQU */
				/* 246: simple_operator: _OP_NEQ */
				/* 247: simple_operator: _OP_LEQ */
				/* 248: simple_operator: _OP_GEQ */
				/* 249: simple_operator: _OP_LAND */
				/* 250: simple_operator: _OP_LOR */
				/* 251: simple_operator: _OP_INC */
				/* 252: simple_operator: _OP_DEC */
				/* 253: simple_operator: _COMMA */
				/* 254: simple_operator: _OP_MPTRTO */
				/* 255: simple_operator: _OP_PTRTO */
				/* 256: operator: simple_operator */
				case 257: /* operator: _LPAREN _RPAREN */
#line 1576 "CParse.y"
					{ yyval = T(OP_FUNC); }
					break;
				case 258: /* operator: _LBRACK _RBRACK */
#line 1578 "CParse.y"
					{ yyval = T(OP_INDEX); }
					break;
				/* 259: literal: _NUM */
				/* 260: literal: _CHAR */
				/* 261: literal: _STR */
				/* 262: literal: _FLOAT */
				/* 263: literal: _BOOL */
				case 264: /* primary_expression: literal */
				yyrule264: /* 1 */
#line 1679 "CParse.y"
					{ do1(yyval, Literal, yyvsp[-1]); }
					break;
				case 265: /* primary_expression: _KW_THIS */
#line 1681 "CParse.y"
					{ do3(yyval, SpecialId, yyvsp[-1], yyvsp[-1], _current_scope); }
					break;
				case 266: /* primary_expression: _LPAREN expression _RPAREN */
#line 1683 "CParse.y"
					goto yyrule93;
					break;
				/* 267: primary_expression: id_expression_val */
				/* 268: id_expression: unqualified_id */
				/* 269: id_expression_val: unqualified_id_val */
				/* 270: unqualified_id: identifier */
				/* 271: unqualified_id: operator_function_id */
				/* 272: unqualified_id: conversion_function_id */
				case 273: /* unqualified_id: _DTOR_ID */
				yyrule273: /* 2 */
#line 1712 "CParse.y"
					{ do3(yyval, SpecialId, T(DTOR_ID), yyvsp[-1], _current_scope); }
					break;
				/* 274: unqualified_id_val: _VAL_ID */
				case 275: /* unqualified_id_val: _ID */
#line 1719 "CParse.y"
					{ do2(yyval, Undeclared, yyvsp[-1], _current_scope); }
					break;
				/* 276: unqualified_id_val: operator_function_id */
				/* 277: unqualified_id_val: conversion_function_id */
				case 278: /* unqualified_id_val: _DTOR_ID */
#line 1723 "CParse.y"
					goto yyrule273;
					break;
				/* 279: postfix_expression: primary_expression */
				case 280: /* postfix_expression: postfix_expression _LBRACK expression _RBRACK */
#line 1761 "CParse.y"
					{ do2(yyval, Index, yyvsp[-4], yyvsp[-2]); }
					break;
				case 281: /* postfix_expression: postfix_expression _LPAREN expression_list_opt _RPAREN */
#line 1763 "CParse.y"
					{ do2(yyval, Call, yyvsp[-4], yyvsp[-2]); }
					break;
				case 282: /* postfix_expression: simple_type_specifier_notypename _LPAREN expression_list_opt _RPAREN */
#line 1765 "CParse.y"
					goto yyrule214;
					break;
				case 283: /* postfix_expression: type_name _LPAREN expression_list_opt _RPAREN */
#line 1767 "CParse.y"
					goto yyrule214;
					break;
				case 284: /* postfix_expression: postfix_expression _OP_DOT id_expression */
				yyrule284: /* 24 */
#line 1769 "CParse.y"
					{ do3(yyval, BinaryOp, yyvsp[-2], yyvsp[-3], yyvsp[-1]); }
					break;
				case 285: /* postfix_expression: postfix_expression _OP_PTRTO id_expression */
#line 1771 "CParse.y"
					goto yyrule284;
					break;
				case 286: /* postfix_expression: postfix_expression _OP_DOT pseudo_destructor_name */
#line 1773 "CParse.y"
					goto yyrule284;
					break;
				case 287: /* postfix_expression: postfix_expression _OP_PTRTO pseudo_destructor_name */
#line 1775 "CParse.y"
					goto yyrule284;
					break;
				case 288: /* postfix_expression: postfix_expression _OP_INC */
#line 1777 "CParse.y"
					{ do2(yyval, UnaryOp, T(OP_INCP), yyvsp[-2]); }
					break;
				case 289: /* postfix_expression: postfix_expression _OP_DEC */
#line 1779 "CParse.y"
					{ do2(yyval, UnaryOp, T(OP_DECP), yyvsp[-2]); }
					break;
				case 290: /* postfix_expression: newcast _OP_LSS type_id _OP_GTR _LPAREN expression _RPAREN */
#line 1781 "CParse.y"
					{ do3(yyval, Cast, yyvsp[-2], yyvsp[-5], yyvsp[-7]); }
					break;
				case 291: /* postfix_expression: _KW_TYPEID _LPAREN expression _RPAREN */
#line 1783 "CParse.y"
					{ UNSUPPORTED /* 8 */ }
					break;
				case 292: /* postfix_expression: _KW_TYPEID _LPAREN type_id _RPAREN */
#line 1785 "CParse.y"
					{ UNSUPPORTED /* 9 */ }
					break;
				/* 293: newcast: _KW_DCAST */
				/* 294: newcast: _KW_SCAST */
				/* 295: newcast: _KW_RCAST */
				/* 296: newcast: _KW_CCAST */
				case 297: /* expression_list: assignment_expression */
#line 1801 "CParse.y"
					{ do2(yyval, BuildExprList, none, yyvsp[-1]); }
					break;
				case 298: /* expression_list: expression_list _COMMA assignment_expression */
#line 1803 "CParse.y"
					{ do2(yyval, BuildExprList, yyvsp[-3], yyvsp[-1]); }
					break;
				case 299: /* expression_list_opt: */
#line 1806 "CParse.y"
					{ do2(yyval, BuildExprList, none, none); }
					break;
				/* 300: expression_list_opt: expression_list */
				case 301: /* pseudo_destructor_name: _DTOR_ID */
#line 1814 "CParse.y"
					goto yyrule273;
					break;
				/* 302: unary_expression: postfix_expression */
				case 303: /* unary_expression: _OP_INC cast_expression */
				yyrule303: /* 5 */
#line 1834 "CParse.y"
					{ do2(yyval, UnaryOp, yyvsp[-2], yyvsp[-1]); }
					break;
				case 304: /* unary_expression: _OP_DEC cast_expression */
#line 1836 "CParse.y"
					goto yyrule303;
					break;
				case 305: /* unary_expression: unary_operator cast_expression */
#line 1838 "CParse.y"
					goto yyrule303;
					break;
				case 306: /* unary_expression: _OP_SIZEOF unary_expression */
#line 1840 "CParse.y"
					goto yyrule303;
					break;
				case 307: /* unary_expression: _OP_SIZEOF _LPAREN type_id _RPAREN */
#line 1842 "CParse.y"
					{ do2(yyval, UnaryOp, T(OP_TSIZEOF), yyvsp[-2]); }
					break;
				/* 308: unary_expression: new_expression */
				/* 309: unary_expression: delete_expression */
				case 310: /* unary_operator: _OP_MUL */
#line 1851 "CParse.y"
					{ yyval = T(OP_PTR); }
					break;
				case 311: /* unary_operator: _OP_AND */
#line 1852 "CParse.y"
					goto yyrule112;
					break;
				case 312: /* unary_operator: _OP_ADD */
#line 1853 "CParse.y"
					{ yyval = T(OP_UPLUS); }
					break;
				case 313: /* unary_operator: _OP_SUB */
#line 1854 "CParse.y"
					{ yyval = T(OP_UMINUS); }
					break;
				/* 314: unary_operator: _OP_LNOT */
				/* 315: unary_operator: _OP_NOT */
				case 316: /* new_expression: _OP_NEW new_placement_opt new_type_id new_initializer_opt */
#line 1865 "CParse.y"
					{ do3(yyval, OpNew, yyvsp[-2], yyvsp[-1], yyvsp[-3]); }
					break;
				case 317: /* new_expression: _OP_NEW new_placement_opt _LPAREN type_id _RPAREN new_initializer_opt */
#line 1867 "CParse.y"
					{ do3(yyval, OpNew, yyvsp[-3], yyvsp[-1], yyvsp[-5]); }
					break;
				case 318: /* new_placement: _LPAREN expression_list _RPAREN */
#line 1875 "CParse.y"
					goto yyrule93;
					break;
				case 319: /* new_placement_opt: */
#line 1878 "CParse.y"
					goto yyrule5;
					break;
				/* 320: new_placement_opt: new_placement */
				case 321: /* new_type_id: type_specifier_seq new_declarator_opt */
#line 1885 "CParse.y"
					goto yyrule120;
					break;
				case 322: /* new_declarator: ptr_operator new_declarator_opt */
#line 1894 "CParse.y"
					goto yyrule101;
					break;
				/* 323: new_declarator: direct_new_declarator */
				case 324: /* new_declarator_opt: */
#line 1898 "CParse.y"
					goto yyrule5;
					break;
				/* 325: new_declarator_opt: new_declarator */
				case 326: /* direct_new_declarator: _LBRACK expression _RBRACK */
#line 1906 "CParse.y"
					{ declArray(yyval, none, yyvsp[-2]); }
					break;
				case 327: /* direct_new_declarator: direct_new_declarator _LBRACK constant_expression _RBRACK */
#line 1908 "CParse.y"
					goto yyrule104;
					break;
				case 328: /* new_initializer: _LPAREN expression_list_opt _RPAREN */
#line 1916 "CParse.y"
					goto yyrule93;
					break;
				case 329: /* new_initializer_opt: */
#line 1919 "CParse.y"
					goto yyrule5;
					break;
				/* 330: new_initializer_opt: new_initializer */
				case 331: /* delete_expression: _OP_DELETE cast_expression */
#line 1928 "CParse.y"
					goto yyrule303;
					break;
				case 332: /* delete_expression: _OP_DELETEA cast_expression */
#line 1930 "CParse.y"
					goto yyrule303;
					break;
				/* 333: cast_expression: unary_expression */
				case 334: /* cast_expression: _LPAREN type_id _RPAREN cast_expression */
#line 1940 "CParse.y"
					{ do3(yyval, Cast, yyvsp[-1], yyvsp[-3], none); }
					break;
				/* 335: pm_expression: cast_expression */
				case 336: /* pm_expression: pm_expression _OP_MDOT cast_expression */
#line 1951 "CParse.y"
					goto yyrule284;
					break;
				case 337: /* pm_expression: pm_expression _OP_MPTRTO cast_expression */
#line 1953 "CParse.y"
					goto yyrule284;
					break;
				/* 338: multiplicative_expression: pm_expression */
				case 339: /* multiplicative_expression: multiplicative_expression _OP_MUL pm_expression */
#line 1965 "CParse.y"
					goto yyrule284;
					break;
				case 340: /* multiplicative_expression: multiplicative_expression _OP_DIV pm_expression */
#line 1967 "CParse.y"
					goto yyrule284;
					break;
				case 341: /* multiplicative_expression: multiplicative_expression _OP_MOD pm_expression */
#line 1969 "CParse.y"
					goto yyrule284;
					break;
				/* 342: additive_expression: multiplicative_expression */
				case 343: /* additive_expression: additive_expression _OP_ADD multiplicative_expression */
#line 1980 "CParse.y"
					goto yyrule284;
					break;
				case 344: /* additive_expression: additive_expression _OP_SUB multiplicative_expression */
#line 1982 "CParse.y"
					goto yyrule284;
					break;
				/* 345: shift_expression: additive_expression */
				case 346: /* shift_expression: shift_expression _OP_LSHIFT additive_expression */
#line 1993 "CParse.y"
					goto yyrule284;
					break;
				case 347: /* shift_expression: shift_expression _OP_RSHIFT additive_expression */
#line 1995 "CParse.y"
					goto yyrule284;
					break;
				/* 348: relational_expression: shift_expression */
				case 349: /* relational_expression: relational_expression _OP_LSS shift_expression */
#line 2008 "CParse.y"
					goto yyrule284;
					break;
				case 350: /* relational_expression: relational_expression _OP_GTR shift_expression */
#line 2010 "CParse.y"
					goto yyrule284;
					break;
				case 351: /* relational_expression: relational_expression _OP_LEQ shift_expression */
#line 2012 "CParse.y"
					goto yyrule284;
					break;
				case 352: /* relational_expression: relational_expression _OP_GEQ shift_expression */
#line 2014 "CParse.y"
					goto yyrule284;
					break;
				/* 353: equality_expression: relational_expression */
				case 354: /* equality_expression: equality_expression _OP_EQU relational_expression */
#line 2025 "CParse.y"
					goto yyrule284;
					break;
				case 355: /* equality_expression: equality_expression _OP_NEQ relational_expression */
#line 2027 "CParse.y"
					goto yyrule284;
					break;
				/* 356: and_expression: equality_expression */
				case 357: /* and_expression: and_expression _OP_AND equality_expression */
#line 2037 "CParse.y"
					goto yyrule284;
					break;
				/* 358: exclusive_or_expression: and_expression */
				case 359: /* exclusive_or_expression: exclusive_or_expression _OP_XOR and_expression */
#line 2048 "CParse.y"
					goto yyrule284;
					break;
				/* 360: inclusive_or_expression: exclusive_or_expression */
				case 361: /* inclusive_or_expression: inclusive_or_expression _OP_OR exclusive_or_expression */
#line 2058 "CParse.y"
					goto yyrule284;
					break;
				/* 362: logical_and_expression: inclusive_or_expression */
				case 363: /* logical_and_expression: logical_and_expression _OP_LAND inclusive_or_expression */
#line 2068 "CParse.y"
					goto yyrule284;
					break;
				/* 364: logical_or_expression: logical_and_expression */
				case 365: /* logical_or_expression: logical_or_expression _OP_LOR logical_and_expression */
#line 2078 "CParse.y"
					goto yyrule284;
					break;
				/* 366: conditional_expression: logical_or_expression */
				case 367: /* conditional_expression: logical_or_expression _QMARK expression colon assignment_expression */
#line 2088 "CParse.y"
					{ do3(yyval, CondExpr, yyvsp[-5], yyvsp[-3], yyvsp[-1]); }
					break;
				/* 368: assignment_expression: conditional_expression */
				case 369: /* assignment_expression: logical_or_expression assignment_operator assignment_expression */
#line 2099 "CParse.y"
					goto yyrule284;
					break;
				/* 370: assignment_operator: _OP_ASN */
				/* 371: assignment_operator: _OP_ASNMUL */
				/* 372: assignment_operator: _OP_ASNDIV */
				/* 373: assignment_operator: _OP_ASNMOD */
				/* 374: assignment_operator: _OP_ASNADD */
				/* 375: assignment_operator: _OP_ASNSUB */
				/* 376: assignment_operator: _OP_ASNRSHIFT */
				/* 377: assignment_operator: _OP_ASNLSHIFT */
				/* 378: assignment_operator: _OP_ASNAND */
				/* 379: assignment_operator: _OP_ASNXOR */
				/* 380: assignment_operator: _OP_ASNOR */
				/* 381: expression: assignment_expression */
				case 382: /* expression: expression _COMMA assignment_expression */
#line 2127 "CParse.y"
					{ do3(yyval, BinaryOp, T(OP_COMMA), yyvsp[-3], yyvsp[-1]); }
					break;
				case 383: /* expression_opt: */
#line 2130 "CParse.y"
					goto yyrule5;
					break;
				/* 384: expression_opt: expression */
				case 385: /* constant_expression: conditional_expression */
#line 2137 "CParse.y"
					goto yyrule264;
					break;
				/* 386: statement: matched_statement */
				/* 387: statement: unmatched_statement */
				/* 388: other_statement: labeled_statement */
				/* 389: other_statement: expression_statement */
				/* 390: other_statement: compound_statement */
				/* 391: other_statement: selection_statement */
				/* 392: other_statement: iteration_statement */
				/* 393: other_statement: jump_statement */
				/* 394: other_statement: declaration_statement */
				/* 395: matched_statement: other_statement */
				case 396: /* matched_statement: _KW_IF _LPAREN condition _RPAREN matched_statement _KW_ELSE matched_statement */
				yyrule396: /* 1 */
#line 2181 "CParse.y"
					{ do3(yyval, IfElse, yyvsp[-5], yyvsp[-3], yyvsp[-1]); }
					break;
				case 397: /* unmatched_statement: _KW_IF _LPAREN condition _RPAREN statement */
#line 2186 "CParse.y"
					{ do3(yyval, IfElse, yyvsp[-3], yyvsp[-1], none); }
					break;
				case 398: /* unmatched_statement: _KW_IF _LPAREN condition _RPAREN matched_statement _KW_ELSE unmatched_statement */
#line 2188 "CParse.y"
					goto yyrule396;
					break;
				case 399: /* selection_statement: _KW_SWITCH _LPAREN condition _RPAREN statement */
#line 2193 "CParse.y"
					{ do2(yyval, Switch, yyvsp[-3], yyvsp[-1]); }
					break;
				case 400: /* labeled_statement: identifier colon statement */
#line 2203 "CParse.y"
					{ NOSCOPE(yyvsp[-3]); do3(yyval, Label, none, yyvsp[-3], yyvsp[-1]); }
					break;
				case 401: /* labeled_statement: _KW_CASE constant_expression colon statement */
#line 2205 "CParse.y"
					{ do3(yyval, Label, yyvsp[-4], yyvsp[-3], yyvsp[-1]); }
					break;
				case 402: /* labeled_statement: _KW_DEFAULT colon statement */
#line 2207 "CParse.y"
					{ do3(yyval, Label, yyvsp[-3], none, yyvsp[-1]); }
					break;
				/* 403: expression_statement: expression_opt _EOS */
				case 404: /* compound_statement: _LBRACE <405> statement_seq_opt _RBRACE */
#line 2227 "CParse.y"
					{ yyval = _stmtlist; _stmtlist = yyvsp[-4]; 
		  _current_scope = yyvsp[-3]; }
					break;
				case 405: /* <405>: */
#line 2472 "CParse.y"
					{ yyval = _current_scope;
		  do3(_current_scope, NewLocal, none, none, yyval); 
		  yyvsp[-1] = _stmtlist;
		  do3(_stmtlist, BuildStmtList, none, none, _current_scope); }
					break;
				/* 406: statement_seq_opt: */
				case 407: /* statement_seq_opt: statement_seq_opt statement */
#line 2237 "CParse.y"
					{ do3(_stmtlist, BuildStmtList, _stmtlist, yyvsp[-1],
							_current_scope); }
					break;
				/* 408: condition: expression */
				case 409: /* condition: type_specifier_seq declarator _OP_ASN assignment_expression */
#line 2248 "CParse.y"
					{ defineSym(yyvsp[-2], yyvsp[-3], yyvsp[-1]); 
		  eval_decl(yyval, yyvsp[-2], yyvsp[-4]); }
					break;
				case 410: /* condition_opt: */
#line 2252 "CParse.y"
					goto yyrule5;
					break;
				/* 411: condition_opt: condition */
				case 412: /* iteration_statement: _KW_WHILE _LPAREN condition _RPAREN statement */
#line 2261 "CParse.y"
					{ do3(yyval, While, yyvsp[-3], yyvsp[-1], none); }
					break;
				case 413: /* iteration_statement: _KW_DO statement _KW_WHILE _LPAREN expression _RPAREN _EOS */
#line 2263 "CParse.y"
					{ do3(yyval, DoWhile, none, yyvsp[-6], yyvsp[-3]); }
					break;
				case 414: /* iteration_statement: _KW_FOR _LPAREN for_init_statement condition_opt _EOS expression_opt _RPAREN statement */
#line 2266 "CParse.y"
					{ do3(yyvsp[-8], ForExpr, yyvsp[-6], yyvsp[-5], yyvsp[-3]);
		  do2(yyval, For, yyvsp[-8], yyvsp[-1]); }
					break;
				/* 415: for_init_statement: expression_statement */
				/* 416: for_init_statement: simple_declaration */
				case 417: /* jump_statement: _KW_BREAK _EOS */
#line 2287 "CParse.y"
					{ do1(yyval, Break, yyvsp[-2]); }
					break;
				case 418: /* jump_statement: _KW_CONTINUE _EOS */
#line 2289 "CParse.y"
					{ do1(yyval, Continue, yyvsp[-2]); }
					break;
				case 419: /* jump_statement: _KW_RETURN expression_opt _EOS */
#line 2291 "CParse.y"
					{ do1(yyval, Return, yyvsp[-2]); }
					break;
				case 420: /* jump_statement: _KW_GOTO identifier _EOS */
#line 2293 "CParse.y"
					{ NOSCOPE(yyvsp[-2]); do1(yyval, Goto, yyvsp[-2]); }
					break;
				case 421: /* declaration_statement: block_declaration */
#line 2301 "CParse.y"
					goto yyrule5;
					break;
				/* 422: typedef_name: _TYPE_ID */
				/* 423: namespace_name: original_namespace_name */
				/* 424: namespace_name: namespace_alias */
				/* 425: original_namespace_name: _NSPACE_ID */
				/* 426: namespace_alias: _ASPACE_ID */
				/* 427: class_name: _AGGR_ID */
				/* 428: enum_name: _ENUM_ID */
				/* 429: template_name: _TPLATE_ID */
				/* 430: identifier: _ID */
				/* 431: identifier: _VAL_ID */
				/* 432: identifier: _TYPE_ID */
				/* 433: identifier: _ENUM_ID */
				case 434: /* identifier: _AGGR_ID */
#line 2370 "CParse.y"
					{ yyval.token = T(AGGR_ID); }
					break;
				/* 435: identifier: _NSPACE_ID */
				/* 436: identifier: _ASPACE_ID */
				/* 437: identifier: _TPLATE_ID */
				case 438: /* identifier_opt: */
#line 2376 "CParse.y"
					goto yyrule5;
					break;
				/* 439: identifier_opt: identifier */
				/* 440: lbrace: _LBRACE */
				case 441: /* lbrace: _LBRACE_SAVE */
#line 2382 "CParse.y"
					{ yyval.token = T(LBRACE); }
					break;
				/* 442: colon: _COLON */
				case 443: /* colon: _COLON_SAVE */
#line 2384 "CParse.y"
					{ yyval.token = T(COLON); }
					break;
				/* 444: ignore_savecode: */
				/* 445: ignore_savecode: _TOKENLIST */
				case 446: /* save_compound_statement: <447> _LBRACE_SAVE save_statement_seq_opt _RBRACE */
#line 2391 "CParse.y"
					{ _rawid = yyvsp[-4].bval;
		  yyvsp[-3].token = T(LBRACE);
		  ((TokenList*)yyvsp[-2].ptr)->prepend(yyvsp[-3]); 
		  ((TokenList*)yyvsp[-2].ptr)->append(yyvsp[-1]);
		  yyval = yyvsp[-2]; }
					break;
				case 447: /* <447>: */
#line 2472 "CParse.y"
					goto yyrule204;
					break;
				case 448: /* save_statement_seq: save_statement_seq_opt any_statement_token */
#line 2400 "CParse.y"
					{ ((TokenList*)yyval.ptr)->append(yyvsp[-1]); }
					break;
				case 449: /* save_statement_seq: save_statement_seq_opt save_statement_list */
#line 2402 "CParse.y"
					{ ((TokenList*)yyval.ptr)->append_delete((TokenList*)yyvsp[-1].ptr); }
					break;
				case 450: /* save_statement_seq_opt: */
#line 2406 "CParse.y"
					{ yyval.token = T(TOKENLIST); yyval.ptr = new TokenList; }
					break;
				/* 451: save_statement_seq_opt: save_statement_seq */
				case 452: /* save_statement_list: lbrace save_statement_seq_opt _RBRACE */
				yyrule452: /* 2 */
#line 2412 "CParse.y"
					{ ((TokenList*)yyvsp[-2].ptr)->prepend(yyvsp[-3]); 
		  ((TokenList*)yyvsp[-2].ptr)->append(yyvsp[-1]);
		  yyval = yyvsp[-2]; }
					break;
				case 453: /* save_statement_list: _LPAREN save_statement_seq_opt _RPAREN */
#line 2416 "CParse.y"
					goto yyrule452;
					break;
				case 454: /* save_statement_list: _LBRACK save_statement_seq_opt _RBRACK */
#line 2420 "CParse.y"
					goto yyrule452;
					break;
				/* 455: any_statement_token: _NUM */
				/* 456: any_statement_token: _CHAR */
				/* 457: any_statement_token: _STR */
				/* 458: any_statement_token: _FLOAT */
				/* 459: any_statement_token: _BOOL */
				/* 460: any_statement_token: _EOS */
				/* 461: any_statement_token: colon */
				/* 462: any_statement_token: _QMARK */
				/* 463: any_statement_token: simple_operator */
				/* 464: any_statement_token: _OP_DOT */
				/* 465: any_statement_token: _OP_MDOT */
				/* 466: any_statement_token: _OP_SCOPE */
				/* 467: any_statement_token: _ELIPSIS */
				/* 468: any_statement_token: cv_qualifier */
				/* 469: any_statement_token: simple_type_specifier_notypename */
				/* 470: any_statement_token: storage_class_specifier */
				/* 471: any_statement_token: _KW_IF */
				/* 472: any_statement_token: _KW_ELSE */
				/* 473: any_statement_token: _KW_DO */
				/* 474: any_statement_token: _KW_WHILE */
				/* 475: any_statement_token: _KW_FOR */
				/* 476: any_statement_token: _KW_SWITCH */
				/* 477: any_statement_token: _KW_CASE */
				/* 478: any_statement_token: _KW_DEFAULT */
				/* 479: any_statement_token: _KW_BREAK */
				/* 480: any_statement_token: _KW_CONTINUE */
				/* 481: any_statement_token: _KW_GOTO */
				/* 482: any_statement_token: _KW_RETURN */
				/* 483: any_statement_token: _KW_ENUM */
				/* 484: any_statement_token: _KW_TYPEDEF */
				/* 485: any_statement_token: class_key */
				/* 486: any_statement_token: function_specifier */
				/* 487: any_statement_token: _OP_SIZEOF */
				/* 488: any_statement_token: _OP_TYPEOF */
				/* 489: any_statement_token: _KW_ASM */
				/* 490: any_statement_token: _KW_OPERATOR */
				/* 491: any_statement_token: _KW_THIS */
				/* 492: any_statement_token: _KW_TYPEID */
				/* 493: any_statement_token: _KW_TYPENAME */
				/* 494: any_statement_token: _KW_NAMESPACE */
				/* 495: any_statement_token: _KW_USING */
				/* 496: any_statement_token: _KW_TEMPLATE */
				/* 497: any_statement_token: access_specifier */
				/* 498: any_statement_token: _KW_FRIEND */
				/* 499: any_statement_token: _KW_THROW */
				/* 500: any_statement_token: _KW_CATCH */
				/* 501: any_statement_token: _KW_TRY */
				/* 502: any_statement_token: newcast */
				case 503: /* any_statement_token: identifier */
#line 2452 "CParse.y"
					{ yyval.token = T(ID); }
					break;
				/* 504: any_statement_token: _DTOR_ID */
				case 505: /* attribute: _KW_ATTRIBUTE _LPAREN _LPAREN save_statement_seq _RPAREN _RPAREN */
#line 2467 "CParse.y"
					{ /* just ignore it for now */ }
					break;
				case 506: /* attribute_opt: */
#line 2470 "CParse.y"
					goto yyrule5;
					break;
				/* 507: attribute_opt: attribute */
#line 808 "/u4/shaggy/u2/src/sos/util/shacc/shacc.c.parser"
				}
				*nvsp = yyval;
				yyvsp = nvsp+1;
				sp = yy.rstk[sp].p;
			}
			if(!pr) {
				while(yy.dsrp < yy.dsp)
					*yyvsp++ = yy.dvsa[yy.dsrp++];
				if(yyvsp-yy.vsa >= YYINITDEPTH) {
					YYERROR("vsa overflow");
					return 0;
				}
				DBG(YYEPRINTF("clearing rstk\n"));
				yy.rsp = 0; 
				yy.dsp = 0;
				yy.dsrp = 0;
			}
			yy.vsp = yyvsp;
		}

		/* reduce all reductions and substitute them back into
		 *	their previous rules.  Also follow all substitution
		 *	links for additional substitutions.
		 *	All reductions get added as reduction chain links.
		 */
		np = yy.rnp;
		yy.rnp = 0;
		ambig = 0;
		if(np) {
		    while(np) {
			YYNode *wnp;
			const YYNode *nnp = np->a;
			const YYNode *pp = np->p;
			const YYNode *sp;
			int nt;
			int srsp = yy.rsp;
			(CCAST(YYNode*)(np))->a = 0;
#ifdef YYPDEBUG
			YYEPRINTF("reducing: ");
			yyprintfullnode(&yy, np);
#endif
			if(yyrule[np->r].uniq && (yy.anp || yy.rnp || nnp)) 
				ambig = 1;
			if(!pp) {
#if YYDEBUG
				if(YYDEBUGFLAG) {
					YYEPRINTF("yydebug: success:\n");
					yyprintactiverules(&yy, np);
				}
#endif
				success++;
			} else {
				int c1;
				nt = yysym[yyrule[pp->r].sym + pp->x];
				if(!yyrule[np->r].hasact
						&& yyrule[np->r].ns == 1) {
					srsp = np->c;
				} else {
					if(yy.rsp+1 > maxrsp)
					    if((maxrsp = yy.rsp+1)
								> rstksize) {
						YYERROR("rstk overflow");
						return 0;
					    }
					yy.rstk[yy.rsp].r = np->r;
					yy.rstk[yy.rsp].sp = yy.dsp;
					yy.rstk[yy.rsp].p = np->c;
#if YYDEBUG
					yy.rstk[yy.rsp].nt = nt;
#endif
					yy.rsp++;
				}
				wnp = yynode_canwrite(&yy, pp);
				++wnp->x;
				wnp->c = srsp;
				if((c1 = np->c1) == -1)
					c1 = srsp;
				wnp->c1 = c1;
				yynode_assign(&yy, &wnp->a, 0);
#ifdef YYPDEBUG
				YYEPRINTF("to: ");
				yyprintfullnode(&yy, wnp);
#endif
				yynode_copy(&yy, pp);
				sp = wnp->s;
				yynode_add(&yy, wnp);
				while(sp) {
					wnp = yynode_canwrite(&yy, sp);
					wnp->c = srsp;
					wnp->c1 = c1;
					++wnp->x;
#ifdef YYPDEBUG
					YYEPRINTF("also to: ");
					yyprintfullnode(&yy, wnp);
#endif
					yynode_copy(&yy, sp);
					sp = wnp->s;
					yynode_add(&yy, wnp);
				}
			}
			yynode_delete(&yy, np);
			np = nnp;
		    }

		    /* attempt to resolve any ambiguities 
		     * via reduction precedence rules
		     */
		    if(ambig) {
			int d = 0;
			for(;;) {
				int c, cc;
				int r, cr;
				int rp, crp, mrp;
				const YYNode *snp;
				const YYNode **pnp;
				const YYNode **psnp;
				np = yy.anp ? yy.anp : yy.rnp;
				snp = yy.anp ? yy.rnp : 0;
				c = -3;
				r = -3;
				rp = 0;
				mrp = 0;
				while(np) {
					int i;
					cc = np->c;
					for(i=0; i<d; i++)
						cc = yy.rstk[cc].p;
					if(c == -3)
						c = cc;
					else if(cc != c) 
						c = -2;
					cr = yy.rstk[cc].r;
					if(r == -3)
						r = cr;
					else if(cr != r)
						r = -2;
					if((crp = yyrule[cr].rrp)) {
						if(crp > mrp)
							mrp = crp;
						if(!rp || crp < rp)
							rp = crp;
					} else
						rp = -1;
					if(!(np = np->a)) {
						np = snp;
						snp = 0;
					}
				}
				if(c > -2) {
					ambig = 0;
					break;
				}
				if(r > -2) {
					++d;
					continue;
				}
				if(rp < 1 || mrp <= rp)
					break;
				pnp = yy.anp ? &yy.anp : &yy.rnp;
				psnp = yy.anp ? &yy.rnp : 0;
				while(pnp && (np = *pnp)) {
					int i;
					cc = np->c;
					for(i=0; i<d; i++)
						cc = yy.rstk[cc].p;
					cr = yy.rstk[cc].r;
					if(yyrule[cr].rrp != rp) {
						*pnp = np->a;
						yynode_delete(&yy, np);
					} else
						pnp = CCAST(const YYNode**)(&np->a);
					if(!*pnp) {
						pnp = psnp;
						psnp = 0;
					}
				}
			}
			if(ambig) {
				YYERROR("ambiguous grammar");
#if YYDEBUG
				if(YYDEBUGFLAG) {
					yyprintactiverules(&yy, yy.rnp);
					yyprintactiverules(&yy, yy.anp);
				}
				return -1;
			}
#endif
		    }
		    continue;	
		}

		if(yy.rnp)
			YYINTERR;

		/* Follow all paths as far as possible,
		 *	seeing if empty rules perhaps help us to reduce
		 */
		np = yy.anp;
		yy.anp = 0;
		while(np) {
			const YYNode *nnp = np->a;
			(CCAST(YYNode*)(np))->a = 0;
			yyfollow(&yy, np, -1);
			np = nnp;
		}
		if(yy.rnp)		/* if we got any reductions */
			continue;	/* go ahead and use them */

		/* Get a new token */
		s = eof ? 0 : YYLEXCALL;
		eof = (s == 0);
#if YYDEBUG
		if(YYDEBUGFLAG)
			YYEPRINTF("yydebug: read token: %d (%s)\n",
					s, yysname[s+YYNNONTERMS-1]);
#endif

		/* Follow all possible paths with this token,
		 *	eliminating ones that no longer match
		 */
		np = yy.anp;
		yy.anp = 0;
		while(np) {
			const YYNode *nnp = np->a;
			(CCAST(YYNode*)(np))->a = 0;
			yyfollow(&yy, np, s);
			np = nnp;
		}

		if(yy.dsp >= YYINITDEPTH) {
			YYERROR("dvsa overflow");
			return 0;
		}
		yy.dvsa[yy.dsp++] = yylval;

		for(i=0; i<YYNNONTERMS; i++)
			yynode_assign(&yy, &yy.nts[i], 0);
		
#ifdef YYPDEBUG
		YYEPRINTF("After following token:\n");
		yyprintactive(&yy, yy.anp, "active");
		yyprintactive(&yy, yy.rnp, "reduce");
#endif
		
	}
	YYFREE(yy.rstk);
	YYFREE(yy.vsa);
	YYFREE(yy.dvsa);
	
	while(yy.alloc) {
		void *a = yy.alloc;
		yy.alloc = *RCAST(void**)(a);
		YYFREE(a);
	}
	
	if(!success) {
		YYERROR("parse error");
		return -1;
	} else if (success > 1) {
		YYERROR("ambiguous grammar");
		return -1;
	}
#ifdef YYPDEBUG
	YYEPRINTF("active=%p reduce=%p\n", yy.rnp, yy.anp);
	YYEPRINTF("nalloc=%d nfree=%d nnew=%d ndelete=%d ncopy=%d nwrite=%d\n",
		nalloc, nfree, nnew, ndelete, ncopy, nwrite);
	YYEPRINTF("aln=%p maxrsp=%d lostmem=%d\n", aln, maxrsp,
				sizeof(YYNode)*(nalloc-nfree));
	while(aln) {
		yyprintfullnode(&yy, aln);
		aln = aln->np;
	}
#endif
	return 0;
}

#ifdef YYCLASS
int
YYCLASS::YYPARSE(YYPARSEPARAMS)
{
	return static_cast<YYYYCLASS*>(this)->YYPARSE(YYPARSEPARAM);
}
#endif

#line 2472 "CParse.y"
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

